Package: skimr
Title: Compact and Flexible Summaries of Data
Version: 2.2.2
Authors@R: 
    c(person(given = "Elin",
             family = "Waring",
             role = c("cre", "aut"),
             email = "elin.waring@gmail.com"),
      person(given = "Michael",
             family = "Quinn",
             role = "aut",
             email = "msquinn@google.com"),
      person(given = "Amelia",
             family = "McNamara",
             role = "aut",
             email = "amcnamara@smith.edu"),
      person(given = "Eduardo",
             family = "Arino de la Rubia",
             role = "aut",
             email = "earino@gmail.com"),
      person(given = "Hao",
             family = "Zhu",
             role = "aut",
             email = "haozhu233@gmail.com"),
      person(given = "Julia",
             family = "Lowndes",
             role = "ctb",
             email = "lowndes@nceas.ucsb.edu"),
      person(given = "Shannon",
             family = "Ellis",
             role = "aut",
             email = "sellis18@jhmi.edu"),
      person(given = "Hope",
             family = "McLeod",
             role = "ctb",
             email = "hmgit2@gmail.com"),
      person(given = "Hadley",
             family = "Wickham",
             role = "ctb",
             email = "hadley@rstudio.com"),
      person(given = "Kirill",
             family = "Müller",
             role = "ctb",
             email = "krlmlr+r@mailbox.org"),
      person(family = "RStudio, Inc.",
             role = "cph",
             comment = "Spark functions"),
      person(given = "Connor",
             family = "Kirkpatrick",
             role = "ctb",
             email = "hello@connorkirkpatrick.com"),
      person(given = "Scott",
             family = "Brenstuhl",
             role = "ctb",
             email = "brenstsr@miamioh.edu"),
      person(given = "Patrick",
             family = "Schratz",
             role = "ctb",
             email = "patrick.schratz@gmail.com"),
      person(given = "lbusett",
             role = "ctb",
             email = "lbusett@gmail.com"),
      person(given = "Mikko",
             family = "Korpela",
             role = "ctb",
             email = "mvkorpel@iki.fi"),
      person(given = "Jennifer",
             family = "Thompson",
             role = "ctb",
             email = "thompson.jennifer@gmail.com"),
      person(given = "Harris",
             family = "McGehee",
             role = "ctb",
             email = "mcgehee.harris@gmail.com"),
      person(given = "Mark",
             family = "Roepke",
             role = "ctb",
             email = "mroepke5@gmail.com"),
      person(given = "Patrick",
             family = "Kennedy",
             role = "ctb",
             email = "pkqstr@protonmail.com"),
      person(given = "Daniel",
             family = "Possenriede",
             role = "ctb",
             email = "possenriede@gmail.com"),
      person(given = "David",
             family = "Zimmermann",
             role = "ctb",
             email = "david_j_zimmermann@hotmail.com"),
      person(given = "Kyle", 
             family = "Butts",
             role ="ctb",
             email = "buttskyle96@gmail.com"),
      person(given = "Bastian", 
             family = "Torges",
             role ="ctb",
             email = "bastian.torges@gmail.com"),
      person(given = "Rick",
             family = "Saporta",
             role = "ctb",
             email = "Rick@TheFarmersDog.com"),
      person(given = "Henry",
             family = "Morgan Stewart",
             role = "ctb",
             email = "henry.morganstewart@gmail.com"),
      person(given = "Olivier",
             family = "Roy",
             role = "ctb")
       )
Description: A simple to use summary function that can be used with pipes
    and displays nicely in the console. The default summary statistics may
    be modified by the user as can the default formatting.  Support for
    data frames and vectors is included, and users can implement their own
    skim methods for specific object types as described in a vignette.
    Default summaries include support for inline spark graphs.
    Instructions for managing these on specific operating systems are
    given in the "Using skimr" vignette and the README.
License: GPL-3
URL: https://docs.ropensci.org/skimr/,
        https://github.com/ropensci/skimr/
BugReports: https://github.com/ropensci/skimr/issues
Depends: R (>= 4.1.0)
Imports: cli, dplyr (>= 1.0.0), knitr (>= 1.2), pillar (>= 1.6.4),
        purrr, repr, rlang, stats, stringr (>= 1.1), tibble (>= 2.0.0),
        tidyr (>= 1.0), tidyselect (>= 1.0.0), utils, vctrs (>= 0.5.0)
Suggests: data.table, dtplyr, extrafont, haven, lubridate, rmarkdown,
        testthat (>= 3.0.0), withr
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.3
Collate: 'deprecated.R' 'dplyr.R' 'stats.R' 'skim_with.R'
        'get_skimmers.R' 'package.R' 'reshape.R' 'sfl.R' 'skim.R'
        'skim_obj.R' 'skim_print.R' 'skimr-package.R' 'summary.R'
        'utils.R' 'vctrs.R'
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-01-09 17:38:40 UTC; elinwaring
Author: Elin Waring [cre, aut],
  Michael Quinn [aut],
  Amelia McNamara [aut],
  Eduardo Arino de la Rubia [aut],
  Hao Zhu [aut],
  Julia Lowndes [ctb],
  Shannon Ellis [aut],
  Hope McLeod [ctb],
  Hadley Wickham [ctb],
  Kirill Müller [ctb],
  RStudio, Inc. [cph] (Spark functions),
  Connor Kirkpatrick [ctb],
  Scott Brenstuhl [ctb],
  Patrick Schratz [ctb],
  lbusett [ctb],
  Mikko Korpela [ctb],
  Jennifer Thompson [ctb],
  Harris McGehee [ctb],
  Mark Roepke [ctb],
  Patrick Kennedy [ctb],
  Daniel Possenriede [ctb],
  David Zimmermann [ctb],
  Kyle Butts [ctb],
  Bastian Torges [ctb],
  Rick Saporta [ctb],
  Henry Morgan Stewart [ctb],
  Olivier Roy [ctb]
Maintainer: Elin Waring <elin.waring@gmail.com>
Repository: CRAN
Date/Publication: 2026-01-10 08:40:02 UTC
Built: R 4.6.0; ; 2026-04-26 15:58:06 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(ctl_new_pillar,one_skim_df)
S3method(format,summary_skim_df)
S3method(get_skimmers,AsIs)
S3method(get_skimmers,Date)
S3method(get_skimmers,POSIXct)
S3method(get_skimmers,Timespan)
S3method(get_skimmers,character)
S3method(get_skimmers,complex)
S3method(get_skimmers,default)
S3method(get_skimmers,difftime)
S3method(get_skimmers,factor)
S3method(get_skimmers,haven_labelled)
S3method(get_skimmers,list)
S3method(get_skimmers,logical)
S3method(get_skimmers,numeric)
S3method(get_skimmers,ts)
S3method(knit_print,one_skim_df)
S3method(knit_print,skim_df)
S3method(knit_print,skim_list)
S3method(knit_print,summary_skim_df)
S3method(mutate,skim_df)
S3method(print,skim_df)
S3method(print,skim_list)
S3method(print,summary_skim_df)
S3method(repr_text,one_skim_df)
S3method(repr_text,skim_df)
S3method(repr_text,skim_list)
S3method(skim_by_type,data.frame)
S3method(skim_by_type,data.table)
S3method(skim_by_type,grouped_df)
S3method(summary,skim_df)
S3method(tbl_format_header,one_skim_df)
S3method(to_long,default)
S3method(to_long,skim_df)
S3method(vec_cast,skim_df.skim_df)
S3method(vec_cast,skim_df.tbl_df)
S3method(vec_cast,tbl_df.skim_df)
S3method(vec_ptype2,skim_df.skim_df)
S3method(vec_ptype2,skim_df.tbl_df)
S3method(vec_ptype2,tbl_df.skim_df)
export("%>%")
export(assert_is_one_skim_df)
export(assert_is_skim_df)
export(assert_is_skim_list)
export(base_skimmers)
export(bind)
export(complete_rate)
export(contains)
export(could_be_skim_df)
export(data_cols)
export(data_rows)
export(df_name)
export(dt_key)
export(ends_with)
export(everything)
export(fix_windows_histograms)
export(focus)
export(get_default_skimmer_names)
export(get_default_skimmers)
export(get_one_default_skimmer)
export(get_one_default_skimmer_names)
export(get_sfl)
export(get_skimmers)
export(group_names)
export(has_skim_type_attribute)
export(has_skimmers)
export(has_skimr_attributes)
export(has_type_column)
export(has_variable_column)
export(inline_hist)
export(inline_linegraph)
export(is_data_frame)
export(is_one_skim_df)
export(is_skim_df)
export(is_skim_list)
export(list_lengths_max)
export(list_lengths_median)
export(list_lengths_min)
export(list_max_length)
export(list_min_length)
export(matches)
export(max_char)
export(min_char)
export(modify_default_skimmers)
export(n_complete)
export(n_empty)
export(n_missing)
export(n_unique)
export(n_whitespace)
export(num_range)
export(one_of)
export(partition)
export(sfl)
export(skim)
export(skim_format)
export(skim_tee)
export(skim_to_list)
export(skim_to_wide)
export(skim_with)
export(skim_without_charts)
export(skimmers_used)
export(sorted_count)
export(starts_with)
export(to_long)
export(top_counts)
export(ts_end)
export(ts_start)
export(yank)
importFrom(dplyr,mutate)
importFrom(knitr,knit_print)
importFrom(magrittr,"%>%")
importFrom(pillar,ctl_new_pillar)
importFrom(pillar,tbl_format_header)
importFrom(repr,repr_text)
importFrom(rlang,"%||%")
importFrom(rlang,.data)
importFrom(tidyselect,contains)
importFrom(tidyselect,ends_with)
importFrom(tidyselect,everything)
importFrom(tidyselect,matches)
importFrom(tidyselect,num_range)
importFrom(tidyselect,one_of)
importFrom(tidyselect,starts_with)
importFrom(vctrs,vec_cast)
importFrom(vctrs,vec_ptype2)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# skimr <a href='https://docs.ropensci.org/skimr/'>

<img src='https://docs.ropensci.org/skimr/reference/figures/logo.png'
align="right" height="139" /></a>

<!-- badges: start -->

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/)
[![R-CMD-check](https://github.com/ropensci/skimr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ropensci/skimr/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/ropensci/skimr/graph/badge.svg)](https://app.codecov.io/gh/ropensci/skimr)
[![This is an ROpenSci Peer reviewed
package](https://badges.ropensci.org/175_status.svg)](https://github.com/ropensci/software-review/issues/175)
[![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version/skimr)](https://cran.r-project.org/package=skimr)
[![cran
checks](https://badges.cranchecks.info/worst/skimr.svg)](https://badges.cranchecks.info/worst/skimr.svg)
<!-- badges: end -->

`skimr` provides a frictionless approach to summary statistics which
conforms to the [principle of least
surprise](https://en.wikipedia.org/wiki/Principle_of_least_astonishment),
displaying summary statistics the user can skim quickly to understand
their data. It handles different data types and returns a `skim_df`
object which can be included in a pipeline or displayed nicely for the
human reader.

**Note: `skimr` version 2 has major changes when skimr is used
programmatically. Upgraders should review this document, the release
notes and vignettes carefully.**

## Installation

The current released version of `skimr` can be installed from CRAN. If
you wish to install the current build of the next release you can do so
using the following:

    # install.packages("devtools")
    devtools::install_github("ropensci/skimr")

The APIs for this branch should be considered reasonably stable but
still subject to change if an issue is discovered.

To install the version with the most recent changes that have not yet
been incorporated in the main branch (and may not be):

    devtools::install_github("ropensci/skimr", ref = "develop")

Do not rely on APIs from the develop branch, as they are likely to
change.

## Skim statistics in the console

`skimr`:

-   Provides a larger set of statistics than `summary()`, including
    missing, complete, n, and sd.
-   reports each data types separately
-   handles dates, logicals, and a variety of other types
-   supports spark-bar and spark-line based on the [pillar
    package](https://github.com/r-lib/pillar).

### Separates variables by class:

    skim(chickwts)

    ## ── Data Summary ────────────────────────
    ##                            Values  
    ## Name                       chickwts
    ## Number of rows             71      
    ## Number of columns          2       
    ## _______________________            
    ## Column type frequency:             
    ##   factor                   1       
    ##   numeric                  1       
    ## ________________________           
    ## Group variables            None    
    ## 
    ## ── Variable type: factor ───────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate ordered n_unique top_counts                        
    ## 1 feed                  0             1 FALSE          6 soy: 14, cas: 12, lin: 12, sun: 12
    ## 
    ## ── Variable type: numeric ──────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate mean   sd  p0  p25 p50  p75 p100 hist 
    ## 1 weight                0             1 261. 78.1 108 204. 258 324.  423 ▆▆▇▇▃

### Presentation is in a compact horizontal format:

    skim(iris)

    ## ── Data Summary ────────────────────────
    ##                            Values
    ## Name                       iris  
    ## Number of rows             150   
    ## Number of columns          5     
    ## _______________________          
    ## Column type frequency:           
    ##   factor                   1     
    ##   numeric                  4     
    ## ________________________         
    ## Group variables            None  
    ## 
    ## ── Variable type: factor ───────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate ordered n_unique top_counts               
    ## 1 Species               0             1 FALSE          3 set: 50, ver: 50, vir: 50
    ## 
    ## ── Variable type: numeric ──────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate mean    sd  p0 p25  p50 p75 p100 hist 
    ## 1 Sepal.Length          0             1 5.84 0.828 4.3 5.1 5.8  6.4  7.9 ▆▇▇▅▂
    ## 2 Sepal.Width           0             1 3.06 0.436 2   2.8 3    3.3  4.4 ▁▆▇▂▁
    ## 3 Petal.Length          0             1 3.76 1.77  1   1.6 4.35 5.1  6.9 ▇▁▆▇▂
    ## 4 Petal.Width           0             1 1.20 0.762 0.1 0.3 1.3  1.8  2.5 ▇▁▇▅▃

### Built in support for strings, lists and other column classes

    skim(dplyr::starwars)

    ## ── Data Summary ────────────────────────
    ##                            Values         
    ## Name                       dplyr::starwars
    ## Number of rows             87             
    ## Number of columns          14             
    ## _______________________                   
    ## Column type frequency:                    
    ##   character                8              
    ##   list                     3              
    ##   numeric                  3              
    ## ________________________                  
    ## Group variables            None           
    ## 
    ## ── Variable type: character ────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate min max empty n_unique whitespace
    ## 1 name                  0         1       3  21     0       87          0
    ## 2 hair_color            5         0.943   4  13     0       11          0
    ## 3 skin_color            0         1       3  19     0       31          0
    ## 4 eye_color             0         1       3  13     0       15          0
    ## 5 sex                   4         0.954   4  14     0        4          0
    ## 6 gender                4         0.954   8   9     0        2          0
    ## 7 homeworld            10         0.885   4  14     0       48          0
    ## 8 species               4         0.954   3  14     0       37          0
    ## 
    ## ── Variable type: list ─────────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate n_unique min_length max_length
    ## 1 films                 0             1       24          1          7
    ## 2 vehicles              0             1       11          0          2
    ## 3 starships             0             1       16          0          5
    ## 
    ## ── Variable type: numeric ──────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate  mean    sd p0   p25 p50   p75 p100 hist 
    ## 1 height                6         0.931 175.   34.8 66 167   180 191    264 ▂▁▇▅▁
    ## 2 mass                 28         0.678  97.3 169.  15  55.6  79  84.5 1358 ▇▁▁▁▁
    ## 3 birth_year           44         0.494  87.6 155.   8  35    52  72    896 ▇▁▁▁▁

### Has a useful summary function

    skim(iris) |>
      summary()

    ## ── Data Summary ────────────────────────
    ##                            Values
    ## Name                       iris  
    ## Number of rows             150   
    ## Number of columns          5     
    ## _______________________          
    ## Column type frequency:           
    ##   factor                   1     
    ##   numeric                  4     
    ## ________________________         
    ## Group variables            None

### Individual columns can be selected using tidyverse-style selectors

    skim(iris, Sepal.Length, Petal.Length)

    ## ── Data Summary ────────────────────────
    ##                            Values
    ## Name                       iris  
    ## Number of rows             150   
    ## Number of columns          5     
    ## _______________________          
    ## Column type frequency:           
    ##   numeric                  2     
    ## ________________________         
    ## Group variables            None  
    ## 
    ## ── Variable type: numeric ──────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate mean    sd  p0 p25  p50 p75 p100 hist 
    ## 1 Sepal.Length          0             1 5.84 0.828 4.3 5.1 5.8  6.4  7.9 ▆▇▇▅▂
    ## 2 Petal.Length          0             1 3.76 1.77  1   1.6 4.35 5.1  6.9 ▇▁▆▇▂

### Handles grouped data

`skim()` can handle data that has been grouped using
`dplyr::group_by()`.

    iris |>
      dplyr::group_by(Species) |>
      skim()

    ## ── Data Summary ────────────────────────
    ##                            Values                      
    ## Name                       dplyr::group_by(iris, Spe...
    ## Number of rows             150                         
    ## Number of columns          5                           
    ## _______________________                                
    ## Column type frequency:                                 
    ##   numeric                  4                           
    ## ________________________                               
    ## Group variables            Species                     
    ## 
    ## ── Variable type: numeric ──────────────────────────────────────────────────────────────────────────
    ##    skim_variable Species    n_missing complete_rate  mean    sd  p0  p25  p50  p75 p100 hist 
    ##  1 Sepal.Length  setosa             0             1 5.01  0.352 4.3 4.8  5    5.2   5.8 ▃▃▇▅▁
    ##  2 Sepal.Length  versicolor         0             1 5.94  0.516 4.9 5.6  5.9  6.3   7   ▂▇▆▃▃
    ##  3 Sepal.Length  virginica          0             1 6.59  0.636 4.9 6.22 6.5  6.9   7.9 ▁▃▇▃▂
    ##  4 Sepal.Width   setosa             0             1 3.43  0.379 2.3 3.2  3.4  3.68  4.4 ▁▃▇▅▂
    ##  5 Sepal.Width   versicolor         0             1 2.77  0.314 2   2.52 2.8  3     3.4 ▁▅▆▇▂
    ##  6 Sepal.Width   virginica          0             1 2.97  0.322 2.2 2.8  3    3.18  3.8 ▂▆▇▅▁
    ##  7 Petal.Length  setosa             0             1 1.46  0.174 1   1.4  1.5  1.58  1.9 ▁▃▇▃▁
    ##  8 Petal.Length  versicolor         0             1 4.26  0.470 3   4    4.35 4.6   5.1 ▂▂▇▇▆
    ##  9 Petal.Length  virginica          0             1 5.55  0.552 4.5 5.1  5.55 5.88  6.9 ▃▇▇▃▂
    ## 10 Petal.Width   setosa             0             1 0.246 0.105 0.1 0.2  0.2  0.3   0.6 ▇▂▂▁▁
    ## 11 Petal.Width   versicolor         0             1 1.33  0.198 1   1.2  1.3  1.5   1.8 ▅▇▃▆▁
    ## 12 Petal.Width   virginica          0             1 2.03  0.275 1.4 1.8  2    2.3   2.5 ▂▇▆▅▇

### Behaves nicely in pipelines

    iris |>
      skim() |>
      dplyr::filter(numeric.sd > 1)

    ## ── Data Summary ────────────────────────
    ##                            Values
    ## Name                       iris  
    ## Number of rows             150   
    ## Number of columns          5     
    ## _______________________          
    ## Column type frequency:           
    ##   numeric                  1     
    ## ________________________         
    ## Group variables            None  
    ## 
    ## ── Variable type: numeric ──────────────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate mean   sd p0 p25  p50 p75 p100 hist 
    ## 1 Petal.Length          0             1 3.76 1.77  1 1.6 4.35 5.1  6.9 ▇▁▆▇▂

## Knitted results

Simply skimming a data frame will produce the horizontal print layout
shown above. We provide a `knit_print` method for the types of objects
in this package so that similar results are produced in documents. To
use this, make sure the `skimmed` object is the last item in your code
chunk.

    faithful |>
      skim()

<table>
<caption>Data summary</caption>
<tbody>
<tr class="odd">
<td style="text-align: left;">Name</td>
<td style="text-align: left;">faithful</td>
</tr>
<tr class="even">
<td style="text-align: left;">Number of rows</td>
<td style="text-align: left;">272</td>
</tr>
<tr class="odd">
<td style="text-align: left;">Number of columns</td>
<td style="text-align: left;">2</td>
</tr>
<tr class="even">
<td style="text-align: left;">_______________________</td>
<td style="text-align: left;"></td>
</tr>
<tr class="odd">
<td style="text-align: left;">Column type frequency:</td>
<td style="text-align: left;"></td>
</tr>
<tr class="even">
<td style="text-align: left;">numeric</td>
<td style="text-align: left;">2</td>
</tr>
<tr class="odd">
<td style="text-align: left;">________________________</td>
<td style="text-align: left;"></td>
</tr>
<tr class="even">
<td style="text-align: left;">Group variables</td>
<td style="text-align: left;">None</td>
</tr>
</tbody>
</table>

Data summary

**Variable type: numeric**

<table style="width:100%;">
<colgroup>
<col style="width: 17%" />
<col style="width: 12%" />
<col style="width: 17%" />
<col style="width: 7%" />
<col style="width: 7%" />
<col style="width: 6%" />
<col style="width: 7%" />
<col style="width: 4%" />
<col style="width: 7%" />
<col style="width: 6%" />
<col style="width: 7%" />
</colgroup>
<thead>
<tr class="header">
<th style="text-align: left;">skim_variable</th>
<th style="text-align: right;">n_missing</th>
<th style="text-align: right;">complete_rate</th>
<th style="text-align: right;">mean</th>
<th style="text-align: right;">sd</th>
<th style="text-align: right;">p0</th>
<th style="text-align: right;">p25</th>
<th style="text-align: right;">p50</th>
<th style="text-align: right;">p75</th>
<th style="text-align: right;">p100</th>
<th style="text-align: left;">hist</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td style="text-align: left;">eruptions</td>
<td style="text-align: right;">0</td>
<td style="text-align: right;">1</td>
<td style="text-align: right;">3.49</td>
<td style="text-align: right;">1.14</td>
<td style="text-align: right;">1.6</td>
<td style="text-align: right;">2.16</td>
<td style="text-align: right;">4</td>
<td style="text-align: right;">4.45</td>
<td style="text-align: right;">5.1</td>
<td style="text-align: left;">▇▂▂▇▇</td>
</tr>
<tr class="even">
<td style="text-align: left;">waiting</td>
<td style="text-align: right;">0</td>
<td style="text-align: right;">1</td>
<td style="text-align: right;">70.90</td>
<td style="text-align: right;">13.59</td>
<td style="text-align: right;">43.0</td>
<td style="text-align: right;">58.00</td>
<td style="text-align: right;">76</td>
<td style="text-align: right;">82.00</td>
<td style="text-align: right;">96.0</td>
<td style="text-align: left;">▃▃▂▇▂</td>
</tr>
</tbody>
</table>

## Customizing skimr

Although skimr provides opinionated defaults, it is highly customizable.
Users can specify their own statistics, change the formatting of
results, create statistics for new classes and develop skimmers for data
structures that are not data frames.

### Specify your own statistics and classes

Users can specify their own statistics using a list combined with the
`skim_with()` function factory. `skim_with()` returns a new `skim`
function that can be called on your data. You can use this factory to
produce summaries for any type of column within your data.

Assignment within a call to `skim_with()` relies on a helper function,
`sfl` or `skimr` function list. By default, functions in the `sfl` call
are appended to the default skimmers, and names are automatically
generated as well.

    my_skim <- skim_with(numeric = sfl(mad))
    my_skim(iris, Sepal.Length)

But you can also helpers from the `tidyverse` to create new anonymous
functions that set particular function arguments. The behavior is the
same as in `purrr` or `dplyr`, with both `.` and `.x` as acceptable
pronouns. Setting the `append = FALSE` argument uses only those
functions that you’ve provided.

    my_skim <- skim_with(
      numeric = sfl(
        iqr = IQR,
        p01 = ~ quantile(.x, probs = .01)
        p99 = ~ quantile(., probs = .99)
      ),
      append = FALSE
    )
    my_skim(iris, Sepal.Length)

And you can remove default skimmers by setting them to `NULL`.

    my_skim <- skim_with(numeric = sfl(hist = NULL))
    my_skim(iris, Sepal.Length)

### Skimming other objects

`skimr` has summary functions for the following types of data by
default:

-   `numeric` (which includes both `double` and `integer`)
-   `character`
-   `factor`
-   `logical`
-   `complex`
-   `Date`
-   `POSIXct`
-   `ts`
-   `AsIs`

`skimr` also provides a small API for writing packages that provide
their own default summary functions for data types not covered above. It
relies on R S3 methods for the `get_skimmers` function. This function
should return a `sfl`, similar to customization within `skim_with()`,
but you should also provide a value for the `class` argument. Here’s an
example.

    get_skimmers.my_data_type <- function(column) {
      sfl(
        .class = "my_data_type",
        p99 = quantile(., probs = .99)
      )
    }

## Limitations of current version

We are aware that there are issues with rendering the inline histograms
and line charts in various contexts, some of which are described below.

### Support for spark histograms

With versions of R before 4.2.1, there are known issues with printing
the spark-histogram characters when printing a data frame. For example,
`"▂▅▇"` is printed as `"<U+2582><U+2585><U+2587>"`. This longstanding
problem [originates in the low-level
code](https://stat.ethz.ch/pipermail/r-devel/2015-May/071250.html) for
printing dataframes. While some cases have been addressed, there are,
for example, reports of this issue in Emacs ESS. While this is a deep
issue, there is [ongoing work to address it in base
R](https://blog.r-project.org/2020/05/02/utf-8-support-on-windows/). We
recommend upgrading to at least R 4.2.1 to address this issue.

This means that while `skimr` can render the histograms to the console
and in RMarkdown documents, it cannot in other circumstances. This
includes:

-   converting a `skimr` data frame to a vanilla R data frame, but
    tibbles render correctly
-   in the context of rendering to a pdf using an engine that does not
    support utf-8.

One workaround for showing these characters in Windows is to set the
CTYPE part of your locale to Chinese/Japanese/Korean with
`Sys.setlocale("LC_CTYPE", "Chinese")`. The helper function
`fix_windows_histograms()` does this for you.

And last but not least, we provide `skim_without_charts()` as a
fallback. This makes it easy to still get summaries of your data, even
if unicode issues continue.

### Printing spark histograms and line graphs in knitted documents

Spark-bar and spark-line work in the console, but may not work when you
knit them to a specific document format. The same session that produces
a correctly rendered HTML document may produce an incorrectly rendered
PDF, for example. This issue can generally be addressed by changing
fonts to one with good building block (for histograms) and Braille
support (for line graphs). For example, the open font “DejaVu Sans” from
the `extrafont` package supports these. You may also want to try
wrapping your results in `knitr::kable()`. Please see the vignette on
using fonts for details.

Displays in documents of different types will vary. For example, one
user found that the font “Yu Gothic UI Semilight” produced consistent
results for Microsoft Word and Libre Office Write.

## Inspirations

-   [TextPlots](https://github.com/sunetos/TextPlots.jl) for use of
    Braille characters

-   [spark](https://github.com/holman/spark) for use of block
    characters.

The earliest use of unicode characters to generate sparklines appears to
be [from
2009](https://blog.jonudell.net/2009/01/13/fuel-prices-and-pageviews/).

Exercising these ideas to their fullest requires a font with good
support for block drawing characters.
[PragamataPro](https://fsd.it/shop/fonts/pragmatapro/) is one such font.

## Contributing

We welcome issue reports and pull requests, including potentially adding
support for commonly used variable classes. However, in general, we
encourage users to take advantage of skimr’s flexibility to add their
own customized classes. Please see the
[contributing](https://docs.ropensci.org/skimr/CONTRIBUTING.html) and
[conduct](https://ropensci.org/code-of-conduct/) documents.

[![ropenci\_footer](https://ropensci.org/public_images/ropensci_footer.png)](https://ropensci.org)

# Simpler expectations for data frames.

expect_n_columns <- function(object, n) {
  stopifnot(is.numeric(n), length(n) == 1)
  act <- testthat::quasi_label(rlang::enquo(object))
  act$n <- length(act$val)
  testthat::expect(act$n == n, sprintf(
    "%s has %i columns, not %i columns.",
    act$lab, act$n, n
  ))
  invisible(act$val)
}

expect_n_rows <- function(object, n) {
  stopifnot(is.numeric(n), length(n) == 1)
  act <- testthat::quasi_label(rlang::enquo(object))
  act$n <- nrow(act$val)
  testthat::expect(
    act$n == n,
    sprintf("%s has %i rows, not %i rows", act$lab, act$n, n)
  )
  invisible(act$val)
}

expect_NA <- function(object) {
  act <- testthat::quasi_label(rlang::enquo(object))
  testthat::expect(is.na(act$val), sprintf("%s is not NA", act$lab))
  invisible(act$val)
}

test_that("skim returns expected response for numeric vectors", {
  input <- skim(mtcars, mpg)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 12)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "numeric.mean", "numeric.sd", "numeric.p0", "numeric.p25",
    "numeric.p50", "numeric.p75", "numeric.p100", "numeric.hist"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 32)
  expect_equal(attrs$data_cols, 11)
  expect_equal(attrs$df_name, "`mtcars`")
  expect_equal(
    attrs$skimmers_used,
    list(
      numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist")
    )
  )

  # values
  expect_identical(unname(input$skim_variable), "mpg")
  expect_identical(input$skim_type, "numeric")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1)
  expect_equal(input$numeric.mean, 20.1, tolerance = 0.1)
  expect_equal(input$numeric.p0, 10.4, tolerance = 0.1)
  expect_equal(input$numeric.p25, 15.4, tolerance = 0.1)
  expect_equal(input$numeric.p50, 19.2, tolerance = 0.1)
  expect_equal(input$numeric.p75, 22.8, tolerance = 0.1)
  expect_equal(input$numeric.p100, 33.9, tolerance = 0.1)
  skip_on_os("windows")
  skip_if_not(l10n_info()$`UTF-8`)
  expect_identical(input$numeric.hist, "▃▇▅▁▂")
})

test_that("skim handles numeric vectors with NAs and extreme numbers", {
  patho <- tibble::tibble(
    vals = c((2^.Machine$double.digits), NA, -(2^.Machine$double.digits))
  )

  input <- skim(patho)

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 3)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`patho`")

  # values
  expect_identical(input$skim_variable, "vals")
  expect_identical(input$skim_type, "numeric")
  expect_identical(input$n_missing, 1L)
  expect_equal(input$complete_rate, .667, tolerance = .001)
  expect_equal(input$numeric.mean, 0)
  expect_equal(input$numeric.sd, 1.27e16, tolerance = 1e14)
  expect_equal(input$numeric.p0, -9.01e+15, tolerance = 1e14)
  expect_equal(input$numeric.p25, -4.5e+15, tolerance = 1e14)
  expect_equal(input$numeric.p50, 0)
  expect_equal(input$numeric.p75, 4.5e+15, tolerance = 1e14)
  expect_equal(input$numeric.p100, 9.01e+15, tolerance = 1e14)
  skip_on_os("windows")
  skip_if_not(l10n_info()$`UTF-8`)
  expect_identical(input$numeric.hist, "▇▁▁▁▇")
})

test_that("numeric skim is calculated correctly when x is all NAs.", {
  nas <- as.numeric(c(NA, NA, NA))
  x <- tibble::tibble(nas)
  input <- skim(x)

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 3)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`x`")

  # values
  expect_identical(input$skim_variable, "nas")
  expect_identical(input$skim_type, "numeric")
  expect_identical(input$n_missing, 3L)
  expect_identical(input$complete_rate, 0)
  expect_equal(input$numeric.mean, NaN)
  expect_equal(input$numeric.sd, NaN)
  expect_NA(input$numeric.p0)
  expect_NA(input$numeric.p25)
  expect_NA(input$numeric.p50)
  expect_NA(input$numeric.p75)
  expect_NA(input$numeric.p100)
  skip_on_os("windows")
  skip_if_not(l10n_info()$`UTF-8`)
  expect_identical(input$numeric.hist, " ")
})

test_that("numeric skim is calculated correctly when x is all zeores or NAs.", {
  na_and_zeros <- as.numeric(c(NA, NA, NA, 0))
  x <- tibble::tibble(na_and_zeros)
  input <- skim(x)

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 4)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`x`")

  # values
  expect_identical(input$skim_variable, "na_and_zeros")
  expect_identical(input$skim_type, "numeric")
  expect_identical(input$n_missing, 3L)
  expect_equal(input$complete_rate, .25)
  expect_equal(input$numeric.mean, 0)
  expect_NA(input$numeric.sd)
  expect_equal(input$numeric.p0, 0)
  expect_equal(input$numeric.p25, 0)
  expect_equal(input$numeric.p50, 0)
  expect_equal(input$numeric.p75, 0)
  expect_equal(input$numeric.p100, 0)
  skip_on_os("windows")
  skip_if_not(l10n_info()$`UTF-8`)
  expect_identical(input$numeric.hist, "▁▁▇▁▁")
})

test_that("Skimming with non-finite values works", {
  inf_vals <- c(Inf, 0, -Inf)
  x <- tibble::tibble(inf_vals)
  expect_warning(input <- skim(x))

  expect_identical(input$skim_variable, "inf_vals")
  expect_identical(input$skim_type, "numeric")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1)
  expect_equal(input$numeric.mean, NaN)
  expect_equal(input$numeric.sd, NaN)
  expect_equal(input$numeric.p0, -Inf)
  expect_equal(input$numeric.p25, -Inf)
  expect_equal(input$numeric.p50, 0)
  expect_equal(input$numeric.p75, Inf)
  expect_equal(input$numeric.p100, Inf)
  skip_on_os("windows")
  skip_if_not(l10n_info()$`UTF-8`)
  expect_identical(input$numeric.hist, "▁▁▇▁▁")
})

test_that("skim returns expected response for factor vectors", {
  input <- skim(iris, Species)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 7)
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "factor.ordered", "factor.n_unique", "factor.top_counts"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 150)
  expect_equal(attrs$data_cols, 5)
  expect_equal(attrs$df_name, "`iris`")
  expect_equal(
    attrs$skimmers_used,
    list(factor = c("ordered", "n_unique", "top_counts"))
  )

  # values
  expect_identical(unname(input$skim_variable), "Species")
  expect_identical(input$skim_type, "factor")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1)
  expect_false(input$factor.ordered)
  expect_identical(input$factor.top_counts, "set: 50, ver: 50, vir: 50")
})

test_that("skim handles factors when NAs are present", {
  dat <- iris
  dat$Species[15:18] <- NA
  input <- skim(dat, Species)
  expect_identical(input$n_missing, 4L)
  expect_equal(input$complete_rate, .973, tolerance = .001)
  expect_identical(input$factor.top_counts, "ver: 50, vir: 50, set: 46")
})


test_that("skim returns expected response for character vectors", {
  dat <- c("AAAB", "ABc", "acb", NA, "", " ")
  x <- tibble::tibble(dat)
  input <- skim(x)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 9)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "character.min", "character.max", "character.empty",
    "character.n_unique", "character.whitespace"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 6)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`x`")
  expect_equal(
    attrs$skimmers_used,
    list(character = c("min", "max", "empty", "n_unique", "whitespace"))
  )

  # values
  expect_identical(input$skim_variable, "dat")
  expect_identical(input$skim_type, "character")
  expect_identical(input$n_missing, 1L)
  expect_equal(input$complete_rate, .833, tolerance = .001)
  expect_identical(input$character.min, 0L)
  expect_identical(input$character.max, 4L)
  expect_identical(input$character.empty, 1L)
  expect_identical(input$character.n_unique, 5L)
  expect_identical(input$character.whitespace, 1L)
})

test_that("skim returns expected response for logical vectors", {
  dat <- dplyr::mutate(chickwts, log_col = stringr::str_detect(feed, "ea"))
  input <- skim(dat, log_col)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 6)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "logical.mean", "logical.count"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 71)
  expect_equal(attrs$data_cols, 3)
  expect_equal(attrs$df_name, "`dat`")
  expect_equal(
    attrs$skimmers_used,
    list(logical = c("mean", "count"))
  )

  # values
  expect_identical(unname(input$skim_variable), "log_col")
  expect_identical(input$skim_type, "logical")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1)
  expect_equal(input$logical.mean, 0.49, tolerance = .01)
  expect_identical(input$logical.count, "FAL: 36, TRU: 35")
})

test_that("skim returns expected response for logical vectors with NA values", {
  dat <- dplyr::mutate(chickwts, log_col = stringr::str_detect(feed, "ea"))
  dat$log_col[15:18] <- NA
  input <- skim(dat, log_col)

  expect_identical(input$n_missing, 4L)
  expect_equal(input$complete_rate, .944, tolerance = .001)
  expect_equal(input$logical.mean, 0.52, tolerance = .01)
  expect_identical(input$logical.count, "TRU: 35, FAL: 32")
})

test_that("skim returns expected response for complex vectors", {
  dat <- dplyr::mutate(chickwts, test_complex = weight)
  dat$test_complex[1:2] <- dat$test_complex[1:2] + 2i
  dat$test_complex[15:18] <- NA_complex_
  input <- skim(dat, test_complex)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 5)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(
    input,
    c(
      "skim_type", "skim_variable", "n_missing", "complete_rate", "complex.mean"
    )
  )

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 71)
  expect_equal(attrs$data_cols, 3)
  expect_equal(attrs$df_name, "`dat`")
  expect_equal(
    attrs$skimmers_used,
    list(complex = "mean")
  )

  # values
  expect_identical(unname(input$skim_variable), "test_complex")
  expect_identical(input$skim_type, "complex")
  expect_identical(input$n_missing, 4L)
  expect_equal(input$complete_rate, .944, tolerance = .001)
  expect_equal(input$complex.mean, 265.2687 + 0.0597i, tolerance = .001)
})

test_that("skim returns expected response for Date vectors", {
  dat <- seq(as.Date("2011-07-01"), by = 1, len = 9)
  x <- tibble::tibble(dat)
  input <- skim(x)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 8)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "Date.min", "Date.max", "Date.median", "Date.n_unique"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 9)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`x`")
  expect_equal(
    attrs$skimmers_used,
    list(Date = c("min", "max", "median", "n_unique"))
  )

  # values
  expect_identical(input$skim_variable, "dat")
  expect_identical(input$skim_type, "Date")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1)
  expect_equal(input$Date.min, as.Date("2011-07-01"))
  expect_equal(input$Date.max, as.Date("2011-07-09"))
  expect_equal(input$Date.median, as.Date("2011-07-05"))
  expect_identical(input$Date.n_unique, 9L)
})

test_that("skim returns expected response for ts vectors", {
  input <- skim(freeny, y)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 14)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "ts.start", "ts.end", "ts.frequency", "ts.deltat", "ts.mean", "ts.sd",
    "ts.min", "ts.max", "ts.median", "ts.line_graph"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 39)
  expect_equal(attrs$data_cols, 5)
  expect_equal(attrs$df_name, "`freeny`")
  expect_equal(
    attrs$skimmers_used,
    list(ts = c(
      "start", "end", "frequency", "deltat", "mean", "sd", "min",
      "max", "median", "line_graph"
    ))
  )

  # values
  expect_identical(unname(input$skim_variable), "y")
  expect_identical(input$skim_type, "ts")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1)
  expect_equal(input$ts.start, 1962)
  expect_equal(input$ts.end, 1971)
  expect_equal(input$ts.frequency, 4)
  expect_equal(input$ts.deltat, 0.25)
  expect_equal(input$ts.mean, 9.31, tolerance = 0.001)
  expect_equal(input$ts.sd, 0.316, tolerance = 0.001)
  expect_equal(input$ts.min, 8.79, tolerance = 0.001)
  expect_equal(input$ts.max, 9.79, tolerance = 0.001)
  expect_equal(input$ts.median, 9.31, tolerance = 0.001)
  expect_identical(input$ts.line_graph, "⣀⣀⠤⠤⠒⠒⠉⠉")
})

test_that("skim returns expected response for POSIXct vectors", {
  dat <- seq(as.POSIXct("2011-07-01 00:00:00", tz = "UTC"), by = 1, len = 10)
  dat[2] <- NA
  posix <- tibble::tibble(dat)
  input <- skim(posix)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 8)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "POSIXct.min", "POSIXct.max", "POSIXct.median",
    "POSIXct.n_unique"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 10)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`posix`")
  expect_equal(
    attrs$skimmers_used,
    list(POSIXct = c("min", "max", "median", "n_unique"))
  )

  # values
  expect_identical(input$skim_variable, "dat")
  expect_identical(input$skim_type, "POSIXct")
  expect_identical(input$n_missing, 1L)
  expect_equal(input$complete_rate, .9, tolerance = .001)
  expect_identical(
    input$POSIXct.min,
    as.POSIXct("2011-07-01 00:00:00", tz = "UTC")
  )
  expect_identical(
    input$POSIXct.max,
    as.POSIXct("2011-07-01 00:00:09", tz = "UTC")
  )
  expect_identical(
    input$POSIXct.median,
    as.POSIXct("2011-07-01 00:00:05", tz = "UTC")
  )
  expect_identical(input$POSIXct.n_unique, 9L)
})

test_that("skim returns expected response for list (not AsIs) vectors", {
  dat <- list(
    list("a", "b", "c"),
    list("d", "b", "d"),
    list("e", "f", "g"),
    d = list("h"),
    e = list("i", "j", "k", "l"),
    f = NA
  )
  list_frame <- tibble::tibble(dat)
  input <- skim(list_frame)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 7)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "list.n_unique", "list.min_length", "list.max_length"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 6)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`list_frame`")
  expect_equal(
    attrs$skimmers_used,
    list(list = c("n_unique", "min_length", "max_length"))
  )

  # values
  expect_identical(input$skim_variable, "dat")
  expect_identical(input$skim_type, "list")
  expect_identical(input$n_missing, 1L)
  expect_equal(input$complete_rate, .833, tolerance = .001)
  expect_identical(input$list.n_unique, 5L)
  expect_identical(input$list.min_length, 1L)
  expect_identical(input$list.max_length, 4L)
})

test_that("skim returns expected response for list with all NA's", {
  dat <- list(NA, NA, NA)
  list_frame <- tibble::tibble(dat)
  input <- skim(list_frame)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 7)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "list.n_unique", "list.min_length", "list.max_length"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 3)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`list_frame`")
  expect_equal(
    attrs$skimmers_used,
    list(list = c("n_unique", "min_length", "max_length"))
  )

  # values
  expect_identical(input$skim_variable, "dat")
  expect_identical(input$skim_type, "list")
  expect_identical(input$n_missing, 3L)
  expect_equal(input$complete_rate, 0)
  expect_identical(input$list.n_unique, 0L)
  expect_identical(input$list.min_length, 1L)
  expect_identical(input$list.max_length, 1L)
})

test_that("skim returns expected response for asis vectors", {
  dat <- I(list(5, 5:6, 5:10, NA))
  asis_frame <- tibble::tibble(dat)
  input <- skim(asis_frame)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 7)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "AsIs.n_unique", "AsIs.min_length", "AsIs.max_length"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 4)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`asis_frame`")
  expect_equal(
    attrs$skimmers_used,
    list(AsIs = c("n_unique", "min_length", "max_length"))
  )

  # values
  expect_identical(input$skim_variable, "dat")
  expect_identical(input$skim_type, "AsIs")
  expect_identical(input$n_missing, 1L)
  expect_equal(input$complete_rate, .75)
  expect_identical(input$AsIs.n_unique, 3L)
  expect_identical(input$AsIs.min_length, 1L)
  expect_identical(input$AsIs.max_length, 6L)
})

test_that("skim returns expected response for difftime vectors", {
  dat_datetime <- as.POSIXct("2011-07-01 00:00:00", tz = "UTC")
  dat <- difftime(dat_datetime, (dat_datetime - seq(-30, 60, 10)))
  dat[2] <- NA
  dt <- tibble::tibble(dat)
  input <- skim(dt)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 8)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "difftime.min", "difftime.max", "difftime.median",
    "difftime.n_unique"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 10)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`dt`")
  expect_equal(
    attrs$skimmers_used,
    list(difftime = c("min", "max", "median", "n_unique"))
  )

  # values
  expect_identical(input$skim_variable, "dat")
  expect_identical(input$skim_type, "difftime")
  expect_identical(input$n_missing, 1L)
  expect_equal(input$complete_rate, .9, tolerance = .001)
  expect_identical(input$difftime.min, as.difftime(-30, units = "secs"))
  expect_identical(input$difftime.max, as.difftime(60, units = "secs"))
  expect_identical(input$difftime.median, as.difftime(20, units = "secs"))
  expect_identical(input$difftime.n_unique, 9L)
})

test_that("skim returns expected response for lubridate Timespan vectors", {
  skip_if_not_installed("lubridate")
  dt <- tibble::tibble(x = lubridate::duration(1))
  input <- skim(dt)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 8)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "Timespan.min", "Timespan.max", "Timespan.median",
    "Timespan.n_unique"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 1)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`dt`")
  expect_equal(
    attrs$skimmers_used,
    list(Timespan = c("min", "max", "median", "n_unique"))
  )

  # values
  expect_identical(input$skim_variable, "x")
  expect_identical(input$skim_type, "Timespan")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1, tolerance = .001)
  expect_identical(input$Timespan.min, 1)
  expect_identical(input$Timespan.max, 1)
  expect_identical(input$Timespan.median, lubridate::duration(1))
  expect_identical(input$Timespan.n_unique, 1L)
})

test_that("skim handles objects containing haven_labelled vectors: double", {
  skip_if_not_installed("haven")
  dt <- tibble::tibble(x = haven::labelled(c(1, 2, 3), c(A = 1, B = 2, C = 3)))
  input <- skimr::skim(dt)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 12)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "numeric.mean", "numeric.sd", "numeric.p0", "numeric.p25",
    "numeric.p50", "numeric.p75", "numeric.p100", "numeric.hist"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 3)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`dt`")
  expect_equal(
    attrs$skimmers_used,
    list(
      numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist")
    )
  )

  # values
  expect_identical(input$skim_type, "numeric")
  expect_identical(input$skim_variable, "x")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1)
  expect_equal(input$numeric.mean, 2, tolerance = 0.1)
  expect_equal(input$numeric.sd, 1, tolerance = 0.1)
  expect_equal(input$numeric.p0, 1, tolerance = 0.1)
  expect_equal(input$numeric.p25, 1.5, tolerance = 0.1)
  expect_equal(input$numeric.p50, 2, tolerance = 0.1)
  expect_equal(input$numeric.p75, 2.5, tolerance = 0.1)
  expect_equal(input$numeric.p100, 3, tolerance = 0.1)
})

test_that("skim handles objects containing haven_labelled vectors: character", {
  skip_if_not_installed("haven")
  dt <- tibble::tibble(x = haven::labelled(
    c("a", "b", "c"),
    c(A = "a", B = "b", C = "c")
  ))
  input <- skimr::skim(dt)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 9)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "character.min", "character.max", "character.empty",
    "character.n_unique", "character.whitespace"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 3)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`dt`")
  expect_equal(
    attrs$skimmers_used,
    list(character = c("min", "max", "empty", "n_unique", "whitespace"))
  )

  # values
  expect_identical(input$skim_variable, "x")
  expect_identical(input$skim_type, "character")
  expect_identical(input$n_missing, 0L)
  expect_equal(input$complete_rate, 1., tolerance = .001)
  expect_identical(input$character.min, 1L)
  expect_identical(input$character.max, 1L)
  expect_identical(input$character.empty, 0L)
  expect_identical(input$character.n_unique, 3L)
  expect_identical(input$character.whitespace, 0L)
})

test_that("skim handles objects with multiple classes", {
  dat <- seq(as.Date("2011-07-01"), by = 1, len = 10)
  dat[2] <- NA
  class(dat) <- c("strange_type", "Date")
  x <- tibble::tibble(dat)
  input <- skim(x)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 8)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "Date.min", "Date.max", "Date.median", "Date.n_unique"
  ))

  # values
  expect_identical(input$skim_variable, "dat")
  expect_identical(input$skim_type, "Date")
})

test_that("skim treats unknown classes as character", {
  weight <- chickwts$weight
  x <- tibble::tibble(weight)
  class(x$weight) <- "strange"
  expect_warning(input <- skim(x))

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "character.min", "character.max", "character.empty",
    "character.n_unique", "character.whitespace"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 71)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`x`")
  expect_equal(
    attrs$skimmers_used,
    list(character = c("min", "max", "empty", "n_unique", "whitespace"))
  )

  # values
  expect_identical(input$skim_variable, "weight")
  expect_identical(input$skim_type, "character")
})

test_that("skim handles objects with two unknown classes", {
  feed <- chickwts$feed
  x <- tibble::tibble(feed)
  class(x$feed) <- c("strange", "stranger")
  expect_warning(input <- skim(x))

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "character.min", "character.max", "character.empty",
    "character.n_unique", "character.whitespace"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 71)
  expect_equal(attrs$data_cols, 1)
  expect_equal(attrs$df_name, "`x`")
  expect_equal(
    attrs$skimmers_used,
    list(character = c("min", "max", "empty", "n_unique", "whitespace"))
  )

  # values
  expect_identical(input$skim_variable, "feed")
  expect_identical(input$skim_type, "character")
})

test_that("Skimming a complete data frame works as expected", {
  input <- skim(chickwts)

  # dimensions
  expect_n_rows(input, 2)
  expect_n_columns(input, 15)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "n_missing", "complete_rate",
    "factor.ordered", "factor.n_unique", "factor.top_counts",
    "numeric.mean", "numeric.sd", "numeric.p0", "numeric.p25", "numeric.p50",
    "numeric.p75", "numeric.p100", "numeric.hist"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 71)
  expect_equal(attrs$data_cols, 2)
  expect_equal(attrs$df_name, "`chickwts`")
  expect_equal(attrs$base_skimmers, c("n_missing", "complete_rate"))
  expect_equal(
    attrs$skimmers_used,
    list(
      numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist"),
      factor = c("ordered", "n_unique", "top_counts")
    )
  )
})

test_that("successfully skim mixed data types with common skimmers", {
  df <- data.frame(
    Date = seq(as.Date("2011-07-01"), by = 1, len = 10),
    POSIXct = as.POSIXct("2011-07-01 00:00:00", tz = "UTC")
  )
  input <- skim(df)
  expect_n_rows(input, 2)
  expect_n_columns(input, 12)

  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(
    input,
    c(
      "skim_type", "skim_variable", "n_missing", "complete_rate",
      "Date.min", "Date.max", "Date.median", "Date.n_unique",
      "POSIXct.min", "POSIXct.max", "POSIXct.median", "POSIXct.n_unique"
    )
  )
})

test_that("skim handles a matrix as input", {
  expect_error(skim(volcano), NA)
})

test_that("skim handles a vector as input", {
  expect_error(skim(iris$Sepal.Length), NA)
})

test_that("You can use tidyselect negation", {
  input <- skim(chickwts, -weight)

  # dimensions
  expect_n_rows(input, 1)
  expect_n_columns(input, 7)

  # values
  expect_identical(unname(input$skim_variable), "feed")
  expect_identical(input$skim_type, "factor")
})

test_that("Tidyselect helpers work as expected", {
  input <- skim(iris, starts_with("Sepal"))

  expect_n_rows(input, 2)
  expect_n_columns(input, 12)
  expect_identical(
    unname(input$skim_variable),
    c("Sepal.Length", "Sepal.Width")
  )
})

test_that("Tidyselect predicates work as expected", {
  input <- skim(iris, where(is.numeric))

  expect_n_rows(input, 4)
  expect_n_columns(input, 12)
  expect_identical(
    unname(input$skim_variable),
    c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  )
})

test_that("Skimming a grouped df with selections works as expected", {
  grouped <- dplyr::group_by(mtcars, cyl, gear)
  input <- skim(grouped, mpg, disp)

  # dimensions
  expect_n_rows(input, 16)
  expect_n_columns(input, 14)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "cyl", "gear", "n_missing",
    "complete_rate", "numeric.mean", "numeric.sd", "numeric.p0",
    "numeric.p25", "numeric.p50", "numeric.p75", "numeric.p100", "numeric.hist"
  ))

  # attributes
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 32)
  expect_equal(attrs$data_cols, 11)
  expect_equal(attrs$df_name, "`grouped`")
  expect_equal(as.character(attrs$groups), c("cyl", "gear"))
  expect_equal(
    attrs$skimmers_used,
    list(numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist"))
  )
})

test_that("Skimming a grouped df works when selecting exactly one variable", {
  grouped <- dplyr::group_by(mtcars, cyl, gear)
  input <- skim(grouped, mpg)

  # dimensions
  expect_n_rows(input, 8)
  expect_n_columns(input, 14)

  # classes
  expect_s3_class(input, "skim_df")
  expect_s3_class(input, "tbl_df")
  expect_s3_class(input, "tbl")
  expect_s3_class(input, "data.frame")
  expect_named(input, c(
    "skim_type", "skim_variable", "cyl", "gear", "n_missing",
    "complete_rate", "numeric.mean", "numeric.sd", "numeric.p0",
    "numeric.p25", "numeric.p50", "numeric.p75", "numeric.p100", "numeric.hist"
  ))
  expect_true(all(input$skim_variable == "mpg"))
})

test_that("Skimming without charts produces no histograms", {
  input <- skim_without_charts(iris, Sepal.Width)
  expect_null(iris$numeric.hist)
})

test_that("Skimming without charts produces no ts line charts", {
  input <- skim_without_charts(freeny, y)
  expect_null(iris$ts.line_graph)
})

test_that("Skimming succeeds when column names are similar", {
  input <- data.frame(
    x = 1:10,
    X = 11:20
  )
  skimmed <- skim(input)
  expect_s3_class(skimmed, "skim_df")
})


test_that("Skim inside a function returns the data name", {
  nested_skim <- function(df) {
    skim(df, .data_name = deparse(substitute(df)))
  }
  input <- nested_skim(iris)
  expect_equal(attr(input, "df_name"), "iris")
})

test_that("The data_name parameter can change the data_name attribute", {
  input <- skim(iris, .data_name = "Anderson")
  expect_equal(attr(input, "df_name"), "Anderson")
})

#' Functions for working with the vctrs package
#'
#' These make it clear that we need to use the tibble behavior when joining,
#' concatenating or casting `skim_df` objects. For a better discussion, on
#' why this is important and how these functions work, see:
#' <https://vctrs.r-lib.org/reference/howto-faq-coercion-data-frame.html>.
#'
#' `vec_ptype2.*` handles finding common prototypes between `skim_df` and
#' similar objects. `vec_cast.*` handles casting between objects. Note that
#' as of `dplyr 1.0.2`, [dplyr::bind_rows()] does not full support combining
#' attributes and [vctrs::vec_rbind()] is preferred instead.
#'
#' @importFrom vctrs vec_ptype2 vec_cast
#' @name skimr-vctrs
#' @keywords internal
NULL

#' @rdname skimr-vctrs
#' @export
vec_ptype2.skim_df.skim_df <- function(x, y, ...) {
  combine_compatible_objects(x, y, ...)
}

#' @rdname skimr-vctrs
#' @export
vec_ptype2.skim_df.tbl_df <- function(x, y, ...) {
  combine_compatible_objects(x, y, ...)
}

#' @rdname skimr-vctrs
#' @export
vec_ptype2.tbl_df.skim_df <- function(x, y, ...) {
  combine_compatible_objects(x, y, ...)
}

#' @rdname skimr-vctrs
#' @export
vec_cast.skim_df.skim_df <- function(x, to, ...) {
  cast_compatible_object(x, to, ...)
}

#' @rdname skimr-vctrs
#' @export
vec_cast.skim_df.tbl_df <- function(x, to, ...) {
  vctrs::tib_cast(x, to, ...)
}

#' @rdname skimr-vctrs
#' @export
vec_cast.tbl_df.skim_df <- function(x, to, ...) {
  vctrs::tib_cast(x, to, ...)
}

#' We only combine skim_df's that were built with the same set of skimmers.
#' @noRd
has_compatible_skimmers <- function(x, y) {
  has_identical_base(x, y) && has_identical_skimmers(x, y)
}

has_identical_base <- function(x, y) {
  base_x <- attr(x, "base_skimmers") %||% attr(y, "base_skimmers")
  base_y <- attr(y, "base_skimmers") %||% attr(x, "base_skimmers")
  identical(base_x, base_y)
}

has_identical_skimmers <- function(x, y) {
  skim_list_x <- attr(x, "skimmers_used") %||% attr(y, "skimmers_used")
  skim_list_y <- attr(y, "skimmers_used") %||% attr(x, "skimmers_used")
  x_names <- names(skim_list_x)
  y_names <- names(skim_list_y)
  all_names <- union(x_names, y_names)
  all(purrr::map_lgl(
    all_names,
    check_identical_skimmers,
    x_names,
    y_names,
    skim_list_x,
    skim_list_y
  ))
}

check_identical_skimmers <- function(name,
                                     x_names,
                                     y_names,
                                     skim_list_x,
                                     skim_list_y) {
  if ((name %in% x_names) && (name %in% y_names)) {
    identical(skim_list_x[name], skim_list_y[name])
  } else {
    TRUE
  }
}

combine_compatible_objects <- function(x, y, ..., x_arg = "", y_arg = "") {
  if (!has_compatible_skimmers(x, y)) {
    vctrs::stop_incompatible_type(
      x,
      y,
      x_arg = x_arg,
      y_arg = y_arg,
      details = "Can't combine different sets of skim functions."
    )
  }

  if (could_be_skim_df(x) && could_be_skim_df(y)) {
    reassign_skim_attrs(
      vctrs::tib_ptype2(x, y, ...),
      x,
      data_rows = data_rows(x) + data_rows(y),
      data_cols = data_cols(x) + data_cols(y),
      df_name = paste0(df_name(x), "+", df_name(y)),
      groups = c(group_names(x), group_names(y)),
      skimmers_used = union(skimmers_used(x), skimmers_used(y))
    )
  } else {
    strip_skim_attrs(vctrs::tib_ptype2(x, y, ...))
  }
}

cast_compatible_object <- function(x, to, ..., x_arg = "", to_arg = "") {
  out <- vctrs::tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
  if (could_be_skim_df(out)) {
    reassign_skim_attrs(out, to)
  } else {
    strip_skim_attrs(out)
  }
}

# Simpler expectations for data frames.

expect_n_columns <- function(object, n) {
  stopifnot(is.numeric(n), length(n) == 1)
  act <- testthat::quasi_label(rlang::enquo(object))
  act$n <- length(act$val)
  testthat::expect(act$n == n, sprintf(
    "%s has %i columns, not %i columns.",
    act$lab, act$n, n
  ))
  invisible(act$val)
}

expect_n_rows <- function(object, n) {
  stopifnot(is.numeric(n), length(n) == 1)
  act <- testthat::quasi_label(rlang::enquo(object))
  act$n <- nrow(act$val)
  testthat::expect(
    act$n == n,
    sprintf("%s has %i rows, not %i rows", act$lab, act$n, n)
  )
  invisible(act$val)
}

expect_NA <- function(object) {
  act <- testthat::quasi_label(rlang::enquo(object))
  testthat::expect(is.na(act$val), sprintf("%s is not NA", act$lab))
  invisible(act$val)
}

expect_print_matches_file <- function(object,
                                      filename,
                                      skip_on_windows = getOption(
                                        "skimr_skip_on_windows",
                                        TRUE
                                      ),
                                      skip_on_cran = getOption(
                                        "skimr_skip_on_cran",
                                        TRUE
                                      ),
                                      width = 100,
                                      update = getOption(
                                        "skimr_update_print",
                                        FALSE
                                      ),
                                      skimr_table_header_width = NULL,
                                      ...) {
  if (skip_on_windows) testthat::skip_on_os("windows")
  if (skip_on_windows) testthat::skip_if_not(l10n_info()$`UTF-8`)
  if (skip_on_cran) testthat::skip_on_cran()
  withr::with_options(list(
    crayon.enabled = FALSE,
    width = width,
    skimr_table_header_width = skimr_table_header_width
  ), {
    testthat::expect_known_output(
      print(object, ...),
      filename,
      update = update,
      width = width
    )
  })
}

expect_matches_file <- function(object,
                                file,
                                update = getOption(
                                  "skimr_update_print",
                                  FALSE
                                ),
                                skip_on_windows = getOption(
                                  "skimr_skip_on_windows",
                                  TRUE
                                ),
                                skip_on_cran = getOption(
                                  "skimr_skip_on_cran",
                                  TRUE
                                ),
                                width = 100,
                                ...) {
  if (skip_on_windows) testthat::skip_on_os("windows")
  if (skip_on_windows) testthat::skip_if_not(l10n_info()$`UTF-8`)
  if (skip_on_cran) testthat::skip_on_cran()
  withr::local_options(list(crayon.enabled = FALSE, width = width))
  act <- testthat::quasi_label(rlang::enquo(object), NULL)

  if (!file.exists(file)) {
    warning("Creating reference value", call. = FALSE)
    writeLines(object, file)
    testthat::succeed()
  } else {
    ref_val <- paste0(readLines(file), collapse = "\n")
    comp <- testthat::compare(as.character(act$val), ref_val, ...)
    if (update && !comp$equal) {
      writeLines(act$val, file)
    }

    testthat::expect(
      comp$equal,
      sprintf(
        "%s has changed from known value recorded in %s.\n%s",
        act$lab, encodeString(file, quote = "'"), comp$message
      ),
      info = NULL
    )
  }

  invisible(act$value)
}

test_that("skim of a simple data.table produces no warnings", {
  skip_if_not_installed("data.table")
  withr::local_options(list(width = 91))
  DT_letters <- data.table::data.table(abc = letters)
  expect_warning(skim(DT_letters), NA)
})

test_that("skim of a simple data.table produces no warnings even with dtplyr", {
  skip_if_not_installed("data.table")
  skip_if_not_installed("dtplyr")
  withr::local_options(list(width = 91))
  DT_letters <- data.table::data.table(abc = letters)
  expect_warning(skim(DT_letters), NA)
})

test_that("skim of a simple data.table produces output as expected", {
  skip_if_not_installed("data.table")
  DT_letters <- data.table::data.table(abc = letters)

  skimmed_DT_letters <- skim(DT_letters)
  withr::local_options(list(cli.unicode = FALSE, width = 91))
  expect_print_matches_file(
    skimmed_DT_letters,
    "data.table/summary_DT_letters.txt"
  )
})


test_that("skim of data.table produces output as expected", {
  skip_if_not_installed("data.table")
  set.seed(1L)

  DT_factors <- data.table::data.table(
    abc = letters,
    grps = factor(sample(c("AA", "BB"), 26, TRUE)),
    values = rnorm(26)
  )

  withr::local_options(list(cli.unicode = FALSE, width = 91))

  expect_print_matches_file(
    skim(DT_factors),
    "data.table/summary_DT_factors_no_key.txt"
  )

  data.table::setkeyv(DT_factors, c("abc", "grps"))
  expect_print_matches_file(
    skim(DT_factors),
    "data.table/summary_DT_factors.txt"
  )

  DF_factors <- as.data.frame(DT_factors)
  expect_print_matches_file(
    skim(DF_factors),
    "data.table/summary_DF_factors.txt"
  )

  tibble_factors <- tibble::as_tibble(DT_factors)
  expect_print_matches_file(
    skim(tibble_factors),
    "data.table/summary_tibble_factors.txt"
  )
})

skimmed_iris <- skim(iris)

test_that("dplyr::filter works as expected", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  input <- dplyr::filter(skimmed_iris, skim_type == "numeric")
  expect_print_matches_file(input, "dplyr/filter-skim.txt")
  no_rows <- dplyr::filter(skimmed_iris, skim_type == "no_type")
  expect_print_matches_file(no_rows, "dplyr/filter-no-skim.txt")
})

test_that("dplyr::select works as expected", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  with_type <- dplyr::select(skimmed_iris, skim_type, skim_variable)
  expect_print_matches_file(with_type, "dplyr/select-skim.txt")

  without_type <- dplyr::select(skimmed_iris, numeric.mean)
  withr::local_options(list(cli.unicode = FALSE))
  expect_print_matches_file(without_type, "dplyr/select-no-skim.txt")
})

test_that("dplyr::mutate works as expected", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  input <- dplyr::mutate(skimmed_iris, mean2 = numeric.mean^2)
  expect_print_matches_file(input, "dplyr/mutate-skim.txt")

  no_variable <- dplyr::mutate(skimmed_iris, skim_variable = NULL)
  identical(
    capture.output(print(no_variable)),
    capture.output(print(tibble::as_tibble(no_variable)))
  )
})

test_that("dplyr::slice works as expected", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  input <- dplyr::slice(skimmed_iris, 1:3)
  expect_print_matches_file(input, "dplyr/slice.txt")
})

test_that("dplyr::arrange works as expected", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  input <- dplyr::arrange(skimmed_iris, desc(numeric.mean))
  expect_print_matches_file(input, "dplyr/arrange.txt")
})

test_that("Using skim_tee prints returns the object", {
  skip_if_not(l10n_info()$`UTF-8`)
  input <- testthat::capture_output(skim_object <- skim_tee(chickwts))
  expect_matches_file(input, "skim_tee/skim_tee.txt")
  expect_identical(chickwts, skim_object)
})

test_that("skim_tee prints only selected columns, but returns full object", {
  skip_if_not(l10n_info()$`UTF-8`)
  input <- testthat::capture_output(obj <- skim_tee(iris, Species))
  expect_matches_file(input, "skim_tee/species.txt")
  expect_identical(obj, iris)
})

test_that("skim_tee supports dplyr helpers", {
  skip_if_not(l10n_info()$`UTF-8`)
  input <- testthat::capture_output(obj <- skim_tee(iris, starts_with("Sepal")))
  expect_matches_file(input, "skim_tee/sepal.txt")
  expect_identical(obj, iris)
})

test_that("Skim_tee works with groups", {
  skip_if_not(l10n_info()$`UTF-8`)
  iris_grouped <- dplyr::group_by(iris, Species)
  my_skim <- skim_with(numeric = sfl(hist = NULL))
  input <- testthat::capture_output(
    obj <- skim_tee(iris_grouped, Sepal.Length, skim_fun = my_skim)
  )
  expect_matches_file(input, "skim_tee/grouped.txt")
  expect_identical(obj, iris_grouped)
})

#' Skim a data frame, getting useful summary statistics
#'
#' `skim()` is an alternative to [summary()], quickly providing a broad
#' overview of a data frame. It handles data of all types, dispatching a
#' different set of summary functions based on the types of columns in the data
#' frame.
#'
#' Each call produces a `skim_df`, which is a fundamentally a tibble with a
#' special print method. One unusual feature of this data frame is pseudo-
#' namespace for columns. `skim()` computes statistics by data type, and it
#' stores them in the data frame as `<type>.<statistic>`. These types are
#' stripped when printing the results. The "base" skimmers (`n_missing` and
#' `complete_rate`) are the only columns that don't follow this behavior.
#' See [skim_with()] for more details on customizing `skim()` and
#' [get_default_skimmers()] for a list of default functions.
#'
#' If you just want to see the printed output, call `skim_tee()` instead.
#' This function returns the original data. `skim_tee()` uses the default
#' `skim()`, but you can replace it with the `skim` argument.
#'
#' The data frame produced by `skim` is wide and sparse. To avoid type coercion
#' `skimr` uses a type namespace for all summary statistics. Columns for numeric
#' summary statistics all begin `numeric`; for factor summary statistics
#' begin `factor`; and so on.
#'
#' See [partition()] and [yank()] for methods for transforming this wide data
#' frame. The first function splits it into a list, with each entry
#' corresponding to a data type. The latter pulls a single subtable for a
#' particular type from the `skim_df`.
#'
#' `skim()` is designed to operate in pipes and to generally play nicely with
#' other `tidyverse` functions. This means that you can use `tidyselect` helpers
#' within `skim` to select or drop specific columns for summary. You can also
#' further work with a `skim_df` using `dplyr` functions in a pipeline.
#'
#' @section Customizing skim:
#' `skim()` is an intentionally simple function, with minimal arguments like
#' [summary()]. Nonetheless, this package provides two broad approaches to
#' how you can customize `skim()`'s behavior. You can customize the functions
#' that are called to produce summary statistics with [skim_with()].
#'
#' @section Unicode rendering:
#' If the rendered examples show unencoded values such as `<U+2587>` you will
#' need to change your locale to allow proper rendering. Please review the
#' *Using Skimr* vignette for more information
#' (`vignette("Using_skimr", package = "skimr")`).
#'
#' Otherwise, we export `skim_without_charts()` to produce summaries without the
#' spark graphs. These are the source of the unicode dependency.
#'
#' @param data A tibble, or an object that can be coerced into a tibble.
#' @param ...  Columns to select for skimming. When none are provided, the
#'   default is to skim all columns.
#' @param .data_name The name to use for the data. Defaults to the same as data.
#' @param skim  The skimming function to use in `skim_tee()`.
#' @return A `skim_df` object, which also inherits the class(es) of the input
#'   data. In many ways, the object behaves like a [tibble::tibble()].
#' @examples
#' skim(iris)
#'
#' # Use tidyselect
#' skim(iris, Species)
#' skim(iris, starts_with("Sepal"))
#' skim(iris, where(is.numeric))
#'
#' # Skim also works groupwise
#' iris %>%
#'   dplyr::group_by(Species) %>%
#'   skim()
#'
#' # Which five numeric columns have the greatest mean value?
#' # Look in the `numeric.mean` column.
#' iris %>%
#'   skim() %>%
#'   dplyr::select(numeric.mean) %>%
#'   dplyr::top_n(5)
#'
#' # Which of my columns have missing values? Use the base skimmer n_missing.
#' iris %>%
#'   skim() %>%
#'   dplyr::filter(n_missing > 0)
#'
#' # Use skim_tee to view the skim results and
#' # continue using the original data.
#' chickwts %>%
#'   skim_tee() %>%
#'   dplyr::filter(feed == "sunflower")
#'
#' # Produce a summary without spark graphs
#' iris %>%
#'   skim_without_charts()
#' @export
skim <- skim_with()

#' @rdname skim
#' @param skim_fun The skim function used.
#' @export
skim_tee <- function(data, ..., skim_fun = skim) {
  skimmed <- skim_fun(data, ...)
  print(skimmed)
  invisible(data)
}

#' @rdname skim
#' @export
skim_without_charts <- skim_with(
  numeric = sfl(hist = NULL),
  ts = sfl(line_graph = NULL)
)

test_that("Using skim_tee prints returns the object", {
  skip_if_not(l10n_info()$`UTF-8`)
  expect_snapshot({
    skim_object <- skim_tee(chickwts)
  })
  expect_identical(chickwts, skim_object)
})

test_that("skim_tee prints only selected columns, but returns full object", {
  skip_if_not(l10n_info()$`UTF-8`)
  expect_snapshot({
    obj <- skim_tee(iris, Species)
  })
  expect_identical(obj, iris)
})

test_that("skim_tee supports dplyr helpers", {
  skip_if_not(l10n_info()$`UTF-8`)
  expect_snapshot({
    obj <- skim_tee(iris, starts_with("Sepal"))
  })
  expect_identical(obj, iris)
})

test_that("Skim_tee works with groups", {
  skip_if_not(l10n_info()$`UTF-8`)
  iris_grouped <- dplyr::group_by(iris, Species)
  my_skim <- skim_with(numeric = sfl(hist = NULL))
  expect_snapshot({
    obj <- skim_tee(iris_grouped, Sepal.Length, skim_fun = my_skim)
  })
  expect_identical(obj, iris_grouped)
})

#' Print `skim` objects
#'
#' `skimr` has custom print methods for all supported objects. Default printing
#' methods for `knitr`/ `rmarkdown` documents is also provided.
#'
#' @section Printing options:
#'
#' For better or for worse, `skimr` often produces more output than can fit in
#' the standard R console. Fortunately, most modern environments like RStudio
#' and Jupyter support more than 80 character outputs. Call
#' `options(width = 90)` to get a better experience with `skimr`.
#'
#' The print methods in `skimr` wrap those in the [tibble][tibble::formatting]
#' package. You can control printing behavior using the same global options.
#'
#' @section Behavior in `dplyr` pipelines:
#'
#' Printing a `skim_df` requires specific columns that might be dropped when
#' using [dplyr::select()] or [dplyr::summarize()] on a `skim_df`. In those
#' cases, this method falls back to [tibble::print.tbl()].
#'
#' @section Options for controlling print behavior:
#'
#' You can control the width rule line for the printed subtables with an option:
#' `skimr_table_header_width`.
#'
#' @inheritParams tibble::print.tbl
#' @seealso [tibble::trunc_mat()] For a list of global options for customizing
#'   print formatting. [cli::num_ansi_colors()] for the variety of issues that
#'   affect tibble's color support.
#' @param include_summary Whether a summary of the data frame should be printed
#' @param summary_rule_width Width of Data Summary cli rule, defaults to 40.
#' @name print
NULL

#' @describeIn print Print a skimmed data frame (`skim_df` from [skim()]).
#' @export
print.skim_df <- function(x,
                          include_summary = TRUE,
                          n = Inf,
                          width = Inf,
                          summary_rule_width = getOption(
                            "skimr_summary_rule_width",
                            default = 40
                          ),
                          ...) {
  if (is_skim_df(x) && nrow(x) > 0) {
    if (include_summary) {
      print(summary(x), .summary_rule_width = summary_rule_width, ...)
    }
    by_type <- partition(x)
    purrr::map(
      by_type,
      print,
      width = width,
      n = n,
      ...
    )
    invisible(x)
  } else {
    NextMethod("print")
  }
}

# Methods for correctly formatting a `one_skim_df`. We leverage the
# customiztion options in `pillar` for this. It divides the results into: a
# header, which we customize; a body, where we strip some values; and a footer,
# which we drop. For more details, see
# https://pillar.r-lib.org/articles/extending.html

#' @importFrom pillar tbl_format_header
#' @export
tbl_format_header.one_skim_df <- function(x, setup, ...) {
  variable_type <- paste("Variable type:", attr(x, "skim_type"))
  rule <- cli::rule(
    line = 1,
    left = variable_type,
    width = getOption("skimr_table_header_width", getOption("width", 80))
  )
  # Add an empty line before the rule
  c("", rule)
}

#' @importFrom pillar ctl_new_pillar
#' @export
ctl_new_pillar.one_skim_df <- function(controller,
                                       x,
                                       width,
                                       ...,
                                       title = NULL) {
  out <- NextMethod()
  out$type <- NULL
  out
}

#' @describeIn print Print a `skim_list`, a list of `skim_df` objects.
#' @export
print.skim_list <- function(x,
                            n = Inf,
                            width = Inf,
                            ...) {
  nms <- names(x)
  attributes(x) <- NULL
  print(rlang::set_names(x, nms))
}


#' @describeIn print Print method for a `summary_skim_df` object.
#' @param .summary_rule_width the width for the main rule above the summary.
#' @param  ...  Passed on to \link[pillar:tbl_format_setup]{pillar::tbl_format_setup()}
#' @export
print.summary_skim_df <- function(x, .summary_rule_width = 40, ...) {
  with_title <- c(
    cli::rule(line = 1, left = "Data Summary", width = .summary_rule_width),
    format(x)
  )
  writeLines(with_title)
}

#' Provide a default printing method for knitr.
#'
#' Instead of standard R output, `knitr` and `RMarkdown` documents will have
#' formatted [knitr::kable()] output on return. You can disable this by setting
#' the chunk option `render = normal_print`.
#'
#' The summary statistics for the original data frame can be disabled by setting
#' the `knitr` chunk option `skimr_include_summary = FALSE`. See
#' [knitr::opts_chunk] for more information. You can change the number of digits
#' shown in the printed table with the `skimr_digits` chunk option.
#'
#' Alternatively, you can call [yank()] to get the particular
#' `skim_df` objects and format them however you like. One warning though.
#' Because histograms contain unicode characters, they can have unexpected
#' print results, as R as varying levels of unicode support. This affects
#' Windows users most commonly. Call `vignette("Using_fonts")` for more details.
#'
#' @seealso [knitr::kable()]
#' @inheritParams knitr::knit_print
#' @param options Options passed into the print function.
#' @return A `knit_asis` object. Which is used by `knitr` when rendered.
#' @importFrom knitr knit_print
#' @name knit_print
NULL

#' @describeIn knit_print Default `knitr` print for `skim_df` objects.
#' @export
knit_print.skim_df <- function(x, options = NULL, ...) {
  if (is_skim_df(x) && nrow(x) > 0) {
    if (options$skimr_include_summary %||% TRUE) {
      summary_stats <- summary(x)
      kabled <- knit_print(summary_stats)
    } else {
      kabled <- c()
    }

    by_type <- partition(x)
    knit_print_by_type(by_type, options, kabled)
  } else {
    NextMethod("knit_print")
  }
}

knit_print_by_type <- function(x, options, summary) {
  all_tables <- purrr::imap(x, knit_print_one, options)
  combined <- c("", summary, "", "", unlist(all_tables))
  knitr::asis_output(paste(combined, collapse = "\n"))
}

knit_print_one <- function(by_type, type, options) {
  kabled <- knitr::kable(
    by_type,
    digits = options$skimr_digits %||% 2
  )
  caption <- sprintf("**Variable type: %s**", type)
  c(caption, "", kabled, "", "")
}

#' @describeIn knit_print Default `knitr` print for a `skim_list`.
#' @export
knit_print.skim_list <- function(x, options = NULL, ...) {
  knit_print_by_type(x, options, NULL)
}

#' @describeIn knit_print Default `knitr` print within a partitioned `skim_df`.
#' @export
knit_print.one_skim_df <- function(x, options = NULL, ...) {
  kabled <- knit_print_one(x, attr(x, "skim_type"), options)
  combined <- c("", "", kabled, "")
  knitr::asis_output(paste(combined, collapse = "\n"))
}

#' @describeIn knit_print Default `knitr` print for `skim_df` summaries.
#' @export
knit_print.summary_skim_df <- function(x, options = NULL, ...) {
  summary_mat <- cbind(
    get_summary_dnames(x),
    get_summary_values(x),
    deparse.level = 0
  )
  kabled <- knitr::kable(
    summary_mat,
    table.attr = "style='width: auto;'
      class='table table-condensed'",
    caption = "Data summary"
  )

  knitr::asis_output(paste(kabled, collapse = "\n"))
}


#' Skimr printing within Jupyter notebooks
#'
#' This reproduces printed results in the console. By default Jupyter kernels
#' render the final object in the cell. We want the version printed by
#' `skimr` instead of the data that it contains.
#'
#' @param obj The object to \link{print} and then return the output.
#' @param ... ignored.
#' @return None. `invisible(NULL)`.
#' @importFrom repr repr_text
#' @name repr

#' @rdname repr
#' @export
repr_text.skim_df <- function(obj, ...) {
  paste(utils::capture.output(print(obj)), collapse = "\n")
}

#' @rdname repr
#' @export
repr_text.skim_list <- repr_text.skim_df

#' @rdname repr
#' @export
repr_text.one_skim_df <- repr_text.skim_df

# Extracted from test-skim_print.R:15

# setup ------------------------------------------------------------------------
library(testthat)
test_env <- simulate_test_env(package = "skimr", path = "..")
attach(test_env, warn.conflicts = FALSE)

# test -------------------------------------------------------------------------
skip_if_not(l10n_info()$`UTF-8`)
withr::local_options(list(cli.unicode = FALSE))
expect_snapshot( skim(dplyr::group_by(iris, Species)))

#' Create a skimr function list
#'
#' This constructor is used to create a named list of functions. It also you
#' also pass `NULL` to identify a skimming function that you wish to remove.
#' Only functions that return a single value, working with [dplyr::summarize()],
#' can be used within `sfl`.
#'
#' `sfl()` will automatically generate callables and names for a variety of
#' inputs, including functions, formulas and strings. Nonetheless, we recommend
#' providing names when reasonable to get better [skim()] output.
#'
#' @param ... Inherited from dplyr::data_masking() for dplyr version 1 or later
#'  or dplyr::funs() for older versions of dplyr.
#'  A list of functions
#'   specified by:
#'
#'  * Their name, `"mean"`
#'  * The function itself, `mean`
#'  * A call to the function with `.` as a dummy argument,
#'    `mean(., na.rm = TRUE)`
#'  * An anonymous function in \pkg{purrr} notation, `~mean(., na.rm = TRUE)`
#'
#' @param skim_type A character scalar. This is used to match locally-provided
#'   skimmers with defaults. See [get_skimmers()] for more detail.
#' @return A `skimr_function_list`, which contains a list of `fun_calls`,
#'   returned by `dplyr::funs()` and a list of skimming functions to drop.
#' @seealso [dplyr::funs()], [skim_with()] and [get_skimmers()].
#' @examples
#' # sfl's can take a variety of input formats and will generate names
#' # if not provided.
#' sfl(mad, "var", ~ length(.)^2)
#'
#' # But these can generate unpredictable names in your output.
#' # Better to set your own names.
#' sfl(mad = mad, variance = "var", length_sq = ~ length(.)^2)
#'
#' # sfl's can remove individual skimmers from defaults by passing NULL.
#' sfl(hist = NULL)
#'
#' # When working interactively, you don't need to set a type.
#' # But you should when defining new defaults with `get_skimmers()`.
#' get_skimmers.my_new_class <- function(column) {
#'   sfl(n_missing, skim_type = "my_new_class")
#' }
#' @export
sfl <- function(..., skim_type = "") {
  stopifnot(length(skim_type) == 1, is.character(skim_type))
  funs <- build_sfl_names(...)
  structure(
    list(funs = funs, skim_type = skim_type),
    class = "skimr_function_list"
  )
}

build_sfl_names <- function(...) {
  labels <- rlang::quos_auto_name(rlang::enquos(...))
  rlang::set_names(rlang::list2(...), names(labels))
}

test_that("Only string scalars for types", {
  expect_error(sfl(mad, skim_type = c("1", "2")))
})

test_that("Zero-length sfl's supported", {
  input <- sfl(skim_type = "numeric")
  expect_length(input$funs, 0)
})

test_that("The interface for sfl's separates keep and drop functions", {
  input <- sfl(mad = mad, hist = NULL, skim_type = "test")
  expect_s3_class(input, "skimr_function_list")
  expect_length(input, 2)
  expect_named(input, c("funs", "skim_type"))
  expect_identical(input$skim_type, "test")

  funs <- input$funs
  expect_type(funs, "list")
  expect_named(funs, c("mad", "hist"))
})

test_that("sfl's support dummy names", {
  input <- sfl(mean = ~ mean(., na.rm = TRUE), skim_type = "test")
  funs <- input$funs
  expect_equal(funs$mean, rlang::quo(mean(., na.rm = TRUE)), ignore_attr = "class")
})

test_that("sfl's automatically generate function names", {
  input <- sfl(mad, hist = NULL, ~ length(.)^2, "var")
  expect_named(input$funs, c("mad", "hist", "~length(.)^2", '"var"'))
})

test_that("Only string scalars for types", {
  expect_error(sfl(mad, skim_type = c("1", "2")))
})

test_that("Zero-length sfl's supported", {
  input <- sfl(skim_type = "numeric")
  expect_length(input$funs, 0)
})

test_that("The interface for sfl's separates keep and drop functions", {
  input <- sfl(mad = mad, hist = NULL, skim_type = "test")
  expect_s3_class(input, "skimr_function_list")
  expect_length(input, 2)
  expect_named(input, c("funs", "skim_type"))
  expect_identical(input$skim_type, "test")

  funs <- input$funs
  expect_type(funs, "list")
  expect_named(funs, c("mad", "hist"))
})

test_that("sfl's support dummy names", {
  input <- sfl(mean = ~ mean(., na.rm = TRUE), skim_type = "test")
  funs <- input$funs
  expect_equal(funs$mean, rlang::quo(mean(., na.rm = TRUE)))
})

test_that("sfl's automatically generate function names", {
  input <- sfl(mad, hist = NULL, ~ length(.)^2, "var")
  expect_named(input$funs, c("mad", "hist", "~length(.)^2", '"var"'))
})

test_that("Skimmer list is updated correctly when changing functions", {
  funs <- sfl(median = median, mad = mad)
  new_skim <- skim_with(numeric = funs, append = FALSE)
  x <- tibble::tibble(rnorm(10))
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(used, list(numeric = c("median", "mad")))
})

test_that("Skimming functions can be changed for multiple types", {
  newfuns1 <- sfl(iqr = IQR, q99 = ~ quantile(., probs = .99))
  newfuns2 <- sfl(n2 = length)
  new_skim <- skim_with(numeric = newfuns1, factor = newfuns2, append = FALSE)
  input <- new_skim(iris)
  used <- attr(input, "skimmers_used")
  expect_identical(used, list(numeric = c("iqr", "q99"), factor = "n2"))
})

test_that("Skimming functions can be changed with a list", {
  newfuns1 <- sfl(iqr = IQR, q99 = ~ quantile(., probs = .99))
  new_skim <- skim_with(list(numeric = newfuns1), append = FALSE)
  input <- new_skim(iris)
  used <- attr(input, "skimmers_used")
  expect_identical(used$numeric, c("iqr", "q99"))
})

test_that("Skimming functions can be changed with a list", {
  newfuns1 <- sfl(iqr = IQR, q99 = ~ quantile(., probs = .99))
  newfuns2 <- sfl(n2 = length)
  new_skim <- skim_with(
    list(numeric = newfuns1, factor = newfuns2),
    append = FALSE
  )
  input <- new_skim(iris)
  used <- attr(input, "skimmers_used")
  expect_identical(
    used, list(numeric = c("iqr", "q99"), factor = c("n2"))
  )
})

test_that("Skimming functions can be appended.", {
  funs <- sfl(iqr = IQR)
  new_skim <- skim_with(numeric = funs)
  x <- tibble::tibble(rnorm(10))
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(
    used,
    list(
      numeric = c(
        "mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist", "iqr"
      )
    )
  )
})

test_that("Setting a statistic to null removes it from skimmers", {
  new_skim <- skim_with(numeric = sfl(hist = NULL))
  x <- tibble::tibble(rnorm(10))
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(
    used, list(numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100"))
  )
})

test_that("Skimmers can be removed and added at the same time", {
  new_skim <- skim_with(numeric = sfl(hist = NULL, iqr = IQR))
  x <- tibble::tibble(rnorm(10))
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(
    used,
    list(numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "iqr"))
  )
})

test_that("Skimming functions for new types can be added", {
  funs <- sfl(iqr = IQR, quantile = ~ quantile(., probs = .99))
  expect_message(new_skim <- skim_with(new_type = funs), "new_type")
  x <- tibble::tibble(x = rnorm(10))
  class(x$x) <- "new_type"
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(used, list(new_type = c("iqr", "quantile")))
})

test_that("Set multiple sets of skimming functions", {
  funs <- sfl(iqr = IQR, quantile = ~ quantile(., probs = .99))
  expect_message(
    new_skim <- skim_with(numeric = funs, new_type = funs),
    "new_type"
  )
  x <- tibble::tibble(x = rnorm(10), y = rnorm(10))
  class(x$x) <- "new_type"
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_named(used, c("new_type", "numeric"))
  expect_identical(used$new_type, c("iqr", "quantile"))
  expect_identical(
    used$numeric,
    c(
      "mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist", "iqr", "quantile"
    )
  )
})


test_that("Set multiple sets of skimming functions, rlang", {
  funs <- sfl(iqr = IQR, quantile = ~ quantile(., probs = .99))
  expect_message(new_skim <- skim_with(!!!list(numeric = funs, new_type = funs),
    append = FALSE
  ))
  x <- tibble::tibble(x = rnorm(10), y = rnorm(10))
  class(x$x) <- "new_type"
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_named(used, c("new_type", "numeric"))
  expect_identical(used$new_type, c("iqr", "quantile"))
  expect_identical(used$numeric, c("iqr", "quantile"))
})

test_that("Skimming functions without a class return a message.", {
  funs_no_class <- sfl(IQR)
  expect_error(skim_with(funs_no_class), "arguments to be named.")
  expect_error(
    skim_with(funs_no_class, numeric = funs_no_class),
    "arguments to be named."
  )
})

test_that("An empty call to skim_with() returns the default skim()", {
  input <- skim_with()
  expect_identical(input(iris), skim(iris))
})

test_that("User-defined defaults require sfl's with class names", {
  local_mocked_bindings(get_skimmers = function(column) sfl(length))
  expect_error(skim(data.frame(1)), "Default skimming functions")
})

test_that("Sfl's can be passed as an unquoted list", {
  my_skimmers <- list(numeric = sfl(mean), factor = sfl(length))
  my_skim <- skim_with(!!!my_skimmers, append = FALSE)
  input <- my_skim(iris)
  expect_named(
    input,
    c(
      "skim_type", "skim_variable", "n_missing", "complete_rate",
      "factor.length", "numeric.mean"
    )
  )
})

test_that("Doubles and integers are both 'numeric'", {
  df <- data.frame(int = 1:3, dbl = 1:3 + 0.5)
  my_skim <- skim_with(numeric = sfl(hist = NULL))
  input <- my_skim(df)

  expect_false("numeric.hist" %in% names(input))
  expect_equal(
    attr(input, "skimmers_used")$numeric,
    c("mean", "sd", "p0", "p25", "p50", "p75", "p100")
  )
})

test_that("Defining an integer sfl changes behavior", {
  df <- data.frame(int = 1:3, dbl = 1:3 + 0.5)
  expect_message(
    my_skim <- skim_with(
      numeric = sfl(hist = NULL), integer = sfl(int_mean = mean)
    )
  )
  input <- my_skim(df)

  expect_false("numeric.hist" %in% names(input))
  expect_true("integer.int_mean" %in% names(input))
  expect_identical(
    attr(input, "skimmers_used"),
    list(
      integer = c("int_mean"),
      numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100")
    )
  )
})

test_that("Base skimmers can be changed", {
  my_skim <- skim_with(base = sfl(length = length))
  skimmed <- my_skim(iris)
  expect_true("length" %in% names(skimmed))
  expect_equal(attr(skimmed, "base_skimmers"), "length")
})

test_that("Base skimmers require an sfl", {
  expect_error(skim_with(base = list(length = length)))
})

test_that("Base skimmers can be removed", {
  my_skim <- skim_with(base = NULL)
  skimmed <- my_skim(iris)
  used <- attr(skimmed, "skimmers_used")
  expect_false("base" %in% names(used))
})

test_that("Skimmer list is updated correctly when changing functions", {
  funs <- sfl(median = median, mad = mad)
  new_skim <- skim_with(numeric = funs, append = FALSE)
  x <- tibble::tibble(rnorm(10))
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(used, list(numeric = c("median", "mad")))
})

test_that("Skimming functions can be changed for multiple types", {
  newfuns1 <- sfl(iqr = IQR, q99 = ~ quantile(., probs = .99))
  newfuns2 <- sfl(n2 = length)
  new_skim <- skim_with(numeric = newfuns1, factor = newfuns2, append = FALSE)
  input <- new_skim(iris)
  used <- attr(input, "skimmers_used")
  expect_identical(used, list(numeric = c("iqr", "q99"), factor = "n2"))
})

test_that("Skimming functions can be changed with a list", {
  newfuns1 <- sfl(iqr = IQR, q99 = ~ quantile(., probs = .99))
  new_skim <- skim_with(list(numeric = newfuns1), append = FALSE)
  input <- new_skim(iris)
  used <- attr(input, "skimmers_used")
  expect_identical(used$numeric, c("iqr", "q99"))
})

test_that("Skimming functions can be changed with a list", {
  newfuns1 <- sfl(iqr = IQR, q99 = ~ quantile(., probs = .99))
  newfuns2 <- sfl(n2 = length)
  new_skim <- skim_with(
    list(numeric = newfuns1, factor = newfuns2),
    append = FALSE
  )
  input <- new_skim(iris)
  used <- attr(input, "skimmers_used")
  expect_identical(
    used, list(numeric = c("iqr", "q99"), factor = c("n2"))
  )
})

test_that("Skimming functions can be appended.", {
  funs <- sfl(iqr = IQR)
  new_skim <- skim_with(numeric = funs)
  x <- tibble::tibble(rnorm(10))
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(
    used,
    list(
      numeric = c(
        "mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist", "iqr"
      )
    )
  )
})

test_that("Setting a statistic to null removes it from skimmers", {
  new_skim <- skim_with(numeric = sfl(hist = NULL))
  x <- tibble::tibble(rnorm(10))
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(
    used, list(numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100"))
  )
})

test_that("Skimmers can be removed and added at the same time", {
  new_skim <- skim_with(numeric = sfl(hist = NULL, iqr = IQR))
  x <- tibble::tibble(rnorm(10))
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(
    used,
    list(numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "iqr"))
  )
})

test_that("Skimming functions for new types can be added", {
  funs <- sfl(iqr = IQR, quantile = ~ quantile(., probs = .99))
  expect_message(new_skim <- skim_with(new_type = funs), "new_type")
  x <- tibble::tibble(x = rnorm(10))
  class(x$x) <- "new_type"
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_identical(used, list(new_type = c("iqr", "quantile")))
})

test_that("Set multiple sets of skimming functions", {
  funs <- sfl(iqr = IQR, quantile = ~ quantile(., probs = .99))
  expect_message(
    new_skim <- skim_with(numeric = funs, new_type = funs),
    "new_type"
  )
  x <- tibble::tibble(x = rnorm(10), y = rnorm(10))
  class(x$x) <- "new_type"
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_named(used, c("new_type", "numeric"))
  expect_identical(used$new_type, c("iqr", "quantile"))
  expect_identical(
    used$numeric,
    c(
      "mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist", "iqr", "quantile"
    )
  )
})


test_that("Set multiple sets of skimming functions, rlang", {
  funs <- sfl(iqr = IQR, quantile = ~ quantile(., probs = .99))
  expect_message(new_skim <- skim_with(!!!list(numeric = funs, new_type = funs),
    append = FALSE
  ))
  x <- tibble::tibble(x = rnorm(10), y = rnorm(10))
  class(x$x) <- "new_type"
  input <- new_skim(x)
  used <- attr(input, "skimmers_used")
  expect_named(used, c("new_type", "numeric"))
  expect_identical(used$new_type, c("iqr", "quantile"))
  expect_identical(used$numeric, c("iqr", "quantile"))
})

test_that("Skimming functions without a class return a message.", {
  funs_no_class <- sfl(IQR)
  expect_error(skim_with(funs_no_class), "arguments to be named.")
  expect_error(
    skim_with(funs_no_class, numeric = funs_no_class),
    "arguments to be named."
  )
})

test_that("An empty call to skim_with() returns the default skim()", {
  input <- skim_with()
  expect_identical(input(iris), skim(iris))
})

test_that("User-defined defaults require sfl's with class names", {
  with_mock(
    get_skimmers = function(column) sfl(length),
    expect_error(skim(data.frame(1)), "Default skimming functions")
  )
})

test_that("Sfl's can be passed as an unquoted list", {
  my_skimmers <- list(numeric = sfl(mean), factor = sfl(length))
  my_skim <- skim_with(!!!my_skimmers, append = FALSE)
  input <- my_skim(iris)
  expect_named(
    input,
    c(
      "skim_type", "skim_variable", "n_missing", "complete_rate",
      "factor.length", "numeric.mean"
    )
  )
})

test_that("Doubles and integers are both 'numeric'", {
  df <- data.frame(int = 1:3, dbl = 1:3 + 0.5)
  my_skim <- skim_with(numeric = sfl(hist = NULL))
  input <- my_skim(df)
  
  expect_false("numeric.hist" %in% names(input))
  expect_equal(
    attr(input, "skimmers_used")$numeric,
    c("mean", "sd", "p0", "p25", "p50", "p75", "p100")
  )
})

test_that("Defining an integer sfl changes behavior", {
  df <- data.frame(int = 1:3, dbl = 1:3 + 0.5)
  my_skim <- expect_message(
    skim_with(
      numeric = sfl(hist = NULL), integer = sfl(int_mean = mean)
    )
  )
  input <- my_skim(df)

  expect_false("numeric.hist" %in% names(input))
  expect_true("integer.int_mean" %in% names(input))
  expect_identical(
    attr(input, "skimmers_used"),
    list(
      integer = c("int_mean"),
      numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100")
    )
  )
})

test_that("Base skimmers can be changed", {
  my_skim <- skim_with(base = sfl(length = length))
  skimmed <- my_skim(iris)
  expect_true("length" %in% names(skimmed))
  expect_equal(attr(skimmed, "base_skimmers"), "length")
})

test_that("Base skimmers require an sfl", {
  expect_error(skim_with(base = list(length = length)))
})

test_that("Base skimmers can be removed", {
  my_skim <- skim_with(base = NULL)
  skimmed <- my_skim(iris)
  used <- attr(skimmed, "skimmers_used")
  expect_false("base" %in% names(used))
})

#' Mutate a skim_df
#'
#' [dplyr::mutate()] currently drops attributes, but we need to keep them around
#' for other skim behaviors. Otherwise the behavior is exactly the same. For
#' more information, see <https://github.com/tidyverse/dplyr/issues/3429>.
#'
#' @param .data A `skim_df`, which behaves like a `tbl.`
#' @param ... Name-value pairs of expressions, each with length 1 or the same
#'   length as the number of rows in the group, if using [dplyr::group_by()], or
#'   in the entire input (if not using groups). The name of each argument will
#'   be the name of a new variable, and the value will be its corresponding
#'   value. Use `NULL` value in [dplyr::mutate()] to drop a variable. New
#'   variables overwrite existing variables of the same name.
#'
#'   The arguments in `...` are automatically quoted with [rlang::quo()] and
#'   evaluated with [rlang::eval_tidy()] in the context of the data frame. They
#'   support unquoting [rlang::quasiquotation] and splicing. See
#'   `vignette("programming", package = "dplyr")` for an introduction to these
#'   concepts.
#' @return A `skim_df` object, which also inherits the class(es) of the input
#'   data. In many ways, the object behaves like a [tibble::tibble()].
#' @seealso [dplyr::mutate()] for the function's expected behavior.
#' @importFrom dplyr mutate
#' @export
mutate.skim_df <- function(.data, ...) {
  mutated <- NextMethod("mutate")
  if (could_be_skim_df(mutated)) {
    reassign_skim_attrs(mutated, .data)
  } else {
    strip_skim_attrs(mutated)
  }
}

#' Summary function for skim_df
#'
#' This is a method of the generic function [summary()].
#'
#' @param object a skim dataframe.
#' @param ... Additional arguments affecting the summary produced. Not used.
#' @return A summary of the skim data frame.
#' @examples
#' a <- skim(mtcars)
#' summary(a)
#' @export
summary.skim_df <- function(object, ...) {
  if (is.null(object)) {
    stop("dataframe is null.")
  }
  duplicated <- duplicated(object$skim_variable)
  counts <- table(type = object$skim_type[!duplicated])
  types <- dimnames(counts)[[1]]

  possible_names <- group_names(object)
  possible_groups <- if (length(possible_names) > 0) {
    paste(possible_names, collapse = ", ")
  } else {
    "None"
  }

  structure(
    list(
      data_name = process_data_name(object),
      counts = counts,
      types = types,
      possible_groups = possible_groups,
      dt_key = dt_key(object),
      data_rows = data_rows(object),
      data_cols = data_cols(object)
    ),
    class = "summary_skim_df"
  )
}

process_data_name <- function(object) {
  raw_name <- df_name(object)
  no_ticks <- gsub("`", "", raw_name)
  if (no_ticks %in% c(".", ".data")) {
    "Piped data"
  } else if (nchar(no_ticks) > 25) {
    paste0(substring(no_ticks, 1, 25), "...")
  } else {
    no_ticks
  }
}

#' @export
format.summary_skim_df <- function(x, ...) {
  dnames <- c("", get_summary_dnames(x))
  summary_values <- c("Values", get_summary_values(x))
  paste(
    format(dnames),
    format(summary_values)
  )
}

get_summary_dnames <- function(summary_object) {
  c(
    "Name",
    "Number of rows ",
    "Number of columns ",
    if (!is.na(summary_object$dt_key)) "Key",
    "_______________________ ",
    "Column type frequency: ",
    paste0("  ", summary_object$types),
    "________________________  ",
    "Group variables"
  )
}

get_summary_values <- function(summary_object) {
  c(
    summary_object$data_name,
    summary_object$data_rows,
    summary_object$data_cols,
    if (!is.na(summary_object$dt_key)) summary_object$dt_key,
    " ",
    " ",
    unname(summary_object$counts),
    " ",
    summary_object$possible_groups
  )
}

test_that("Summary creates the correct summary object", {
  withr::local_options(list(cli.unicode = FALSE))
  # Test it creates the correct 4 parts
  skim_input <- skim(iris)
  summary_input <- summary(skim_input)
  expect_named(
    summary_input,
    c(
      "data_name", "counts", "types", "possible_groups",
      "dt_key", "data_rows", "data_cols"
    )
  )
  expect_identical(summary_input$data_name, "iris")
  expect_identical(summary_input$types, c("factor", "numeric"))
  expect_identical(summary_input$data_rows, 150L)
  expect_identical(summary_input$data_cols, 5L)
})

test_that("The summary print method prints the correct object", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skim_summary_input <- summary(skim(iris))
  expect_known_output(
    skim_summary_input, "summary/summary_iris.txt",
    update = FALSE,
    print = TRUE
  )
})

test_that("The summary print method prints the correct object when piped", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  # Test that the correct lines are output, no name should be output.
  summary_input <- iris %>%
    skim() %>%
    summary()
  expect_known_output(summary_input, "summary/summary_iris_piped.txt",
    update = FALSE, print = TRUE
  )
})

test_that("null object gets expected message", {
  input <- NULL
  expect_error(summary.skim_df(input), "dataframe is null.")
})

#' Separate a big `skim_df` into smaller data frames, by type.
#'
#' The data frames produced by [skim()] are wide and sparse, filled with
#' columns that are mostly `NA`. For that reason, it can be convenient to
#' work with "by type" subsets of the original data frame. These smaller
#' subsets have their `NA` columns removed.
#'
#' `partition()` creates a list of smaller `skim_df` data frames. Each entry
#' in the list is a data type from the original `skim_df`. The inverse of
#' `partition()` is `bind()`, which takes the list and produces the original
#' `skim_df`. While `partition()` keeps all of the subtables as list entries,
#' `yank()` gives you a single subtable for a data type.
#'
#' @param data A `skim_df`.
#' @param skim_type A character scalar. The subtable to extract from a
#'   `skim_df`.
#' @return A `skim_list` of `skim_df`'s, by type.
#' @examples
#' # Create a wide skimmed data frame (a skim_df)
#' skimmed <- skim(iris)
#'
#' # Separate into a list of subtables by type
#' separate <- partition(skimmed)
#'
#' # Put back together
#' identical(bind(separate), skimmed)
#' # > TRUE
#'
#' # Alternatively, get the subtable of a particular type
#' yank(skimmed, "factor")
#' @export
partition <- function(data) {
  assert_is_skim_df(data)
  data_as_list <- split(data, data$skim_type)
  groups <- group_names(data)
  base <- base_skimmers(data)

  skimmers <- reconcile_skimmers(data, groups, base)

  # Check to see that there are at least one, known used skim function
  # within the data. This can be reduced after using focus(). This indexing
  # is used so that the lookup order matches that of `data_as_list`.
  has_skim_data <- lengths(skimmers)[names(data_as_list)] > 0
  elements_to_keep <- has_skim_data | any(base %in% names(data))
  reduced <- purrr::imap(
    data_as_list[elements_to_keep],
    simplify_skimdf,
    skimmers = skimmers,
    groups = groups,
    base = base
  )

  reassign_skim_attrs(
    reduced,
    data,
    class = c("skim_list", "list"),
    skimmers_used = skimmers
  )
}

#' Align the skimmers_used attribute with the current columns in the data
#'
#' This catches the case where users partition (or more likely print) a data
#' frame that has had columns added after skimming.
#' @noRd
reconcile_skimmers <- function(data, groups, base) {
  all_columns <- colnames(data)
  skimmers_used <- skimmers_used(data)
  if (length(skimmers_used) == 0) {
    return(skimmers_used)
  }
  skimmers_from_names <- skimmers_from_names(all_columns, skimmers_used)
  with_base_columns <- c(
    "skim_variable",
    "skim_type",
    groups,
    base,
    collapse_skimmers(skimmers_used)
  )
  matched_cols <- dplyr::intersect(all_columns, with_base_columns)
  extra_cols <- dplyr::setdiff(all_columns, with_base_columns)
  if (length(extra_cols) > 0) {
    grouped <- dplyr::group_by(data, .data$skim_type)
    complete_by_type <- dplyr::summarise(
      grouped,
      dplyr::across(tidyselect::all_of(extra_cols),  ~ !all(is.na(.x)))
    )
    complete_cols <- purrr::pmap(
      complete_by_type,
      get_complete_columns,
      names = extra_cols
    )
    new_cols_by_type <- rlang::set_names(
      complete_cols,
      complete_by_type$skim_type
    )
    skimmers_from_names <- purrr::list_merge(
      skimmers_from_names,
      !!!new_cols_by_type
    )
  }

  skimmers_from_names
}

skimmers_from_names <- function(names, skimmers) {
  matched <- purrr::imap(skimmers, match_skimmers, names)
  purrr::set_names(matched, names(skimmers))
}

match_skimmers <- function(fun_names, type, values_to_match) {
  stripped_values <- stringr::str_remove(
    values_to_match,
    paste0("^", type, "\\.")
  )
  dplyr::intersect(stripped_values, fun_names)
}

collapse_skimmers <- function(skimmers_used) {
  with_type <- purrr::imap(skimmers_used, ~ paste(.y, .x, sep = "."))
  purrr::flatten_chr(with_type)
}

get_complete_columns <- function(skim_type, ..., names) {
  names[c(...)]
}

#' For each type subtable, only select columns generated by this type's skimmers
#' This function also catches the case where the user removed columns from
#' the skim_df
#' @noRd
simplify_skimdf <- function(data, skim_type, skimmers, groups, base) {
  stopifnot(has_variable_column(data))
  names(data) <- stringr::str_remove(names(data), paste0(skim_type, "\\."))
  keep <- c("skim_variable", groups, base, skimmers[[skim_type]])
  out <- dplyr::select(data, tidyselect::any_of(keep))

  structure(
    out,
    class = c("one_skim_df", "tbl_df", "tbl", "data.frame"),
    skim_type = skim_type
  )
}

#' @describeIn partition The inverse of a `partition()`. Rebuild the original
#'   `skim_df`.
#' @export
bind <- function(data) {
  assert_is_skim_list(data)
  with_namespaces <- purrr::imap(data, add_namespaces)
  combined <- dplyr::bind_rows(!!!with_namespaces, .id = "skim_type")
  reassign_skim_attrs(combined, data)
}

add_namespaces <- function(data, skim_type) {
  base <- base_skimmers(data)
  meta_columns <- c("skim_variable", dplyr::group_vars(data), base)
  no_meta_columns <- dplyr::setdiff(names(data), meta_columns)
  with_namespace <- paste(skim_type, no_meta_columns, sep = ".")

  # TODO(michaelquinn32): Drop this when vctrs interface works correctly.
  names(data) <- c(meta_columns, with_namespace)
  attr(data, "skim_type") <- NULL
  tibble::as_tibble(data)
}

#' @describeIn partition Extract a subtable from a `skim_df` with a particular
#'   type.
#' @export
yank <- function(data, skim_type) {
  partition(data)[[skim_type]]
}

#' Only show a subset of summary statistics after skimming
#'
#' This function is a variant of [dplyr::select()] designed to work with
#' `skim_df` objects. When using `focus()`, `skimr` metadata columns are kept,
#' and `skimr` print methods are still utilized. Otherwise, the signature and
#' behavior is identical to [dplyr::select()].
#'
#' @param .data A `skim_df` object.
#' @param ...  One or more unquoted expressions separated by commas. Variable
#'   names can be used as if they were positions in the data frame, so
#'   expressions like x:y can be used to select a range of variables.
#' @examples
#' # Compare
#' iris |>
#'   skim() |>
#'   dplyr::select(n_missing)
#'
#' iris |>
#'   skim() |>
#'   focus(n_missing)
#'
#' # This is equivalent to
#' iris |>
#'   skim() |>
#'   dplyr::select(skim_variable, skim_type, n_missing)
#' @export
focus <- function(.data, ...) {
  assert_is_skim_df(.data)
  reduced <- dplyr::select(.data, "skim_type", "skim_variable", ...)
  if (could_be_skim_df(reduced)) {
    skimmers <- reconcile_skimmers(
      reduced,
      group_names(.data),
      base_skimmers(.data)
    )
    reassign_skim_attrs(reduced, .data, skimmers_used = skimmers)
  } else {
    stop("Cannot drop 'skim_variable' or 'skim_type' columns")
  }
}

#' Create "long" skim output
#'
#' Skim results returned as a tidy long data frame with four columns:
#' variable, type, stat and formatted.
#'
#' @param .data A data frame or an object that can be coerced into a data frame.
#' @param ...  Columns to select for skimming. When none are provided, the
#'   default is to skim all columns.
#' @param skim_fun The skim function used.
#' @return A tibble
#' @examples
#' to_long(iris)
#' to_long(skim(iris))
#' @export
to_long <- function(.data, ..., skim_fun = skim) {
  UseMethod("to_long")
}

#' @describeIn to_long Skim a data frame and convert the results to a
#'   long data frame.
#' @export
to_long.default <- function(.data, ..., skim_fun = skim) {
  skimmed <- skim_fun(.data, ...)
  to_long(skimmed, ..., skim_fun)
}

#' @describeIn  to_long Transform a skim_df to a long data frame.
#' @export
to_long.skim_df <- function(.data, ..., skim_fun = skim) {
  .data <- dplyr::mutate(.data, dplyr::across(dplyr::everything(), as.character))
  tidyr::pivot_longer(
    .data,
    cols = c(-"skim_type", -"skim_variable"),
    names_to = "stat",
    values_to = "formatted",
    values_drop_na = TRUE
  )
}

test_that("You can parition a skim_df", {
  skimmed <- skim(iris)
  input <- partition(skimmed)
  expect_s3_class(input, "skim_list")
  expect_length(input, 2)
  expect_named(input, c("factor", "numeric"))
  attrs <- attributes(input)
  expect_equal(attrs$data_rows, 150)
  expect_equal(attrs$data_cols, 5)
  expect_identical(attrs$df_name, "`iris`")
  expect_identical(attrs$base_skimmers, c("n_missing", "complete_rate"))
  expect_identical(
    attrs$skimmers_used,
    list(
      numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist"),
      factor = c("ordered", "n_unique", "top_counts")
    )
  )

  # Subtables
  expect_s3_class(input$factor, c("one_skim_df", "tbl_df", "tbl", "data.frame"))
  expect_n_rows(input$factor, 1)
  expect_n_columns(input$factor, 6)
  expect_named(input$factor, c(
    "skim_variable", "n_missing", "complete_rate", "ordered", "n_unique",
    "top_counts"
  ))

  expect_s3_class(input$numeric, c("one_skim_df", "tbl_df", "tbl", "data.frame"))
  expect_n_rows(input$numeric, 4)
  expect_n_columns(input$numeric, 11)
  expect_named(input$numeric, c(
    "skim_variable", "n_missing", "complete_rate", "mean",
    "sd", "p0", "p25", "p50", "p75", "p100",
    "hist"
  ))
})

test_that("Partitioning works in a round trip", {
  skimmed <- skim(iris)
  partitioned <- partition(skimmed)
  input <- bind(partitioned)
  expect_identical(input, skimmed)
})

test_that("You can yank a subtable from a skim_df", {
  skimmed <- skim(iris)
  input <- yank(skimmed, "numeric")
  expect_s3_class(input, c("one_skim_df", "tbl_df", "tbl", "data.frame"))
  expect_n_rows(input, 4)
  expect_n_columns(input, 11)
  expect_named(input, c(
    "skim_variable", "n_missing", "complete_rate", "mean",
    "sd", "p0", "p25", "p50", "p75", "p100",
    "hist"
  ))
})

test_that("Partition is safe if some skimmers are missing", {
  skimmed <- skim(iris)
  reduced <- dplyr::select(skimmed, skim_variable, skim_type, n_missing)
  partitioned <- partition(reduced)
  expect_length(partitioned, 2)
  expect_named(partitioned, c("factor", "numeric"))
  expect_named(partitioned$numeric, c("skim_variable", "n_missing"))
})

test_that("Partition handles new columns", {
  skimmed <- skim(iris)
  expanded <- dplyr::mutate(
    skimmed,
    mean2 = numeric.mean^2,
    complete2 = complete_rate^2
  )
  partitioned <- partition(expanded)
  expect_named(partitioned$numeric, c(
    "skim_variable", "n_missing", "complete_rate", "mean",
    "sd", "p0", "p25", "p50", "p75", "p100",
    "hist", "mean2", "complete2"
  ))
})

test_that("focus() matches select(data, skim_type, skim_variable, ...)", {
  skimmed <- skim(iris)
  expected <- dplyr::select(
    skimmed, skim_type, skim_variable, n_missing
  )
  expect_equal(focus(skimmed, n_missing), expected, check.attributes = FALSE)
})

test_that("focus() does not allow dropping skim metadata columns", {
  skimmed <- skim(iris)
  expect_error(focus(skimmed, -skim_variable), "Cannot drop")
  expect_error(focus(skimmed, -skim_type), "Cannot drop")
})

test_that("skim_to_wide() returns a deprecation warning", {
  expect_warning(skim_to_wide(iris))
})

test_that("skim_to_list() returns a deprecation warning", {
  expect_warning(skim_to_list(iris))
})

test_that("to_long() returns a long tidy data frame with 4 columns", {
  skimmed_long <- to_long(iris)
  # Statistics from the skim_df  with values of NA are not included
  expect_n_rows(skimmed_long, 45)
  expect_equal(
    names(skimmed_long),
    c("skim_type", "skim_variable", "stat", "formatted")
  )
  expect_equal(length(unique(skimmed_long$stat)), 13)
  expect_equal(length(unique(skimmed_long$skim_type)), 2)
  expect_equal(length(unique(skimmed_long$skim_variable)), 5)
})

test_that("to_long() on a skim_df returns a long tidy df with 4 columns", {
  skimmed_long <- to_long(skim(iris))
  # Statistics from the skim_df  with values of NA are not included
  expect_n_rows(skimmed_long, 45)
  expect_equal(
    names(skimmed_long),
    c("skim_type", "skim_variable", "stat", "formatted")
  )
  expect_equal(length(unique(skimmed_long$stat)), 13)
  expect_equal(length(unique(skimmed_long$skim_type)), 2)
  expect_equal(length(unique(skimmed_long$skim_variable)), 5)
})

test_that("to_long() on a df and a skim_df from same df are identical", {
  expect_identical(to_long(skim(chickwts)), to_long(chickwts))
})

test_that("Skim prints a header for the entire output and each type", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  expect_snapshot({
    input <- skim(iris)
    input
    input$numeric.hist <- NULL
    input
  })
})

test_that("Skim prints a special header for grouped data frames", {
  skip_if_not(l10n_info()$`UTF-8`)
  skip_on_os("mac")
  withr::local_options(list(cli.unicode = FALSE))
  expect_snapshot( skim(dplyr::group_by(iris, Species)))
})

test_that("Skim lists print as expected", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  expect_snapshot(partition(skimmed))
})

test_that("knit_print produces expected results", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  input <- knit_print(skimmed)
  expect_s3_class(input, "knit_asis")
  expect_length(input, 1)
  expect_snapshot(cat(input))
})

test_that("knit_print works with skim summaries", {
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  summarized <- summary(skimmed)
  expect_snapshot(cat(knitr::knit_print(summarized)))
})

test_that("knit_print appropriately falls back to tibble printing", {
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  reduced <- dplyr::select(skimmed, skim_variable, numeric.mean)
  expect_snapshot({
    input <- knitr::knit_print(reduced)
  })
  expect_s3_class(input, "data.frame")
})

test_that("Summaries can be suppressed within knitr", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  options <- list(skimr_include_summary = FALSE)
  expect_snapshot(cat(knitr::knit_print(skimmed, options = options)))
})

test_that("Skim lists have a separate knit_print method", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  skim_list <- partition(skimmed)
  expect_snapshot(cat(knit_print(skim_list)))
})

test_that("You can yank a type from a skim_df and call knit_print", {
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  skim_one <- yank(skimmed, "factor")
  expect_snapshot(cat(knit_print(skim_one)))
})

test_that("Skim falls back to tibble::print.tbl() appropriately", {
  withr::local_options(list(cli.unicode = FALSE))
  
  expect_snapshot({
    input <- skim(iris)
    dplyr::select(input, numeric.mean)
  })
})

test_that("Print focused objects appropriately", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  skimmed <- skim(iris)
  expect_snapshot(focus(skimmed, n_missing))
})

test_that("Support for smaller consoles can be set with the width option", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  expect_snapshot(skim(iris))
})

test_that("Table header width can be controlled by an option", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  skimmed <- skim(iris)
  expect_snapshot(skimmed)
})

test_that("skimr creates appropriate output for Jupyter", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  skimmed <- skim(iris)
  expect_snapshot(skimmed)
})

test_that("Skim prints a header for the entire output and each type", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  input <- skim(iris)
  expect_print_matches_file(input, "print/default.txt")

  input$numeric.hist <- NULL
  expect_print_matches_file(input, "print/no-hist.txt",
    skip_on_windows = FALSE
  )
})

test_that("Skim prints a special header for grouped data frames", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  input <- skim(dplyr::group_by(iris, Species))
  expect_print_matches_file(input, "print/groups.txt")
})

test_that("Skim lists print as expected", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  input <- partition(skimmed)
  expect_print_matches_file(input, "print/list.txt")
})

test_that("knit_print produces expected results", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  input <- knit_print(skimmed)
  expect_s3_class(input, "knit_asis")
  expect_length(input, 1)
  expect_matches_file(input, "print/knit_print.txt")
})

test_that("knit_print works with skim summaries", {
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  summarized <- summary(skimmed)
  input <- knitr::knit_print(summarized)
  expect_matches_file(input, "print/knit_print-summary.txt")
})

test_that("knit_print appropriately falls back to tibble printing", {
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  reduced <- dplyr::select(skimmed, skim_variable, numeric.mean)
  printed <- testthat::capture_output(input <- knitr::knit_print(reduced))
  expect_matches_file(printed, "print/knit_print-fallback.txt")
  expect_s3_class(input, "data.frame")
})

test_that("Summaries can be suppressed within knitr", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  options <- list(skimr_include_summary = FALSE)
  input <- knitr::knit_print(skimmed, options = options)
  expect_matches_file(input, "print/knit_print-suppressed.txt")
})

test_that("Skim lists have a separate knit_print method", {
  skip_if_not(l10n_info()$`UTF-8`)
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  skim_list <- partition(skimmed)
  input <- knit_print(skim_list)
  expect_matches_file(input, "print/knit_print-skim_list.txt")
})

test_that("You can yank a type from a skim_df and call knit_print", {
  withr::local_options(list(cli.unicode = FALSE))
  skimmed <- skim(iris)
  skim_one <- yank(skimmed, "factor")
  input <- knit_print(skim_one)
  expect_matches_file(input, "print/knit_print-yank.txt")
})

test_that("make_utf8 produces the correct result ", {
  withr::local_options(list(cli.unicode = FALSE))
  input <- make_utf8(c("<U+2585><U+2587>"))
  correct <- "▅"
  expect_identical(input, correct)
})

test_that("Skim falls back to tibble::print.tbl() appropriately", {
  withr::local_options(list(cli.unicode = FALSE))
  input <- skim(iris)
  mean_only <- dplyr::select(input, numeric.mean)
  expect_print_matches_file(mean_only, "print/fallback.txt")
})

test_that("Print focused objects appropriately", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  skimmed <- skim(iris)
  input <- focus(skimmed, n_missing)
  expect_print_matches_file(input, "print/focus.txt")
})

test_that("Support for smaller consoles can be set with the width option", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  skimmed <- skim(iris)
  expect_print_matches_file(skimmed, "print/smaller.txt", width = 50)
})

test_that("Table header width can be controlled by an option", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  skimmed <- skim(iris)
  expect_print_matches_file(
    skimmed,
    "print/narrow-header.txt",
    skimr_table_header_width = 50
  )
})

test_that("skimr creates appropriate output for Jupyter", {
  withr::local_options(list(cli.unicode = FALSE))
  skip_if_not(l10n_info()$`UTF-8`)
  skimmed <- skim(iris)
  input <- testthat::capture_output(repr_text(skimmed))
  expect_matches_file(input, "print/repr.txt")
})

#' Fix unicode histograms on Windows
#'
#' This functions changes your session's locale to address issues with printing
#' histograms on Windows on versions of R below 4.2.1.
#'
#' There are known issues with printing the spark-histogram characters when
#' printing a data frame, appearing like this: "<U+2582><U+2585><U+2587>".
#' This longstanding problem originates in the low-level code for printing
#' dataframes. This was addressed in R version 4.2.1.
#'
#' @seealso [skim_without_charts()]
#' @export
fix_windows_histograms <- function() {
  message(
    "This function will change your system locale. It may have other ",
    "unintended effects."
  )
  response <- readline("Continue? (Y/n)")
  if (tolower(response) != "n") {
    Sys.setlocale("LC_CTYPE", "Chinese")
  } else {
    message("Locale was not changed.")
  }
  invisible(NULL)
}

#' Print `skim` objects
#'
#' `skimr` has custom print methods for all supported objects. Default printing
#' methods for `knitr`/ `rmarkdown` documents is also provided.
#'
#' @section Printing options:
#'
#' For better or for worse, `skimr` often produces more output than can fit in
#' the standard R console. Fortunately, most modern environments like RStudio
#' and Jupyter support more than 80 character outputs. Call
#' `options(width = 90)` to get a better experience with `skimr`.
#'
#' The print methods in `skimr` wrap those in the [tibble][tibble::formatting]
#' package. You can control printing behavior using the same global options.
#'
#' @section Behavior in `dplyr` pipelines:
#'
#' Printing a `skim_df` requires specific columns that might be dropped when
#' using [dplyr::select()] or [dplyr::summarize()] on a `skim_df`. In those
#' cases, this method falls back to [tibble::print.tbl()].
#'
#' @section Options for controlling print behavior:
#'
#' You can control the width rule line for the printed subtables with an option:
#' `skimr_table_header_width`.
#'
#' @inheritParams tibble::print.tbl
#' @seealso [tibble::trunc_mat()] For a list of global options for customizing
#'   print formatting. [crayon::has_color()] for the variety of issues that
#'   affect tibble's color support.
#' @param include_summary Whether a summary of the data frame should be printed
#' @param summary_rule_width Width of Data Summary cli rule, defaults to 40.
#' @name print
NULL

#' @describeIn print Print a skimmed data frame (`skim_df` from [skim()]).
#' @export
print.skim_df <- function(x,
                          include_summary = TRUE,
                          n = Inf,
                          width = Inf,
                          summary_rule_width = getOption(
                            "skimr_summary_rule_width",
                            default = 40
                          ),
                          ...) {
  if (is_skim_df(x) && nrow(x) > 0) {
    if (include_summary) {
      print(summary(x), .summary_rule_width = summary_rule_width, ...)
    }
    by_type <- partition(x)
    purrr::map(
      by_type,
      print,
      width = width,
      n = n,
      ...
    )
    invisible(x)
  } else {
    NextMethod("print")
  }
}

# Methods for correctly formatting a a `one_skim_df`. We leverage the
# customiztion options in `pillar` for this. It divides the results into: a
# header, which we customize; a body, where we strip some values; and a footer,
# which we drop. For more details, see
# https://pillar.r-lib.org/articles/extending.html

#' @importFrom pillar tbl_format_header
#' @export
tbl_format_header.one_skim_df <- function(x, setup, ...) {
  variable_type <- paste("Variable type:", attr(x, "skim_type"))
  rule <- cli::rule(
    line = 1,
    left = variable_type,
    width = getOption("skimr_table_header_width", getOption("width", 80))
  )
  # Add an empty line before the rule
  c("", rule)
}

#' @importFrom pillar ctl_new_pillar
#' @export
ctl_new_pillar.one_skim_df <- function(controller,
                                       x,
                                       width,
                                       ...,
                                       title = NULL) {
  out <- NextMethod()
  out$type <- NULL
  out
}

#' @describeIn print Print a `skim_list`, a list of `skim_df` objects.
#' @export
print.skim_list <- function(x,
                            n = Inf,
                            width = Inf,
                            ...) {
  nms <- names(x)
  attributes(x) <- NULL
  print(rlang::set_names(x, nms))
}


#' @describeIn print Print method for a `summary_skim_df` object.
#' @param .summary_rule_width the width for the main rule above the summary.
#' @export
print.summary_skim_df <- function(x, .summary_rule_width = 40, ...) {
  with_title <- c(
    cli::rule(line = 1, left = "Data Summary", width = .summary_rule_width),
    format(x)
  )
  writeLines(with_title)
}

#' Provide a default printing method for knitr.
#'
#' Instead of standard R output, `knitr` and `RMarkdown` documents will have
#' formatted [knitr::kable()] output on return. You can disable this by setting
#' the chunk option `render = normal_print`.
#'
#' The summary statistics for the original data frame can be disabled by setting
#' the `knitr` chunk option `skimr_include_summary = FALSE`. See
#' [knitr::opts_chunk] for more information. You can change the number of digits
#' shown in the printed table with the `skimr_digits` chunk option.
#'
#' Alternatively, you can call [collapse()] or [yank()] to get the particular
#' `skim_df` objects and format them however you like. One warning though.
#' Because histograms contain unicode characters, they can have unexpected
#' print results, as R as varying levels of unicode support. This affects
#' Windows users most commonly. Call `vignette("Using_fonts")` for more details.
#'
#' @seealso [knitr::kable()]
#' @inheritParams knitr::knit_print
#' @param options Options passed into the print function.
#' @return A `knit_asis` object. Which is used by `knitr` when rendered.
#' @importFrom knitr knit_print
#' @name knit_print
NULL

#' @describeIn knit_print Default `knitr` print for `skim_df` objects.
#' @export
knit_print.skim_df <- function(x, options = NULL, ...) {
  if (is_skim_df(x) && nrow(x) > 0) {
    if (options$skimr_include_summary %||% TRUE) {
      summary_stats <- summary(x)
      kabled <- knit_print(summary_stats)
    } else {
      kabled <- c()
    }

    by_type <- partition(x)
    knit_print_by_type(by_type, options, kabled)
  } else {
    NextMethod("knit_print")
  }
}

knit_print_by_type <- function(x, options, summary) {
  all_tables <- purrr::imap(x, knit_print_one, options)
  combined <- c("", summary, "", "", unlist(all_tables))
  knitr::asis_output(paste(combined, collapse = "\n"))
}

knit_print_one <- function(by_type, type, options) {
  kabled <- knitr::kable(
    by_type,
    digits = options$skimr_digits %||% 2
  )
  if (is_windows()) {
    kabled[] <- fix_unicode(kabled)
  }
  caption <- sprintf("**Variable type: %s**", type)
  c(caption, "", kabled, "", "")
}

#' @describeIn knit_print Default `knitr` print for a `skim_list`.
#' @export
knit_print.skim_list <- function(x, options = NULL, ...) {
  knit_print_by_type(x, options, NULL)
}

#' @describeIn knit_print Default `knitr` print within a partitioned `skim_df`.
#' @export
knit_print.one_skim_df <- function(x, options = NULL, ...) {
  kabled <- knit_print_one(x, attr(x, "skim_type"), options)
  combined <- c("", "", kabled, "")
  knitr::asis_output(paste(combined, collapse = "\n"))
}

#' @describeIn knit_print Default `knitr` print for `skim_df` summaries.
#' @export
knit_print.summary_skim_df <- function(x, options = NULL, ...) {
  summary_mat <- cbind(
    get_summary_dnames(x),
    get_summary_values(x),
    deparse.level = 0
  )
  kabled <- knitr::kable(
    summary_mat,
    table.attr = "style='width: auto;'
      class='table table-condensed'",
    caption = "Data summary"
  )

  knitr::asis_output(paste(kabled, collapse = "\n"))
}


#' Skimr printing within Jupyter notebooks
#'
#' This reproduces printed results in the console. By default Jupyter kernels
#' render the final object in the cell. We want the version printed by
#' `skimr` instead of the data that it contains.
#'
#' @param obj The object to \link{print} and then return the output.
#' @param ... ignored.
#' @return None. `invisible(NULL)`.
#' @importFrom repr repr_text
#' @name repr

#' @rdname repr
#' @export
repr_text.skim_df <- function(obj, ...) {
  print(obj)
}

#' @rdname repr
#' @export
repr_text.skim_list <- repr_text.skim_df

#' @rdname repr
#' @export
repr_text.one_skim_df <- repr_text.skim_df

#' Separate a big `skim_df` into smaller data frames, by type.
#'
#' The data frames produced by [skim()] are wide and sparse, filled with
#' columns that are mostly `NA`. For that reason, it can be convenient to
#' work with "by type" subsets of the original data frame. These smaller
#' subsets have their `NA` columns removed.
#'
#' `partition()` creates a list of smaller `skim_df` data frames. Each entry
#' in the list is a data type from the original `skim_df`. The inverse of
#' `partition()` is `bind()`, which takes the list and produces the original
#' `skim_df`. While `partition()` keeps all of the subtables as list entries,
#' `yank()` gives you a single subtable for a data type.
#'
#' @param data A `skim_df`.
#' @param skim_type A character scalar. The subtable to extract from a
#'   `skim_df`.
#' @return A `skim_list` of `skim_df`'s, by type.
#' @examples
#' # Create a wide skimmed data frame (a skim_df)
#' skimmed <- skim(iris)
#'
#' # Separate into a list of subtables by type
#' separate <- partition(skimmed)
#'
#' # Put back together
#' identical(bind(separate), skimmed)
#' # > TRUE
#'
#' # Alternatively, get the subtable of a particular type
#' yank(skimmed, "factor")
#' @export
partition <- function(data) {
  assert_is_skim_df(data)
  data_as_list <- split(data, data$skim_type)
  groups <- group_names(data)
  base <- base_skimmers(data)

  skimmers <- reconcile_skimmers(data, groups, base)

  # Check to see that there are at least one, known used skim function
  # within the data. This can be reduced after using focus(). This indexing
  # is used so that the lookup order matches that of `data_as_list`.
  has_skim_data <- lengths(skimmers)[names(data_as_list)] > 0
  elements_to_keep <- has_skim_data | any(base %in% names(data))
  reduced <- purrr::imap(
    data_as_list[elements_to_keep],
    simplify_skimdf,
    skimmers = skimmers,
    groups = groups,
    base = base
  )

  reassign_skim_attrs(
    reduced,
    data,
    class = c("skim_list", "list"),
    skimmers_used = skimmers
  )
}

#' Align the skimmers_used attribute with the current columns in the data
#'
#' This catches the case where users partition (or more likely print) a data
#' frame that has had columns added after skimming.
#' @noRd
reconcile_skimmers <- function(data, groups, base) {
  all_columns <- colnames(data)
  skimmers_used <- skimmers_used(data)
  if (length(skimmers_used) == 0) {
    return(skimmers_used)
  }
  skimmers_from_names <- skimmers_from_names(all_columns, skimmers_used)
  with_base_columns <- c(
    "skim_variable",
    "skim_type",
    groups,
    base,
    collapse_skimmers(skimmers_used)
  )
  matched_cols <- dplyr::intersect(all_columns, with_base_columns)
  extra_cols <- dplyr::setdiff(all_columns, with_base_columns)
  if (length(extra_cols) > 0) {
    grouped <- dplyr::group_by(data, .data$skim_type)
    complete_by_type <- dplyr::summarize_at(
      grouped,
      extra_cols,
      ~ !all(is.na(.x))
    )
    complete_cols <- purrr::pmap(
      complete_by_type,
      get_complete_columns,
      names = extra_cols
    )
    new_cols_by_type <- rlang::set_names(
      complete_cols,
      complete_by_type$skim_type
    )
    skimmers_from_names <- purrr::list_merge(
      skimmers_from_names,
      !!!new_cols_by_type
    )
  }

  skimmers_from_names
}

skimmers_from_names <- function(names, skimmers) {
  matched <- purrr::imap(skimmers, match_skimmers, names)
  purrr::set_names(matched, names(skimmers))
}

match_skimmers <- function(fun_names, type, values_to_match) {
  stripped_values <- stringr::str_remove(
    values_to_match,
    paste0("^", type, "\\.")
  )
  dplyr::intersect(stripped_values, fun_names)
}

collapse_skimmers <- function(skimmers_used) {
  with_type <- purrr::imap(skimmers_used, ~ paste(.y, .x, sep = "."))
  purrr::flatten_chr(with_type)
}

get_complete_columns <- function(skim_type, ..., names) {
  names[c(...)]
}

#' For each type subtable, only select columns generated by this type's skimmers
#' This function also catches the case where the user removed columns from
#' the skim_df
#' @noRd
simplify_skimdf <- function(data, skim_type, skimmers, groups, base) {
  stopifnot(has_variable_column(data))
  names(data) <- stringr::str_remove(names(data), paste0(skim_type, "\\."))
  keep <- c("skim_variable", groups, base, skimmers[[skim_type]])
  out <- dplyr::select(data, tidyselect::any_of(keep))

  structure(
    out,
    class = c("one_skim_df", "tbl_df", "tbl", "data.frame"),
    skim_type = skim_type
  )
}

#' @describeIn partition The inverse of a `partition()`. Rebuild the original
#'   `skim_df`.
#' @export
bind <- function(data) {
  assert_is_skim_list(data)
  with_namespaces <- purrr::imap(data, add_namespaces)
  combined <- dplyr::bind_rows(!!!with_namespaces, .id = "skim_type")
  reassign_skim_attrs(combined, data)
}

add_namespaces <- function(data, skim_type) {
  base <- base_skimmers(data)
  meta_columns <- c("skim_variable", dplyr::group_vars(data), base)
  no_meta_columns <- dplyr::setdiff(names(data), meta_columns)
  with_namespace <- paste(skim_type, no_meta_columns, sep = ".")

  # TODO(michaelquinn32): Drop this when vctrs interface works correctly.
  names(data) <- c(meta_columns, with_namespace)
  attr(data, "skim_type") <- NULL
  tibble::as_tibble(data)
}

#' @describeIn partition Extract a subtable from a `skim_df` with a particular
#'   type.
#' @export
yank <- function(data, skim_type) {
  partition(data)[[skim_type]]
}

#' Only show a subset of summary statistics after skimming
#'
#' This function is a variant of [dplyr::select()] designed to work with
#' `skim_df` objects. When using `focus()`, `skimr` metadata columns are kept,
#' and `skimr` print methods are still utilized. Otherwise, the signature and
#' behavior is identical to [dplyr::select()].
#'
#' @param .data A `skim_df` object.
#' @param ...  One or more unquoted expressions separated by commas. Variable
#'   names can be used as if they were positions in the data frame, so
#'   expressions like x:y can be used to select a range of variables.
#' @examples
#' # Compare
#' iris %>%
#'   skim() %>%
#'   dplyr::select(n_missing)
#'
#' iris %>%
#'   skim() %>%
#'   focus(n_missing)
#'
#' # This is equivalent to
#' iris %>%
#'   skim() %>%
#'   dplyr::select(skim_variable, skim_type, n_missing)
#' @export
focus <- function(.data, ...) {
  assert_is_skim_df(.data)
  reduced <- dplyr::select(.data, "skim_type", "skim_variable", ...)
  if (could_be_skim_df(reduced)) {
    skimmers <- reconcile_skimmers(
      reduced,
      group_names(.data),
      base_skimmers(.data)
    )
    reassign_skim_attrs(reduced, .data, skimmers_used = skimmers)
  } else {
    stop("Cannot drop 'skim_variable' or 'skim_type' columns")
  }
}

#' Create "long" skim output
#'
#' Skim results returned as a tidy long data frame with four columns:
#' variable, type, stat and formatted.
#'
#' @param .data A data frame or an object that can be coerced into a data frame.
#' @param ...  Columns to select for skimming. When none are provided, the
#'   default is to skim all columns.
#' @param skim_fun The skim function used.
#' @return A tibble
#' @examples
#' to_long(iris)
#' to_long(skim(iris))
#' @export
to_long <- function(.data, ..., skim_fun = skim) {
  UseMethod("to_long")
}

#' @describeIn to_long Skim a data frame and convert the results to a
#'   long data frame.
#' @export
to_long.default <- function(.data, ..., skim_fun = skim) {
  skimmed <- skim_fun(.data, ...)
  to_long(skimmed, ..., skim_fun)
}

#' @describeIn  to_long Transform a skim_df to a long data frame.
#' @export
to_long.skim_df <- function(.data, ..., skim_fun = skim) {
  tidyr::gather(
    .data,
    key = "stat",
    value = "formatted",
    na.rm = TRUE,
    -"skim_type",
    -"skim_variable"
  )
}

#' Functions for accessing skim_df attributes
#'
#' These functions simplify access to attributes contained within a `skim_df`.
#' While all attributes are read-only, being able to extract this information
#' is useful for different analyses. These functions should always be preferred
#' over calling base R's attribute functions.
#'
#' @param object A `skim_df` or `skim_list`.
#' @return Data contained within the requested `skimr` attribute.
#' @name skim-attr
NULL

#' @describeIn skim-attr Get the number of rows in the skimmed data frame.
#' @export
data_rows <- function(object) {
  attr(object, "data_rows")
}

#' @describeIn skim-attr Get the number of columns in the skimmed data frame.
#' @export
data_cols <- function(object) {
  attr(object, "data_cols")
}

#' @describeIn skim-attr Get the name of the skimmed data frame. This is only
#'   available in contexts where the name can be looked up. This is often not
#'   the case within a pipeline.
#' @export
df_name <- function(object) {
  attr(object, "df_name")
}

#' @describeIn skim-attr Get the key of the skimmed data.table. This is only
#'   available in contexts where `data` is of class `data.table`.
#' @export
dt_key <- function(object) {
  attr(object, "dt_key")
}

#' @describeIn skim-attr Get the names of the groups in the original data frame.
#'   Only available if the data was grouped. Otherwise, `NULL`.
#' @export
group_names <- function(object) {
  attr(object, "groups")
}

#' @describeIn skim-attr Get the names of the base skimming functions used.
#' @export
base_skimmers <- function(object) {
  attr(object, "base_skimmers")
}

#' @describeIn skim-attr Get the names of the skimming functions used, separated
#'   by data type.
#' @export
skimmers_used <- function(object) {
  attr(object, "skimmers_used")
}


#' Test if an object is compatible with `skimr`
#'
#' Objects within `skimr` are identified by a class, but they require additional
#' attributes and data columns for all operations to succeed. These checks help
#' ensure this. While they have some application externally, they are mostly
#' used internally.
#'
#' Most notably, a `skim_df` has columns `skim_type` and `skim_variable`. And
#' has the following special attributes
#'
#' * `data_rows`: n rows in the original data
#' * `data_cols`: original number of columns
#' * `df_name`: name of the original data frame
#' * `dt_key`: name of the key if original is a data.table
#' * `groups`: if there were group variables
#' * `base_skimmers`: names of functions applied to all skim types
#' * `skimmers_used`: names of functions used to skim each type
#'
#' The functions in these checks work like [all.equal()]. The return `TRUE` if
#' the check passes, or otherwise notifies why the check failed. This makes them
#' more useful when throwing errors.
#'
#' @param object Any `R` object.
#' @name skim-obj
NULL

#' @describeIn skim-obj Does the object have the `skim_type` column?
#' @export
has_type_column <- function(object) {
  make_issue("skim_type" %in% names(object), "missing column `skim_type`")
}

#' @describeIn skim-obj Does the object have the `skim_variable` column?
#' @export
has_variable_column <- function(object) {
  make_issue(
    "skim_variable" %in% names(object), "missing column `skim_variable`"
  )
}

#' @describeIn skim-obj Does the object have the appropriate `skimr` attributes?
#' @export
has_skimr_attributes <- function(object) {
  skimr_attrs <- c(
    "data_rows", "data_cols", "df_name", "dt_key", "base_skimmers",
    "skimmers_used"
  )
  missing <- !(skimr_attrs %in% names(attributes(object)))
  make_issue(!any(missing), show_missing_attributes(skimr_attrs, missing))
}

show_missing_attributes <- function(attributes, missing) {
  missing_msg <- if (length(missing) > 0) {
    paste0(attributes[missing], collapse = ", ")
  } else {
    ""
  }
  paste("missing attributes:", missing_msg)
}

#' @describeIn skim-obj Does the object have a `skim_type` attribute? This makes
#'   it a `one_skim_df`.
#' @export
has_skim_type_attribute <- function(object) {
  make_issue(
    "skim_type" %in% names(attributes(object)),
    "missing attribute: `skim_type`"
  )
}

#' @describeIn skim-obj Does the object have skimmers?
#' @export
has_skimmers <- function(object) {
  base <- base_skimmers(object)
  skimmers <- reconcile_skimmers(
    object,
    group_names(object),
    base
  )
  make_issue(
    any(lengths(skimmers) > 0, base %in% names(object)),
    "does not have defined skimmers."
  )
}

#' @describeIn skim-obj Is the object a data frame?
#' @export
is_data_frame <- function(object) {
  make_issue(inherits(object, "data.frame"), "not a data.frame")
}

make_issue <- function(check, message) {
  structure(check, message = if (!check) message else character())
}

#' @describeIn skim-obj Is the object a `skim_df`?
#' @export
is_skim_df <- function(object) {
  check_issues(
    "Object is not a `skim_df`",
    is_data_frame(object),
    has_type_column(object),
    has_variable_column(object),
    has_skimr_attributes(object),
    has_skimmers(object)
  )
}

#' @describeIn skim-obj Is the object a `one_skim_df`? This is similar to a
#'   `skim_df`, but does not have the `type` column. That is stored as an
#'   attribute instead.
#' @export
is_one_skim_df <- function(object) {
  check_issues(
    "Object is not a `one_skim_df`",
    is_data_frame(object),
    has_skim_type_attribute(object),
    has_variable_column(object),
    has_skimr_attributes(object)
  )
}

#' @describeIn skim-obj Is the object a `skim_list`?
#' @export
is_skim_list <- function(object) {
  check_issues(
    "Object is not a `skim_list`",
    has_skimr_attributes(object),
    !!!purrr::map(object, is_one_skim_df)
  )
}

#' @describeIn skim-obj Is this a data frame with `skim_variable` and
#'  `skim_type` columns?
#' @export
could_be_skim_df <- function(object) {
  check_issues(
    "Object cannot behave like a `skim_df`",
    is_data_frame(object),
    has_variable_column(object),
    has_type_column(object)
  )
}

check_issues <- function(condition, ...) {
  issues <- rlang::list2(...)
  msgs <- purrr::map(issues, ~ attr(.x, "message"))
  check <- all(purrr::flatten_lgl(issues))
  message <- if (check) {
    character()
  } else {
    paste0(condition, ": ", paste0(unlist(msgs), collapse = "; "))
  }
  structure(check, message = message)
}

#' @describeIn skim-obj Stop if the object is not a `skim_df`.
#' @export
assert_is_skim_df <- function(object) {
  assert(is_skim_df(object))
}

#' @describeIn skim-obj Stop if the object is not a `skim_list`.
#' @export
assert_is_skim_list <- function(object) {
  assert(is_skim_list(object))
}

#' @describeIn skim-obj Stop if the object is not a `one_skim_df`.
#' @export
assert_is_one_skim_df <- function(object) {
  assert(is_one_skim_df(object))
}

assert <- function(check) {
  if (check) {
    invisible(NULL)
  } else {
    stop(attr(check, "message"))
  }
}

#' Remove `skimr` class if not needed.
#' @noRd
strip_skim_attrs <- function(object) {
  attrs <- attributes(object)
  stripped <- purrr::list_modify(
    attrs,
    class = class(object)[-1],
    data_rows = NULL,
    data_cols = NULL,
    df_name = NULL,
    dt_key = NULL,
    base_skimmers = NULL,
    skimmers_used = NULL
  )
  attributes(object) <- stripped
  object
}

#' Pass attributes from a `skimr` object to a new object.
#' @noRd
reassign_skim_attrs <- function(object, skim_df, ...) {
  defaults <- list(
    class = c("skim_df", "tbl_df", "tbl", "data.frame"),
    data_rows = data_rows(skim_df),
    data_cols = data_cols(skim_df),
    df_name = df_name(skim_df),
    dt_key = dt_key(skim_df),
    groups = group_names(skim_df),
    base_skimmers = base_skimmers(skim_df),
    skimmers_used = skimmers_used(skim_df)
  )
  updated <- purrr::list_modify(defaults, ...)
  assign_new_attributes(object, !!!updated)
}

assign_new_attributes <- function(object, ...) {
  original <- attributes(object)
  attributes(object) <- purrr::list_modify(original, ...)
  object
}

test_that("skim produces skim_df objects", {
  skimmed <- skim(iris)
  expect_true(is_skim_df(skimmed))
  expect_false(is_skim_list(skimmed))
  expect_error(assert_is_skim_df(skimmed), NA)
  expect_error(assert_is_skim_list(skimmed))
})

test_that("Partition produces skim_list objects", {
  skimmed <- skim(iris)
  split <- partition(skimmed)
  expect_false(is_skim_df(split))
  expect_true(is_skim_list(split))
  expect_error(assert_is_skim_df(split), "not a data.frame")
  expect_error(assert_is_skim_list(split), NA)

  expect_true(is_one_skim_df(split[[1]]))
  expect_error(assert_is_one_skim_df(split[[1]]), NA)
})

test_that("Error messages are correct", {
  expect_error(assert_is_skim_df(iris), "missing column `skim_type`")
  expect_error(assert_is_skim_df(iris), "missing column `skim_variable`")
  expect_error(assert_is_skim_df(iris), "missing attributes:")
})

test_that("Bind produces skim_df objects", {
  skimmed <- skim(iris)
  split <- partition(skimmed)
  combined <- bind(split)
  expect_true(is_skim_df(combined))
  expect_false(is_skim_list(combined))
  expect_error(assert_is_skim_df(combined), NA)
  expect_error(assert_is_skim_list(combined))
})

test_that("Creating a skim_df requires skim_type and skim_variable columns", {
  expect_true(could_be_skim_df(
    data.frame(
      skim_type = "t",
      skim_variable = "v"
    )
  ))
  expect_false(could_be_skim_df(
    list(
      type = "t",
      variable = "v"
    )
  ))
  expect_false(could_be_skim_df(
    data.frame(
      variable = "v"
    )
  ))
  expect_false(could_be_skim_df(
    data.frame(
      type = "t"
    )
  ))
})

# Adapted from
# https://stackoverflow.com/questions/28248457
fix_unicode <- function(char) {
  m <- gregexpr("<U\\+[0-9a-fA-F]{4}>", char)
  codes <- regmatches(char, m)
  chars <- lapply(codes, make_utf8)
  regmatches(char, m) <- chars
  enc2utf8(char)
}

make_utf8 <- function(x) {
  if (length(x) < 1) {
    x
  } else {
    digits <- substring(x, 4, 7)
    integers <- strtoi(sprintf("0x%s", digits))
    purrr::map_chr(integers, intToUtf8)
  }
}

is_windows <- function() {
  tolower(Sys.info()[["sysname"]]) == "windows"
}

#' Fix unicode histograms on Windows
#'
#' This functions changes your session's locale to address issues with printing
#' histograms on Windows.
#'
#' There are known issues with printing the spark-histogram characters when
#' printing a data frame, appearing like this: "<U+2582><U+2585><U+2587>".
#' This longstanding problem originates in the low-level code for printing
#' dataframes.
#'
#' @seealso [skim_without_charts()]
#' @export
fix_windows_histograms <- function() {
  message(
    "This function will change your system locale. It may have other ",
    "unintended effects."
  )
  response <- readline("Continue? (Y/n)")
  if (tolower(response) != "n") {
    Sys.setlocale("LC_CTYPE", "Chinese")
  } else {
    message("Locale was not changed.")
  }
  invisible(NULL)
}

#' Deprecated functions from skimr v1
#'
#' Skimr used to offer functions that combined skimming with a secondary effect,
#' like reshaping the data, building a list or printing the results. Some of
#' these behaviors are no longer necessary. [skim()] always returns a wide
#' data frame. Others have been replaced by functions that do a single thing.
#' [partition()] creates a list-like object from a skimmed data frame.
#'
#' @param .data A tibble, or an object that can be coerced into a tibble.
#' @param ...  Columns to select for skimming. When none are provided, the
#'   default is to skim all columns.
#' @return Either A `skim_df` or a `skim_list` object.
#' @name deprecated-v1
NULL

#' @describeIn deprecated-v1 [skim()] always produces a wide data frame.
#' @export
skim_to_wide <- function(.data, ...) {
  .Deprecated("skim()")
  skim(.data, ...)
}

#' @describeIn deprecated-v1 [partition()] creates a list.
#' @export
skim_to_list <- function(.data, ...) {
  .Deprecated("partition(skim())")
  skim(.data, ...) %>% partition()
}

#' @describeIn deprecated-v1 [print()] and [skim_with()] set options.
#' @export
skim_format <- function(...) {
  .Deprecated(
    "print()",
    msg = paste(
      "Formatting options are now in print() or as function",
      "arguments in skim_with()."
    )
  )
}

#' Set or add the summary functions for a particular type of data
#'
#' While skim is designed around having an opinionated set of defaults, you
#' can use this function to change the summary statistics that it returns.
#'
#' `skim_with()` is a closure: a function that returns a new function. This
#' lets you have several skimming functions in a single R session, but it
#' also means that you need to assign the return of `skim_with()` before
#' you can use it.
#'
#' You assign values within `skim_with` by using the [sfl()] helper (`skimr`
#' function list). This helper behaves mostly like [dplyr::funs()], but lets
#' you also identify which skimming functions you want to remove, by setting
#' them to `NULL`. Assign an `sfl` to each column type that you wish to modify.
#'
#' Functions that summarize all data types, and always return the same type
#' of value, can be assigned to the `base` argument. The default base skimmers
#' compute the number of missing values [n_missing()] and the rate of values
#' being complete, i.e. not missing, [complete_rate()].
#'
#' When `append = TRUE` and local skimmers have names matching the names of
#' entries in the default `skim_function_list`, the values in the default list
#' are overwritten. Similarly, if `NULL` values are passed within `sfl()`, these
#' default skimmers are dropped. Otherwise, if `append = FALSE`, only the
#' locally-provided skimming functions are used.
#'
#' Note that `append` only applies to the `typed` skimmers (i.e. non-base).
#' See [get_default_skimmer_names()] for a list of defaults.
#'
#' @param ... One or more (`sfl`) `skimmer_function_list` objects, with an
#'   argument name that matches a particular data type.
#' @param base An `sfl` that sets skimmers for all column types.
#' @param append Whether the provided options should be in addition to the
#'   defaults already in `skim`. Default is `TRUE`.
#' @return A new `skim()` function. This is callable. See [skim()] for more
#'   details.
#' @examples
#' # Use new functions for numeric functions. If you don't provide a name,
#' # one will be automatically generated.
#' my_skim <- skim_with(numeric = sfl(median, mad), append = FALSE)
#' my_skim(faithful)
#'
#' # If you want to remove a particular skimmer, set it to NULL
#' # This removes the inline histogram
#' my_skim <- skim_with(numeric = sfl(hist = NULL))
#' my_skim(faithful)
#'
#' # This works with multiple skimmers. Just match names to overwrite
#' my_skim <- skim_with(numeric = sfl(iqr = IQR, p25 = NULL, p75 = NULL))
#' my_skim(faithful)
#'
#' # Alternatively, set `append = FALSE` to replace the skimmers of a type.
#' my_skim <- skim_with(numeric = sfl(mean = mean, sd = sd), append = FALSE)
#'
#' # Skimmers are unary functions. Partially apply arguments during assigment.
#' # For example, you might want to remove NA values.
#' my_skim <- skim_with(numeric = sfl(iqr = ~ IQR(., na.rm = TRUE)))
#'
#' # Set multiple types of skimmers simultaneously.
#' my_skim <- skim_with(numeric = sfl(mean), character = sfl(length))
#'
#' # Or pass the same as a list, unquoting the input.
#' my_skimmers <- list(numeric = sfl(mean), character = sfl(length))
#' my_skim <- skim_with(!!!my_skimmers)
#'
#' # Use the v1 base skimmers instead.
#' my_skim <- skim_with(base = sfl(
#'   missing = n_missing,
#'   complete = n_complete,
#'   n = length
#' ))
#'
#' # Remove the base skimmers entirely
#' my_skim <- skim_with(base = NULL)
#' @importFrom rlang .data
#' @export
skim_with <- function(...,
                      base = sfl(
                        n_missing = n_missing,
                        complete_rate = complete_rate
                      ),
                      append = TRUE) {
  stopifnot(is.null(base) || inherits(base, "skimr_function_list"))
  local_skimmers <- validate_assignment(...)

  function(data, ..., .data_name = NULL) {
    if (is.null(.data_name)) {
      .data_name <- rlang::expr_label(substitute(data))
    }

    if (!inherits(data, "data.frame")) {
      data <- as.data.frame(data)
    }
    stopifnot(inherits(data, "data.frame"))

    selected <- names(tidyselect::eval_select(rlang::expr(c(...)), data))
    if (length(selected) == 0) {
      selected <- names(data)
    }

    grps <- dplyr::groups(data)
    if (length(grps) > 0) {
      group_variables <- selected %in% as.character(grps)
      selected <- selected[!group_variables]
    } else {
      attr(data, "groups") <- list()
    }


    skimmers <- purrr::map(
      selected, get_final_skimmers, data, local_skimmers, append
    )
    types <- purrr::map_chr(skimmers, "skim_type")
    unique_skimmers <- reduce_skimmers(skimmers, types)
    combined_skimmers <- purrr::map(unique_skimmers, join_with_base, base)
    ready_to_skim <- tibble::tibble(
      skim_type = unique(types),
      skimmers = purrr::map(combined_skimmers, mangle_names, names(base$funs)),
      skim_variable = split(selected, types)[unique(types)]
    )
    
    grouped <- dplyr::group_by(ready_to_skim, .data$skim_type)
    nested <- dplyr::summarize(
      grouped,
      skimmed = purrr::map2(
      .data$skimmers, .data$skim_variable, skim_by_type, data
      )
    )
    structure(
      tidyr::unnest(nested, "skimmed"),
      class = c("skim_df", "tbl_df", "tbl", "data.frame"),
      data_rows = nrow(data),
      data_cols = ncol(data),
      df_name = .data_name,
      dt_key = get_dt_key(data),
      groups = dplyr::group_vars(data),
      base_skimmers = names(base$funs),
      skimmers_used = get_skimmers_used(unique_skimmers)
    )
  }
}

#' Process arguments provided in `skim_with`
#'
#' Make sure that arguments provided in `skim_with()` have names. Also,
#' check if we are defining a new skimming type dynamically.
#'
#' @keywords internal
#' @noRd
validate_assignment <- function(...) {
  to_assign <- rlang::list2(...)

  if (length(to_assign) < 1) {
    return(to_assign)
  }

  # Need to cope with case where ... is a list already
  if (!inherits(to_assign[[1]], "skimr_function_list")) {
    to_assign <- to_assign[[1]]
  }

  proposed_names <- names(to_assign)
  if (!all(nzchar(proposed_names)) || is.null(proposed_names) ||
    anyNA(proposed_names)) {
    stop("skim_with requires all arguments to be named.", call. = FALSE)
  }

  defaults <- get_default_skimmers()
  existing <- proposed_names %in% names(defaults)

  if (!all(existing) && length(defaults) > 0) {
    collapsed <- paste(proposed_names[!existing], collapse = ", ")
    message(
      "Creating new skimming functions for the following classes: ",
      collapsed, ".\nThey did not have recognized defaults. Call ",
      "get_default_skimmers() for more information."
    )
  }
  to_assign
}

#' Sets the appropriate key value when working with `data.table`
#' @keywords internal
#' @noRd
get_dt_key <- function(data) {
  if (inherits(data, "data.table")) {
    dt_key <- data.table::key(data)
    if (is.null(dt_key)) {
      dt_key <- "NULL"
    }
    paste(dt_key, collapse = ", ")
  } else {
    NA # Will never be NA if `data` is a data.table
  }
}

#' Combine local and default skimmers for each column
#'
#' Get the default skimmers for the current column using S3 dispatch for
#' [get_skimmers()]. Get the user-provided local skimmers from [skim_with()].
#' If no local skimmers are provided, use the defaults. Otherwise, merge the
#' local and default skimmers with the following rules.
#'
#'   - If `append = FALSE` of if the local and default types differ, use only
#'     the locals.
#'   - Else, replace the default values with the local values.
#'
#' @param column A character scalar. The column name.
#' @param data The data frame to summarize.
#' @param local_skimmers A list of `sfl` objects. Skimmers defined using
#'   `skim_with()`
#' @param append Same as above.
#' @param base Same as above.
#' @noRd
get_final_skimmers <- function(column, data, local_skimmers, append) {
  defaults <- get_skimmers(data[[column]])
  all_classes <- skim_class(data[[column]])
  locals <- get_local_skimmers(all_classes, local_skimmers)

  if (!nzchar(defaults$skim_type)) {
    msg <- sprintf(
      "Default skimming functions for column [%s] with class [%s]",
      column, paste(all_classes, collapse = ", ")
    )
    stop(msg, " did not have value for its `.class` argument. Please ",
      "investigate the definition of the associated S3 method.",
      call. = FALSE
    )
  }

  if (is.null(locals$funs)) {
    if (defaults$skim_type == "default") {
      msg <- sprintf(
        "Couldn't find skimmers for class: %s;",
        paste(all_classes, collapse = ", ")
      )
      warning(msg,
        " No user-defined `sfl` provided. Falling back to `character`.",
        call. = FALSE
      )
      data[[column]] <- as.character(data[[column]])
      skimmers <- defaults
      skimmers$skim_type <- "character"
    } else {
      skimmers <- defaults
    }
  } else {
    skimmers <- merge_skimmers(locals, defaults, append)
  }
  skimmers
}

skim_class <- function(column) {
  base_class <- class(column)
  if (any(base_class %in% c("double", "integer"))) {
    c(base_class, "numeric")
  } else {
    base_class
  }
}

get_local_skimmers <- function(classes, local_skimmers) {
  local_classes <- intersect(classes, names(local_skimmers))
  if (length(local_classes) == 0) {
    return(NULL)
  }

  first_class <- local_classes[[1]]
  out <- local_skimmers[[first_class]]
  out$skim_type <- first_class
  out
}

merge_skimmers <- function(locals, defaults, append) {
  if (!append || locals$skim_type != defaults$skim_type) {
    locals
  } else {
    defaults$funs <- purrr::compact(purrr::list_modify(
      defaults$funs, !!!locals$funs
    ))
    defaults
  }
}

reduce_skimmers <- function(skimmers, types) {
  named <- rlang::set_names(skimmers, types)
  named[unique(types)]
}

join_with_base <- function(skimmers, base) {
  skimmers$funs <- c(base$funs, skimmers$funs)
  skimmers
}

get_skimmers_used <- function(skimmers) {
  types <- names(skimmers)
  function_names <- purrr::map(skimmers, ~ names(.x$funs))
  rlang::set_names(function_names, types)
}

NAME_DELIMETER <- "~!@#$%^&*()-+"
mangle_names <- function(skimmers, base_names) {
  fun_names <- names(skimmers$funs)
  prefixes <- ifelse(
    fun_names %in% base_names,
    NAME_DELIMETER,
    paste0(NAME_DELIMETER, skimmers$skim_type, ".")
  )
  mangled <- paste0(prefixes, fun_names)
  skimmers$funs <- rlang::set_names(skimmers$funs, mangled)
  skimmers
}


#' Generate one or more rows of a `skim_df`, using one column
#'
#' Call all of the skimming functions on the single column, using grouped
#' variants, if necessary.
#'
#' We expect one row per variable/ group. To do this we need to take the
#' processed results, find the appropriate columns for each variable and
#' restack them. This uses a small hack that rests on the naming convention
#' of data frame produced by `summarize_at`, which uses the following scheme:
#'
#'  - `variable_name` + `_` + `function_name`
#'
#' To avoid inappropriately assigning the columns to the wrong variable, we
#' mangle the function names. That way, each set of relevant columns begin
#' with the column name + `_` + our internal delimiter.
#'
#' @param mangled_skimmers The `sfl`'s whose function names have been mangled.
#' @param variable_names The names of columns in the original data, matching a
#'   data type, that will be summarized.
#' @param data The original data.
#' @keywords internal
#' @noRd
skim_by_type <- function(mangled_skimmers, variable_names, data) {
  UseMethod("skim_by_type", data)
}

#' @export
skim_by_type.grouped_df <- function(mangled_skimmers, variable_names, data) {
  group_columns <- dplyr::groups(data)
  grouped <- dplyr::group_by(data, !!!group_columns)
  skimmed <- dplyr::summarize(
    grouped,
    dplyr::across(tidyselect::any_of(variable_names), mangled_skimmers$funs)
  )
  build_results(skimmed, variable_names, group_columns)
}

#' @export
skim_by_type.data.frame <- function(mangled_skimmers, variable_names, data) {
  skimmed <- dplyr::summarize(
    data,
    dplyr::across(tidyselect::any_of(variable_names), mangled_skimmers$funs)
  )
  build_results(skimmed, variable_names, NULL)
}

#' @export
skim_by_type.data.table <- function(mangled_skimmers, variable_names, data) {
  data <- tibble::as_tibble(data)
  skimmed <- dplyr::summarize(
    data,
    dplyr::across(tidyselect::any_of(variable_names), mangled_skimmers$funs)
  )
  build_results(skimmed, variable_names, NULL)
}

#' Summarize returns a single row data frame, make it tall.
#' @noRd
build_results <- function(skimmed, variable_names, groups) {
  if (length(variable_names) > 1) {
    out <- tibble::tibble(
      skim_variable = variable_names,
      by_variable = purrr::map(variable_names, reshape_skimmed, skimmed, groups)
    )
    tidyr::unnest(out, "by_variable")
  } else {
    out <- dplyr::select(
      as.data.frame(skimmed),
      !!!groups,
      tidyselect::contains(NAME_DELIMETER)
    )
    tibble::tibble(
      skim_variable = variable_names,
      !!!set_clean_names(out)
    )
  }
}

reshape_skimmed <- function(column, skimmed, groups) {
  delim_name <- paste0(column, "_", NAME_DELIMETER)
  out <- dplyr::select(
    tibble::as_tibble(skimmed),
    !!!groups,
    tidyselect::starts_with(delim_name, ignore.case = FALSE)
  )
  set_clean_names(out)
}

set_clean_names <- function(out) {
  separated <- strsplit(names(out), NAME_DELIMETER, fixed = TRUE)
  clean_names <- purrr::map_chr(separated, ~ .x[length(.x)])
  rlang::set_names(out, clean_names)
}

#' Skim a data frame
#'
#' This package provides an alternative to the default summary functions
#' within R. The package's API is tidy, functions take data frames, return
#' data frames and can work as part of a pipeline. The returned `skimr`
#' object is subsettable and offers a human readable output.
#'
#' `skimr` is opinionated, providing a strong set of summary statistics
#' that are generated for a variety of different data types. It is also
#' provides an API for customization. Users can change both the functions
#' dispatched and the way the results are formatted.
#'
#' @importFrom rlang .data
#' @name skimr-package
#' @aliases skimr
#' @docType package
NULL


# Imports -----------------------------------------------------------------

#' @importFrom rlang %||%

#' @importFrom magrittr %>%
#' @export
magrittr::`%>%`

#' @importFrom tidyselect contains
#' @export
tidyselect::contains

#' @importFrom tidyselect ends_with
#' @export
tidyselect::ends_with

#' @importFrom tidyselect everything
#' @export
tidyselect::everything

#' @importFrom tidyselect matches
#' @export
tidyselect::matches

#' @importFrom tidyselect num_range
#' @export
tidyselect::num_range

#' @importFrom tidyselect one_of
#' @export
tidyselect::one_of

#' @importFrom tidyselect starts_with
#' @export
tidyselect::starts_with

#' @include skim_with.R stats.R
NULL

#' Retrieve the summary functions for a specific data type
#'
#' These functions are used to set the default skimming functions for a data
#' type. They are combined with the base skim function list (`sfl`) in
#' [skim_with()], to create the summary tibble for each type.
#'
#' When creating your own set of skimming functions, call [sfl()] within a
#' [get_skimmers()] method for your particular type. Your call to [sfl()] should
#' also provide a matching class in the `skim_type` argument.  Otherwise, it
#' will not be possible to dynamically reassign your default functions when
#' working interactively.
#'
#' Call [get_default_skimmers()] to see the functions for each type of summary
#' function currently supported. Call [get_default_skimmer_names()] to just see
#' the names of these functions. Use [modify_default_skimmers()] for a method
#' for changing the `skim_type` or functions for a default `sfl`. This is useful
#' for creating new default `sfl`'s.
#'
#' @param column An atomic vector or list. A column from a data frame.
#' @param skim_type A character scalar. The class of the object with default
#'   skimmers.
#' @return A `skim_function_list` object.
#' @seealso [sfl()]
#' @examples
#' # Defining default skimming functions for a new class, `my_class`.
#' # Note that the class argument is required for dynamic reassignment.
#' get_skimmers.my_class <- function(column) {
#'   sfl(
#'     skim_type = "my_class",
#'     mean,
#'     sd
#'   )
#' }
#'
#' # Integer and double columns are both "numeric" and are treated the same
#' # by default. To switch this behavior in another package, add a method.
#' get_skimmers.integer <- function(column) {
#'   sfl(
#'     skim_type = "integer",
#'     p50 = ~ stats::quantile(
#'       .,
#'       probs = .50, na.rm = TRUE, names = FALSE, type = 1
#'     )
#'   )
#' }
#' x <- mtcars[c("gear", "carb")]
#' class(x$carb) <- "integer"
#' skim(x)
#' \dontrun{
#' # In a package, to revert to the V1 behavior of skimming separately with the
#' # same functions, assign the numeric `get_skimmers`.
#' get_skimmers.integer <- skimr::get_skimmers.numeric
#'
#' # Or, in a local session, use `skim_with` to create a different `skim`.
#' new_skim <- skim_with(integer = skimr::get_skimmers.numeric())
#'
#' # To apply a set of skimmers from an old type to a new type
#' get_skimmers.new_type <- function(column) {
#'   modify_default_skimmers("old_type", new_skim_type = "new_type")
#' }
#' }
#' @export
get_skimmers <- function(column) {
  UseMethod("get_skimmers")
}

#' @describeIn get_skimmers The default method for skimming data. Only used when
#'   a column's data type doesn't match currently installed types. Call
#'   [get_default_skimmer_names] to see these defaults.
#' @export
get_skimmers.default <- function(column) {
  modify_default_skimmers("character", new_skim_type = "default")
}

#' @describeIn get_skimmers Summary functions for numeric columns, covering both
#'   [double()] and [integer()] classes: [mean()], [sd()], [quantile()] and
#'   [inline_hist()].
#' @export
get_skimmers.numeric <- function(column) {
  sfl(
    skim_type = "numeric",
    mean = ~ mean(., na.rm = TRUE),
    sd = ~ stats::sd(., na.rm = TRUE),
    p0 = ~ stats::quantile(., probs = 0, na.rm = TRUE, names = FALSE),
    p25 = ~ stats::quantile(., probs = .25, na.rm = TRUE, names = FALSE),
    p50 = ~ stats::quantile(., probs = .50, na.rm = TRUE, names = FALSE),
    p75 = ~ stats::quantile(., probs = .75, na.rm = TRUE, names = FALSE),
    p100 = ~ stats::quantile(., probs = 1, na.rm = TRUE, names = FALSE),
    hist = ~ inline_hist(., 5)
  )
}

#' @describeIn get_skimmers Summary functions for factor columns:
#'   [is.ordered()], [n_unique()] and [top_counts()].
#' @export
get_skimmers.factor <- function(column) {
  sfl(
    skim_type = "factor",
    ordered = is.ordered,
    n_unique = n_unique,
    top_counts = top_counts
  )
}

#' @describeIn get_skimmers Summary functions for character columns. Also, the
#'   default for unknown columns: [min_char()], [max_char()], [n_empty()],
#'   [n_unique()] and [n_whitespace()].
#' @export
get_skimmers.character <- function(column) {
  sfl(
    skim_type = "character",
    min = min_char,
    max = max_char,
    empty = n_empty,
    n_unique = n_unique,
    whitespace = n_whitespace
  )
}

#' @describeIn get_skimmers Summary functions for logical/ boolean columns:
#'   [mean()], which produces rates for each value, and [top_counts()].
#' @export
get_skimmers.logical <- function(column) {
  sfl(
    skim_type = "logical",
    mean = ~ mean(., na.rm = TRUE),
    count = top_counts
  )
}

#' @describeIn get_skimmers Summary functions for complex columns: [mean()].
#' @export
get_skimmers.complex <- function(column) {
  sfl(
    skim_type = "complex",
    mean = ~ mean(., na.rm = TRUE)
  )
}

#' @describeIn get_skimmers Summary functions for `Date` columns: [min()],
#'   [max()], [median()] and [n_unique()].
#' @export
get_skimmers.Date <- function(column) {
  sfl(
    skim_type = "Date",
    min = ~ min(., na.rm = TRUE),
    max = ~ max(., na.rm = TRUE),
    median = ~ stats::median(., na.rm = TRUE),
    n_unique = n_unique
  )
}

#' @describeIn get_skimmers Summary functions for `POSIXct` columns: [min()],
#'   [max()], [median()] and [n_unique()].
#' @export
get_skimmers.POSIXct <- function(column) {
  modify_default_skimmers("Date", new_skim_type = "POSIXct")
}

#' @describeIn get_skimmers Summary functions for `difftime` columns: [min()],
#'   [max()], [median()] and [n_unique()].
#' @export
get_skimmers.difftime <- function(column) {
  modify_default_skimmers("Date", new_skim_type = "difftime")
}

#' @describeIn get_skimmers Summary functions for `Timespan` columns: [min()],
#'   [max()], [median()] and [n_unique()].
#' @export
get_skimmers.Timespan <- function(column) {
  modify_default_skimmers("difftime", new_skim_type = "Timespan")
}

#' @describeIn get_skimmers Summary functions for `ts` columns: [min()],
#'   [max()], [median()] and [n_unique()].
#' @export
get_skimmers.ts <- function(column) {
  sfl(
    skim_type = "ts",
    start = ts_start,
    end = ts_end,
    frequency = stats::frequency,
    deltat = stats::deltat,
    mean = ~ mean(., na.rm = TRUE),
    sd = ~ stats::sd(., na.rm = TRUE),
    min = ~ min(., na.rm = TRUE),
    max = ~ max(., na.rm = TRUE),
    median = ~ stats::median(., na.rm = TRUE),
    line_graph = ~ inline_linegraph(., 16)
  )
}

#' @describeIn get_skimmers Summary functions for `list` columns: [n_unique()],
#'  [list_min_length()] and [list_max_length()].
#' @export
get_skimmers.list <- function(column) {
  sfl(
    skim_type = "list",
    n_unique = n_unique,
    min_length = list_min_length,
    max_length = list_max_length
  )
}

#' @describeIn get_skimmers Summary functions for `AsIs` columns: [n_unique()],
#'  [list_min_length()] and [list_max_length()].
#' @export
get_skimmers.AsIs <- function(column) {
  modify_default_skimmers("list", new_skim_type = "AsIs")
}

#' @describeIn  get_skimmers Summary functions for `haven_labelled` columns.
#'  Finds the appropriate skimmers for the underlying data in the vector.
#' @export
get_skimmers.haven_labelled <- function(column) {
  get_skimmers(vctrs::vec_data(column))
}

#' @rdname get_skimmers
#' @param new_skim_type The type to assign to the looked up set of skimmers.
#' @param new_funs Replacement functions for those in
#' @export
modify_default_skimmers <- function(skim_type,
                                    new_skim_type = NULL,
                                    new_funs = list()) {
  funs <- get_one_default_skimmer(skim_type)
  new_funs <- purrr::list_modify(funs, !!!new_funs)
  sfl(skim_type = new_skim_type %||% skim_type, !!!new_funs)
}

#' View default skimmer names and functions
#'
#' These utility functions look up the currently-available defaults for one or
#' more `skim_type`'s. They work with all defaults in the `skimr` package, as
#' well as the defaults in any package that extends `skimr`. See
#' [get_skimmers()] for writing lookup methods for different.
#'
#' The functions differ in return type and whether or not the result is in
#' a list. [get_default_skimmers()] and [get_one_default_skimmer()] return
#' functions. The former returns functions within a typed list, i.e.
#' `list(numeric = list(...functions...))`.
#'
#' The functions differ in return type and whether or not the result is in
#' a list. [get_default_skimmer_names()] and [get_one_default_skimmer_names()]
#' return a list of character vectors or a single character vector.
#'
#' [get_sfl()] returns the skimmer function list for a particular `skim_type`.
#' It differs from [get_default_skimmers()] in that the returned `sfl` contains
#' a list of functions and a `skim_type`.
#'
#' @param skim_type The class of the column being skimmed.
#' @export
get_default_skimmers <- function(skim_type = NULL) {
  if (is.null(skim_type)) {
    defaults <- as.character(utils::methods("get_skimmers"))
    types <- stringr::str_replace(defaults, "get_skimmers.", "")
    no_default <- purrr::discard(types, ~ .x == "default")
    iter <- rlang::set_names(no_default)
  } else {
    iter <- rlang::set_names(skim_type)
  }

  results <- purrr::map(iter, get_one_default_skimmer)
  no_defaults_for_class <- purrr::map_lgl(results, anyNA)
  results[!no_defaults_for_class]
}

#' @describeIn get_default_skimmers Get the functions associated with one
#'   `skim_type`.
#' @export
get_one_default_skimmer <- function(skim_type) {
  skimmers <- get_sfl(skim_type)
  skimmers$funs
}

#' @describeIn get_default_skimmers Get the names of the functions used in one
#'   or more `skim_type`'s.
#' @export
get_default_skimmer_names <- function(skim_type = NULL) {
  skimmers <- get_default_skimmers(skim_type)
  purrr::map(skimmers, names)
}

#' @describeIn get_default_skimmers Get the names of the functions used in one
#'   `skim_type`.
#' @export
get_one_default_skimmer_names <- function(skim_type) {
  skimmers <- get_one_default_skimmer(skim_type)
  names(skimmers)
}

#' @describeIn get_default_skimmers Get the `sfl` for a `skim_type`.
#' @export
get_sfl <- function(skim_type) {
  skimmers <- get_skimmers(structure(integer(), class = skim_type))
  if (skimmers$skim_type == "default") {
    warning("There are no default skimmers for type: ", skim_type)
    sfl(NA, skim_type = NA_character_)
  } else {
    skimmers
  }
}

test_that("get_sfl() behaves correctly", {
  my_sfl <- get_sfl("numeric")
  expect_s3_class(my_sfl, "skimr_function_list")
  expect_equal(my_sfl$skim_type, "numeric")
  expect_named(my_sfl$funs, c(
    "mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist"
  ))

  expect_warning(get_sfl("missing class"), "no default")
})

test_that("get_default_skimmer_names() has a correct list of defaults", {
  defaults <- get_default_skimmer_names()
  expect_setequal(
    names(defaults),
    c(
      "AsIs", "character", "complex", "Date", "difftime", "factor",
      "haven_labelled", "list", "logical", "numeric", "POSIXct", "Timespan",
      "ts"
    )
  )
  expect_identical(defaults$AsIs, c("n_unique", "min_length", "max_length"))
  expect_identical(
    defaults$character, c("min", "max", "empty", "n_unique", "whitespace")
  )
  expect_identical(defaults$complex, c("mean"))
  expect_identical(defaults$Date, c("min", "max", "median", "n_unique"))
  expect_identical(defaults$difftime, c("min", "max", "median", "n_unique"))
  expect_identical(defaults$Timespan, c("min", "max", "median", "n_unique"))
  expect_identical(defaults$factor, c("ordered", "n_unique", "top_counts"))
  expect_identical(defaults$list, c("n_unique", "min_length", "max_length"))
  expect_identical(defaults$logical, c("mean", "count"))
  expect_identical(
    defaults$numeric, c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist")
  )
  expect_identical(defaults$POSIXct, c("min", "max", "median", "n_unique"))
  expect_identical(
    defaults$ts,
    c(
      "start", "end", "frequency", "deltat", "mean", "sd", "min", "max",
      "median", "line_graph"
    )
  )
})

test_that("You can get the default skimmers for a particular class", {
  input <- get_default_skimmer_names("numeric")
  expect_named(input, "numeric")
  expect_identical(
    input$numeric, c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist")
  )
})

test_that("You can get the default skimmers for multiple classes", {
  input <- get_default_skimmer_names(c("list", "AsIs"))
  expect_named(input, c("list", "AsIs"))
  expect_identical(input$AsIs, c("n_unique", "min_length", "max_length"))
  expect_identical(input$list, c("n_unique", "min_length", "max_length"))
})

test_that("A warning is given for classes that don't have defaults", {
  expect_warning(get_default_skimmer_names("no_class"), "no default")
})

test_that("get_one_default_skimmer_names() behaves as expected", {
  expect_identical(
    get_one_default_skimmer_names("list"),
    c("n_unique", "min_length", "max_length")
  )
})

#' Summary statistic functions
#'
#' `skimr` provides extensions to a variety of functions with R's stats package
#' to simplify creating summaries of data. All functions are vectorized over the
#' first argument. Additional arguments should be set in the [sfl()] that sets
#' the appropriate skimmers for a data type. You can use these, along with other
#' vectorized R functions, for creating custom sets of summary functions for
#' a given data type.
#'
#' @seealso [get_skimmers()] for customizing the functions called by [skim()].
#' @param x A vector
#' @param n_bins In `inline_hist`, the number of histogram bars.
#' @param length.out In `inline_linegraph`, the length of the character time
#'   series.
#' @param max_char In `top` = 3, max_levels = 4
#' @name stats
NULL

#' @describeIn stats Calculate the sum of `NA` and `NULL` (i.e. missing) values.
#' @export
n_missing <- function(x) {
  sum(is.na(x) | is.null(x))
}

#' @describeIn stats Calculate the sum of not `NA` and `NULL` (i.e. missing)
#'   values.
#' @export
n_complete <- function(x) {
  sum(!is.na(x) & !is.null(x))
}

#' @describeIn stats Calculate complete values; complete values are not missing.
#' @export
complete_rate <- function(x) {
  1 - n_missing(x) / length(x)
}

#' @describeIn stats Calculate the number of rows containing only whitespace
#'   values using s+ regex.
#' @export
n_whitespace <- function(x) {
  whitespace <- grepl("^\\s+$", x)
  sum(whitespace)
}

#' @describeIn stats Create a contingency table and arrange its levels in
#'   descending order. In case of ties, the ordering of results is alphabetical
#'   and depends upon the locale. `NA` is treated as a ordinary value for
#'   sorting.
#' @export
sorted_count <- function(x) {
  tab <- table(x, useNA = "no")
  names_tab <- names(tab)
  if (is.element("", names_tab)) {
    names_tab[names_tab == ""] <- "empty"
    warning(
      "Variable contains value(s) of \"\" that have been ",
      "converted to \"empty\"."
    )
  }
  out <- rlang::set_names(as.integer(tab), names_tab)
  sort(out, decreasing = TRUE)
}

#' @describeIn stats Compute and collapse a contingency table into a single
#'   character scalar. Wraps [sorted_count()].
#' @param max_levels The maximum number of levels to be displayed.
#' @export
top_counts <- function(x, max_char = 3, max_levels = 4) {
  counts <- sorted_count(x)
  if (length(counts) > max_levels) {
    top <- counts[seq_len(max_levels)]
  } else {
    top <- counts
  }
  top_names <- substr(names(top), 1, max_char)
  paste0(top_names, ": ", top, collapse = ", ")
}

#' @describeIn stats Generate inline histogram for numeric variables. The
#'   character length of the histogram is controlled by the formatting options
#'   for character vectors.
#' @export
inline_hist <- function(x, n_bins = 8) {
  # For the purposes of the histogram, treat infinite as NA
  # (before the test for all NA)
  if (any(is.infinite(x))) {
    x[is.infinite(x)] <- NA
    warning(
      "Variable contains Inf or -Inf value(s) that were converted to NA."
    )
  }

  # Handle empty and NA vectors (is.na is TRUE for NaN)
  if (length(x) < 1 || all(is.na(x))) {
    return(" ")
  }

  # Addresses a known bug in cut()
  if (all(x == 0, na.rm = TRUE)) x <- x + 1
  hist_dt <- table(cut(x, n_bins))
  hist_dt <- hist_dt / max(hist_dt)
  spark_bar(hist_dt)
}

#' Draw a sparkline bar graph with unicode block characters
#'
#' Rendered using
#' [block elements](https://en.wikipedia.org/wiki/Block_Elements).
#' In most common fixed width fonts these are rendered wider than regular
#' characters which means they are not suitable if you need precise alignment.
#' Based on the function in the pillar package.
#'
#' @param x A numeric vector between 0 and 1
#' @param safe Nominally there are 8 block elements from 1/8 height to full
#'   height (8/8). However, the half-height and full-height blocks appear
#'   to be rendered inconsistently (possibly due to font substitution).
#' @examples
#' \dontrun{
#' x <- seq(0, 1, length = 6)
#' spark_bar(x)
#' spark_bar(sample(x))
#'
#' # This might work if you're lucky
#' spark_bar(seq(0, 1, length = 8), safe = FALSE)
#'
#' spark_bar(c(0, NA, 0.5, NA, 1))
#' }
#' @noRd
spark_bar <- function(x, safe = TRUE) {
  stopifnot(is.numeric(x))

  bars <- vapply(0x2581:0x2588, intToUtf8, character(1))
  if (safe) {
    bars <- bars[-c(4, 8)]
  }

  factor <- cut(
    x,
    breaks = seq(0, 1, length.out = length(bars) + 1),
    labels = bars,
    include.lowest = TRUE
  )
  chars <- as.character(factor)
  chars[is.na(chars)] <- bars[length(bars)]
  paste0(chars, collapse = "")
}

#' @describeIn stats Calculate the number of blank values in a character vector.
#'   A "blank" is equal to "".
#' @export
n_empty <- function(x) {
  empty_strings <- c("")
  sum(x %in% empty_strings)
}

#' @describeIn stats Calculate the minimum number of characters within a
#'   character vector.
#' @export
min_char <- function(x) {
  if (all(is.na(x))) {
    return(NA)
  }
  characters <- nchar(x, allowNA = TRUE)
  min(characters, na.rm = TRUE)
}

#' @describeIn stats Calculate the maximum number of characters within a
#'   character vector.
#' @export
max_char <- function(x) {
  if (all(is.na(x))) {
    return(NA)
  }
  characters <- nchar(x, allowNA = TRUE)
  max(characters, na.rm = TRUE)
}

#' @describeIn stats Calculate the number of unique elements but remove `NA`.
#' @export
n_unique <- function(x) {
  un <- x[!is.na(x)]
  un <- unique(un)
  length(un)
}

#' @describeIn stats Get the start for a time series without the frequency.
#' @export
ts_start <- function(x) {
  stats::start(x)[1]
}

#' @describeIn stats Get the finish for a time series without the frequency.
#' @export
ts_end <- function(x) {
  stats::end(x)[1]
}

#' @describeIn stats Generate inline line graph for time series variables. The
#'   character length of the line graph is controlled by the formatting options
#'   for character vectors.
#'   Based on the function in the pillar package.
#' @export
inline_linegraph <- function(x, length.out = 16) {
  t <- x[!is.na(x)]
  id <- seq(1, length(t), length.out = length.out)
  normalized <- normalize01(t[floor(id)])
  spark_line(normalized)
}

# Rescale data to be between 0 and 1
normalize01 <- function(x) {
  (x - min(x)) / (max(x) - min(x))
}

#' Draw a sparkline line graph with Braille characters.
#'
#' @inheritParams spark_bar
#' @examples
#' \dontrun{
#' x <- seq(0, 1, length = 10)
#' spark_line(x)
#' }
#' @noRd
spark_line <- function(x) {
  stopifnot(is.numeric(x))

  y <- findInterval(x, seq(0, 1, length.out = 5), all.inside = TRUE)

  ind <- matrix(y, ncol = 2, byrow = TRUE)
  ind[, 2] <- ind[, 2] + 4

  chars <- apply(ind, 1, braille)
  paste0(chars, collapse = "")
}

# https://en.wikipedia.org/wiki/Braille_Patterns
braille <- function(x) {
  # remap to braille sequence
  x <- c(7L, 3L, 2L, 1L, 8L, 6L, 5L, 4L)[x]

  raised <- 1:8 %in% x

  # offset in hex is 2800
  val <- 10240 + sum(raised * 2^(0:7))

  intToUtf8(val)
}

#' @describeIn stats Get the length of the shortest list in a vector of lists.
#' @export
list_lengths_min <- function(x) {
  x <- x[!is.na(x)]
  l <- lengths(x)
  if (length(l) > 0) {
    min(l)
  } else {
    NA
  }
}

#' @describeIn stats Get the median length of the lists.
#' @export
list_lengths_median <- function(x) {
  x <- x[!is.na(x)]
  l <- lengths(x)
  if (length(l) > 0) {
    stats::median(l)
  } else {
    NA
  }
}

#' @describeIn stats Get the maximum length of the lists.
#' @export
list_lengths_max <- function(x) {
  x <- x[!is.na(x)]
  l <- lengths(x)
  if (length(l) > 0) {
    max(l)
  } else {
    NA
  }
}

#' @describeIn stats Get the length of the shortest list in a vector of lists.
#' @export
list_min_length <- function(x) {
  l <- lengths(x)
  min(l)
}

#' @describeIn stats Get the length of the longest list in a vector of lists.
#' @export
list_max_length <- function(x) {
  l <- lengths(x)
  max(l)
}

test_that("inline_linegraph returns expected response for a ts vector", {
  input <- inline_linegraph(freeny$y)
  expect_identical(input, "⣀⣀⠤⠤⠒⠒⠉⠉")
})

test_that("inline_linegraph returns expected response for a long ts vector.", {
  input <- inline_linegraph(AirPassengers)
  expect_identical(input, "⣀⣀⣀⠔⠤⠊⠑⠊")
})

test_that("inline_linegraph returns expected response for an NA ts vector.", {
  input <- inline_linegraph(ts(c(NA, NA, NA, NA, NA, NA)))
  expect_identical(input, "⠀")
})

test_that("ts_start returns expected response for a ts vector", {
  input <- ts_start(freeny$y)
  expect_identical(input, 1962)
})

test_that("ts_start returns expected response for a ts vector", {
  input <- ts_end(freeny$y)
  expect_identical(input, 1971)
})

test_that("n_missing is calculated correctly.", {
  data <- c("a", "b", "c", NA, " ")
  input <- n_missing(data)
  expect_identical(input, 1L)
})

test_that("n_complete is calculated correctly.", {
  data <- c("a", "b", "c", NA)
  input <- n_complete(data)
  expect_identical(input, 3L)
})

test_that("complete_rate is calculated correctly.", {
  data <- c("a", "b", "c", NA, " ")
  input <- complete_rate(data)
  expect_equal(input, .8, tolerance = .001)
})

test_that("n_whitespace is calculated correctly.", {
  data <- c("a", "b", "c", NA, " ", "  a", "   ", "   a")
  input <- n_whitespace(data)
  expect_identical(input, 2L)
})

test_that("inline histogram is calculated correctly.", {
  input <- inline_hist(iris$Sepal.Length)
  expect_identical(input, "▂▇▅▇▆▅▂▂")
})

test_that("inline histogram is calculated correctly when x is all zeros.", {
  input <- inline_hist(numeric(10))
  expect_identical(input, "▁▁▁▇▁▁▁▁")
})

test_that("inline histogram returns an empty string when x is length 0.", {
  input <- inline_hist(numeric(0))
  expect_identical(input, " ")
})

test_that("inline hist is calculated correctly when x is all zeores or NAs", {
  input <- inline_hist(as.numeric(c(NA, NA, NA, 0, 0)))
  expect_identical(input, "▁▁▁▇▁▁▁▁")
})

test_that("inline histogram is calculated correctly when x is all 1s.", {
  input <- inline_hist(c(1, 1, 1, 1, 1, 1))
  expect_identical(input, "▁▁▁▇▁▁▁▁")
})

test_that("inline histogram returns empty string when x is all NAs.", {
  input <- inline_hist(as.numeric(rep(NA, 10)))
  expect_identical(input, " ")
})

test_that("inline histogram is returns empty string when x is all NaN.", {
  input <- inline_hist(rep(NaN, 10))
  expect_identical(input, " ")
})

test_that("inline histogram is correct when x is evenly distributed.", {
  input <- inline_hist(c(1, 2, 3, 4, 5, 6, 7, 8))
  expect_identical(input, "▇▇▇▇▇▇▇▇")
})

test_that("inline histogram is calculated correctly with NaN.", {
  input <- inline_hist(c(1, 2, 3, 3, 6, 6, 6, 8, NaN))
  expect_identical(input, "▂▂▅▁▁▇▁▂")
})

test_that("inline histogram is calculated correctly with NA.", {
  input <- inline_hist(c(1, 2, 3, 3, 6, 6, 6, 8, NA))
  expect_identical(input, "▂▂▅▁▁▇▁▂")
})

test_that("inline histogram is calculated correctly with Inf.", {
  expect_warning(input <-
    inline_hist(c(1, 2, 3, 3, 6, 6, 6, 8, Inf, -Inf)))
  expect_identical(input, "▂▂▅▁▁▇▁▂")
})

test_that("n_empty is calculated correctly.", {
  data <- c("a", "ab", "abc", "")
  correct <- as.integer(1)
  input <- n_empty(data)
  expect_identical(input, correct)
})

test_that("min_char is calculated correctly, including empty strings.", {
  data <- c("a", "ab", "abc", "")
  correct <- as.integer(0)
  input <- min_char(data)
  expect_identical(input, correct)
})

test_that("min_char with a multibyte character does not throw an error.", {
  data <- c("a", "ab", "abc", "Coleophora asteris M\x9fhl.")
  expect_error(min_char(data), NA)
})

test_that("min_char returns NA when there are only NA values.", {
  data <- as.character(c(NA, NA, NA, NA))
  input <- min_char(data)
  expect_equal(input, NA)
})

test_that("max_char is calculated correctly, including empty strings.", {
  data <- c("a", "ab", "abc", "")
  correct <- as.integer(3)
  input <- max_char(data)
  expect_identical(input, correct)
})

test_that("max_char with a multibyte character does not throw an error.", {
  data <- c("a", "ab", "abc", "Coleophora asteris M\x9fhl.")
  # correct <- as.integer(3)
  # input <- max_char(data)
  # expect_identical(input, correct)
  expect_error(max_char(data, NA))
})

test_that("max_char returns NA when there are only NA values.", {
  data <- as.character(c(NA, NA, NA, NA))
  input <- max_char(data)
  expect_equal(input, NA)
})

test_that("n_unique is calculated correctly.", {
  correct <- 4L
  data <- c("a", "ab", "abc", "")
  input <- n_unique(data)
  expect_identical(input, correct)
})

test_that("n_unique handles NA as expected.", {
  correct <- 4L
  data <- c("a", "ab", "abc", "", NA)
  input <- n_unique(data)
  expect_identical(input, correct)
})

test_that("list_lengths_min is calculated correctly.", {
  dat <- list(
    list("a", "b", "c"),
    list("d", "b", "d"),
    list("e", "f", "g"),
    d = list("h"),
    e = list("i", "j", "k", "l"),
    f = NA
  )
  expect_identical(list_lengths_min(dat), 1L)
})

test_that("list_lengths_max is calculated correctly.", {
  dat <- list(
    list("a", "b", "c"),
    list("d", "b", "d"),
    list("e", "f", "g"),
    d = list("h"),
    e = list("i", "j", "k", "l"),
    f = NA
  )
  expect_identical(list_lengths_max(dat), 4L)
})

test_that("list_lengths_median is calculated correctly.", {
  dat <- list(
    list("a", "b", "c"),
    list("d", "b", "d"),
    list("e", "f", "g"),
    d = list("h"),
    e = list("i", "j", "k", "l"),
    f = NA
  )
  expect_identical(list_lengths_median(dat), 3L)
})

test_that("list_min_length is calculated correctly.", {
  dat <- I(list(5, 5:6, 5:10, NA))
  expect_identical(list_min_length(dat), 1L)
})

test_that("list_max_length is calculated correctly.", {
  dat <- I(list(5, 5:6, 5:10, NA))
  expect_identical(list_max_length(dat), 6L)
})

test_that("sorted count is calculated correctly.", {
  dat <- c("A", "A", "A", "B", "C", "C")
  expect_equal(sorted_count(dat)[1:3], c("A" = 3, "C" = 2, "B" = 1))
  expect_equal(names(sorted_count(dat)), c("A", "C", "B"))
})

test_that("top counts is calculated correctly", {
  dat <- c("A", "A", "A", "B", "C", "C")
  expect_equal(top_counts(dat), "A: 3, C: 2, B: 1")
})

test_that("sorted count is calculated correctly with a NA.", {
  # NA should be sorted as if it is a regular value
  dat <- c("A", "A", "A", "A", "B", NA, NA, "C", "C", "C")
  expect_equal(unname(sorted_count(dat)), c(4, 3, 1))
  expect_equal(names(sorted_count(dat)), c("A", "C", "B"))
})

test_that("top counts is calculated correctly with a NA", {
  dat <- c("A", "A", "A", "B", "C", "C")
  expect_equal(top_counts(dat), "A: 3, C: 2, B: 1")
})

test_that("sorted count is calculated correctly with \"\".", {
  # \"\" should be converted to \"empty \" and a warning issued.
  dat <- c("A", "A", "A", "A", "B", "", "", "C", "C", "C")
  dat <- as.factor(dat)
  expect_warning(expected <- sorted_count(dat))
  expect_equal(unname(expected), c(4, 3, 2, 1))
  expect_equal(names(expected), c("A", "C", "empty", "B"))
})

#' Set or add the summary functions for a particular type of data
#'
#' While skim is designed around having an opinionated set of defaults, you
#' can use this function to change the summary statistics that it returns.
#'
#' `skim_with()` is a closure: a function that returns a new function. This
#' lets you have several skimming functions in a single R session, but it
#' also means that you need to assign the return of `skim_with()` before
#' you can use it.
#'
#' You assign values within `skim_with` by using the [sfl()] helper (`skimr`
#' function list). This helper behaves mostly like [dplyr::funs()], but lets
#' you also identify which skimming functions you want to remove, by setting
#' them to `NULL`. Assign an `sfl` to each column type that you wish to modify.
#'
#' Functions that summarize all data types, and always return the same type
#' of value, can be assigned to the `base` argument. The default base skimmers
#' compute the number of missing values [n_missing()] and the rate of values
#' being complete, i.e. not missing, [complete_rate()].
#'
#' When `append = TRUE` and local skimmers have names matching the names of
#' entries in the default `skim_function_list`, the values in the default list
#' are overwritten. Similarly, if `NULL` values are passed within `sfl()`, these
#' default skimmers are dropped. Otherwise, if `append = FALSE`, only the
#' locally-provided skimming functions are used.
#'
#' Note that `append` only applies to the `typed` skimmers (i.e. non-base).
#' See [get_default_skimmer_names()] for a list of defaults.
#'
#' @param ... One or more (`sfl`) `skimmer_function_list` objects, with an
#'   argument name that matches a particular data type.
#' @param base An `sfl` that sets skimmers for all column types.
#' @param append Whether the provided options should be in addition to the
#'   defaults already in `skim`. Default is `TRUE`.
#' @return A new `skim()` function. This is callable. See [skim()] for more
#'   details.
#' @examples
#' # Use new functions for numeric functions. If you don't provide a name,
#' # one will be automatically generated.
#' my_skim <- skim_with(numeric = sfl(median, mad), append = FALSE)
#' my_skim(faithful)
#'
#' # If you want to remove a particular skimmer, set it to NULL
#' # This removes the inline histogram
#' my_skim <- skim_with(numeric = sfl(hist = NULL))
#' my_skim(faithful)
#'
#' # This works with multiple skimmers. Just match names to overwrite
#' my_skim <- skim_with(numeric = sfl(iqr = IQR, p25 = NULL, p75 = NULL))
#' my_skim(faithful)
#'
#' # Alternatively, set `append = FALSE` to replace the skimmers of a type.
#' my_skim <- skim_with(numeric = sfl(mean = mean, sd = sd), append = FALSE)
#'
#' # Skimmers are unary functions. Partially apply arguments during assigment.
#' # For example, you might want to remove NA values.
#' my_skim <- skim_with(numeric = sfl(iqr = ~ IQR(., na.rm = TRUE)))
#'
#' # Set multiple types of skimmers simultaneously.
#' my_skim <- skim_with(numeric = sfl(mean), character = sfl(length))
#'
#' # Or pass the same as a list, unquoting the input.
#' my_skimmers <- list(numeric = sfl(mean), character = sfl(length))
#' my_skim <- skim_with(!!!my_skimmers)
#'
#' # Use the v1 base skimmers instead.
#' my_skim <- skim_with(base = sfl(
#'   missing = n_missing,
#'   complete = n_complete,
#'   n = length
#' ))
#'
#' # Remove the base skimmers entirely
#' my_skim <- skim_with(base = NULL)
#' @export
skim_with <- function(...,
                      base = sfl(
                        n_missing = n_missing,
                        complete_rate = complete_rate
                      ),
                      append = TRUE) {
  stopifnot(is.null(base) || inherits(base, "skimr_function_list"))
  local_skimmers <- validate_assignment(...)

  function(data, ..., .data_name = NULL) {
    if (is.null(.data_name)) {
      .data_name <- rlang::expr_label(substitute(data))
    }

    if (!inherits(data, "data.frame")) {
      data <- as.data.frame(data)
    }
    stopifnot(inherits(data, "data.frame"))

    selected <- names(tidyselect::eval_select(rlang::expr(c(...)), data))
    if (length(selected) == 0) {
      selected <- names(data)
    }

    grps <- dplyr::groups(data)
    if (length(grps) > 0) {
      group_variables <- selected %in% as.character(grps)
      selected <- selected[!group_variables]
    } else {
      attr(data, "groups") <- list()
    }


    skimmers <- purrr::map(
      selected, get_final_skimmers, data, local_skimmers, append
    )
    types <- purrr::map_chr(skimmers, "skim_type")
    unique_skimmers <- reduce_skimmers(skimmers, types)
    combined_skimmers <- purrr::map(unique_skimmers, join_with_base, base)
    ready_to_skim <- tibble::tibble(
      skim_type = unique(types),
      skimmers = purrr::map(combined_skimmers, mangle_names, names(base$funs)),
      skim_variable = split(selected, types)[unique(types)]
    )
    grouped <- dplyr::group_by(ready_to_skim, .data$skim_type)
    nested <- dplyr::summarize(
      grouped,
      skimmed = purrr::map2(
        .data$skimmers, .data$skim_variable, skim_by_type, data
      )
    )
    structure(
      tidyr::unnest(nested, "skimmed"),
      class = c("skim_df", "tbl_df", "tbl", "data.frame"),
      data_rows = nrow(data),
      data_cols = ncol(data),
      df_name = .data_name,
      dt_key  = get_dt_key(data),
      groups = dplyr::group_vars(data),
      base_skimmers = names(base$funs),
      skimmers_used = get_skimmers_used(unique_skimmers)
    )
  }
}

#' Process arguments provided in `skim_with`
#'
#' Make sure that arguments provided in `skim_with()` have names. Also,
#' check if we are defining a new skimming type dynamically.
#'
#' @keywords internal
#' @noRd
validate_assignment <- function(...) {
  to_assign <- rlang::list2(...)

  if (length(to_assign) < 1) {
    return(to_assign)
  }

  # Need to cope with case where ... is a list already
  if (!inherits(to_assign[[1]], "skimr_function_list")) {
    to_assign <- to_assign[[1]]
  }

  proposed_names <- names(to_assign)
  if (!all(nzchar(proposed_names)) || is.null(proposed_names) ||
    anyNA(proposed_names)) {
    stop("skim_with requires all arguments to be named.", call. = FALSE)
  }

  defaults <- get_default_skimmers()
  existing <- proposed_names %in% names(defaults)

  if (!all(existing) & length(defaults) > 0) {
    collapsed <- paste(proposed_names[!existing], collapse = ", ")
    message(
      "Creating new skimming functions for the following classes: ",
      collapsed, ".\nThey did not have recognized defaults. Call ",
      "get_default_skimmers() for more information."
    )
  }
  to_assign
}

#' Sets the appropriate key value when working with `data.table`
#' @keywords internal
#' @noRd
get_dt_key <- function(data) {
  if (inherits(data, "data.table")) {
    dt_key <- data.table::key(data)
    if (is.null(dt_key)) {
      dt_key <- "NULL"
    }
    paste(dt_key, collapse = ", ")
  } else {
    NA # Will never be NA if `data` is a data.table
  }
}

#' Combine local and default skimmers for each column
#'
#' Get the default skimmers for the current column using S3 dispatch for
#' [get_skimmers()]. Get the user-provided local skimmers from [skim_with()].
#' If no local skimmers are provided, use the defaults. Otherwise, merge the
#' local and default skimmers with the following rules.
#'
#'   - If `append = FALSE` of if the local and default types differ, use only
#'     the locals.
#'   - Else, replace the default values with the local values.
#'
#' @param column A character scalar. The column name.
#' @param data The data frame to summarize.
#' @param local_skimmers A list of `sfl` objects. Skimmers defined using
#'   `skim_with()`
#' @param append Same as above.
#' @param base Same as above.
#' @noRd
get_final_skimmers <- function(column, data, local_skimmers, append) {
  defaults <- get_skimmers(data[[column]])
  all_classes <- skim_class(data[[column]])
  locals <- get_local_skimmers(all_classes, local_skimmers)

  if (!nzchar(defaults$skim_type)) {
    msg <- sprintf(
      "Default skimming functions for column [%s] with class [%s]",
      column, paste(all_classes, collapse = ", ")
    )
    stop(msg, " did not have value for its `.class` argument. Please ",
      "investigate the definition of the associated S3 method.",
      call. = FALSE
    )
  }

  if (is.null(locals$funs)) {
    if (defaults$skim_type == "default") {
      msg <- sprintf(
        "Couldn't find skimmers for class: %s;",
        paste(all_classes, collapse = ", ")
      )
      warning(msg,
        " No user-defined `sfl` provided. Falling back to `character`.",
        call. = FALSE
      )
      data[[column]] <- as.character(data[[column]])
      skimmers <- defaults
      skimmers$skim_type <- "character"
    } else {
      skimmers <- defaults
    }
  } else {
    skimmers <- merge_skimmers(locals, defaults, append)
  }
  skimmers
}

skim_class <- function(column) {
  base_class <- class(column)
  if (any(base_class %in% c("double", "integer"))) {
    c(base_class, "numeric")
  } else {
    base_class
  }
}

get_local_skimmers <- function(classes, local_skimmers) {
  local_classes <- intersect(classes, names(local_skimmers))
  if (length(local_classes) == 0) {
    return(NULL)
  }
  
  first_class <- local_classes[[1]]
  
  out <- local_skimmers[[first_class]]
  out$skim_type <- first_class
  out
}

merge_skimmers <- function(locals, defaults, append) {
  if (!append || locals$skim_type != defaults$skim_type) {
    locals
  } else {
    defaults$funs <- purrr::compact(purrr::list_modify(defaults$funs, !!!locals$funs))
    defaults
  }
}

reduce_skimmers <- function(skimmers, types) {
  named <- rlang::set_names(skimmers, types)
  named[unique(types)]
}

join_with_base <- function(skimmers, base) {
  skimmers$funs <- c(base$funs, skimmers$funs)
  skimmers
}

get_skimmers_used <- function(skimmers) {
  types <- names(skimmers)
  function_names <- purrr::map(skimmers, ~ names(.x$funs))
  rlang::set_names(function_names, types)
}

NAME_DELIMETER <- "~!@#$%^&*()-+"
mangle_names <- function(skimmers, base_names) {
  fun_names <- names(skimmers$funs)
  prefixes <- ifelse(
    fun_names %in% base_names,
    NAME_DELIMETER,
    paste0(NAME_DELIMETER, skimmers$skim_type, ".")
  )
  mangled <- paste0(prefixes, fun_names)
  skimmers$funs <- rlang::set_names(skimmers$funs, mangled)
  skimmers
}


#' Generate one or more rows of a `skim_df`, using one column
#'
#' Call all of the skimming functions on the single column, using grouped
#' variants, if necessary.
#'
#' We expect one row per variable/ group. To do this we need to take the
#' processed results, find the appropriate columns for each variable and
#' restack them. This uses a small hack that rests on the naming convention
#' of data frame produced by `summarize_at`, which uses the following scheme:
#'
#'  - `variable_name` + `_` + `function_name`
#'
#' To avoid inappropriately assigning the columns to the wrong variable, we
#' mangle the function names. That way, each set of relevant columns begin
#' with the column name + `_` + our internal delimiter.
#'
#' @param mangled_skimmers The `sfl`'s whose function names have been mangled.
#' @param variable_names The names of columns in the original data, matching a
#'   data type, that will be summarized.
#' @param data The original data.
#' @keywords internal
#' @noRd
skim_by_type <- function(mangled_skimmers, variable_names, data) {
  UseMethod("skim_by_type", data)
}

#' @export
skim_by_type.grouped_df <- function(mangled_skimmers, variable_names, data) {
  group_columns <- dplyr::groups(data)
  grouped <- dplyr::group_by(data, !!!group_columns)
  skimmed <- dplyr::summarize(
    grouped,
    dplyr::across(tidyselect::any_of(variable_names), mangled_skimmers$funs)
  )
  build_results(skimmed, variable_names, group_columns)
}

#' @export
skim_by_type.data.frame <- function(mangled_skimmers, variable_names, data) {
  skimmed <- dplyr::summarize(
    data,
    dplyr::across(tidyselect::any_of(variable_names), mangled_skimmers$funs)
  )
  build_results(skimmed, variable_names, NULL)
}

#' @export
skim_by_type.data.table <- function(mangled_skimmers, variable_names, data) {
  data <- tibble::as_tibble(data)
  skimmed <- dplyr::summarize(
    data,
    dplyr::across(tidyselect::any_of(variable_names), mangled_skimmers$funs)
  )
  build_results(skimmed, variable_names, NULL)
}

#' Summarize returns a single row data frame, make it tall.
#' @noRd
build_results <- function(skimmed, variable_names, groups) {
  if (length(variable_names) > 1) {
    out <- tibble::tibble(
      skim_variable = variable_names,
      by_variable = purrr::map(variable_names, reshape_skimmed, skimmed, groups)
    )
    tidyr::unnest(out, "by_variable")
  } else {
    out <- dplyr::select(
      as.data.frame(skimmed),
      !!!groups,
      tidyselect::contains(NAME_DELIMETER)
    )
    tibble::tibble(
      skim_variable = variable_names,
      !!!set_clean_names(out)
    )
  }
}

reshape_skimmed <- function(column, skimmed, groups) {
  delim_name <- paste0(column, "_", NAME_DELIMETER)
  out <- dplyr::select(
    tibble::as_tibble(skimmed),
    !!!groups,
    tidyselect::starts_with(delim_name, ignore.case = FALSE)
  )
  set_clean_names(out)
}

set_clean_names <- function(out) {
  separated <- strsplit(names(out), NAME_DELIMETER, fixed = TRUE)
  clean_names <- purrr::map_chr(separated, ~ .x[length(.x)])
  rlang::set_names(out, clean_names)
}

test_that("skim of a simple data.table produces no warnings", {
  skip_if_not_installed("data.table")
  withr::local_options(list(width = 91))
  DT_letters <- data.table::data.table(abc = letters)
  expect_no_warning(skim(DT_letters))
})

test_that("skim of a simple data.table produces no warnings even with dtplyr", {
  skip_if_not_installed("data.table")
  skip_if_not_installed("dtplyr")
  withr::local_options(list(width = 91))
  DT_letters <- data.table::data.table(abc = letters)
  expect_no_warning(skim(DT_letters))
})

test_that("skim of a simple data.table produces output as expected", {
  skip_if_not_installed("data.table")
  DT_letters <- data.table::data.table(abc = letters)

  skimmed_DT_letters <- skim(DT_letters)
  withr::local_options(list(cli.unicode = FALSE, width = 91))
  expect_snapshot(skimmed_DT_letters)
})


test_that("skim of data.table produces output as expected", {
  skip_if_not_installed("data.table")
  set.seed(1L)

  DT_factors <- data.table::data.table(
    abc = letters,
    grps = factor(sample(c("AA", "BB"), 26, TRUE)),
    values = rnorm(26)
  )

  withr::local_options(list(cli.unicode = FALSE, width = 91))
  expect_snapshot(skim(DT_factors))

  data.table::setkeyv(DT_factors, c("abc", "grps"))
  expect_snapshot(skim(DT_factors))

  DF_factors <- as.data.frame(DT_factors)
  expect_snapshot(skim(DF_factors))

  tibble_factors <- tibble::as_tibble(DT_factors)
  expect_snapshot(skim(tibble_factors))
})

test_that("You can bind skim_df rows", {
  skimmed1 <- skim(iris)
  skimmed2 <- skim(mtcars)
  combined <- vctrs::vec_rbind(skimmed1, skimmed2)
  expect_s3_class(combined, "skim_df")

  attrs <- attributes(combined)
  expect_equal(attrs$data_rows, 182)
  expect_equal(attrs$data_cols, 16)
  expect_equal(attrs$df_name, "`iris`+`mtcars`")
  expect_equal(
    attrs$skimmers_used,
    list(
      numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist"),
      factor = c("ordered", "n_unique", "top_counts")
    )
  )
})

test_that("When binding columns, fall back to tbl_df", {
  skimmed <- skim(iris)
  combined <- vctrs::vec_cbind(skimmed, skimmed, .name_repair = "universal_quiet")
  expect_s3_class(combined, "tbl")
  expect_false("skim_df" %in% class(combined))
})

test_that("You can bind skim_df rows", {
  skimmed1 <- skim(iris)
  skimmed2 <- skim(mtcars)
  combined <- vctrs::vec_rbind(skimmed1, skimmed2)
  expect_s3_class(combined, "skim_df")

  attrs <- attributes(combined)
  expect_equal(attrs$data_rows, 182)
  expect_equal(attrs$data_cols, 16)
  expect_equal(attrs$df_name, "`iris`+`mtcars`")
  expect_equal(
    attrs$skimmers_used,
    list(
      numeric = c("mean", "sd", "p0", "p25", "p50", "p75", "p100", "hist"),
      factor = c("ordered", "n_unique", "top_counts")
    )
  )
})

test_that("When binding columns, fall back to tbl_df", {
  skimmed <- skim(iris)
  combined <- vctrs::vec_cbind(skimmed, skimmed)
  expect_s3_class(combined, "tbl")
  expect_false("skim_df" %in% class(combined))
})

---
title: "Extending skimr"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Extending skimr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

The `skim()` function summarizes data types contained within data frames and
objects that have `as.data.frame()` methods to coerce them into data frames. It
comes with a set of default summary functions for a wide variety of data types,
but this is not comprehensive. 

Package authors (and advanced users) can add support for skimming their specific
non-data-frame objects in their packages, and they can provide different
defaults in their own summary functions. This will require including skimr as a
dependency.

# Skimming objects that are not coercible to data frames

This example will illustrate this by creating support for the `lm` object
produced by `lm()`. For any object this involves two required elements and one
optional element.  This is a simple example, but for other types of objects
there may be much more complexity

If you are adding skim support to a package you will also need to add `skimr` to
the list of imports. 

```{r}
library(skimr)
```

The `lm()` function produces a complex object with class "lm".

```{r}
results <- lm(weight ~ feed, data = chickwts)
class(results)
attributes(results)
```

There is no as.data.frame method for an `lm` object.

```{r, eval = FALSE}
as.data.frame(results)
#> Error in as.data.frame.default(results) :
#>  cannot coerce class ‘"lm"’ to a data.frame
```


Unlike the example of having a new type of data in a column of a simple data
frame (for which we would create a `sfl`) frame in the "Using skimr" vignette,
this is a different type of challenge: an object that we might wish to skim, but
that cannot be directly skimmed. Therefore we need to make it into an object
that is either a data frame or coercible to a data frame.

In the case of the lm object, the `model` attribute is already a data frame. So
a very simple way to solve the challenge is to skim `results$model` directly.

```{r}
skim(results$model)
```

This is works, but we could go one step further and create a new function for
doing this directly.

```{r}
skim_lm <- function(.data) {
  .data <- .data$model
  skimr::skim(.data)
}

lm(weight ~ feed, data = chickwts) %>% skim_lm()
```

If desired, a more complex function can be created.  For example, the lm object
also contains fitted values and residuals.  We could incorporate these in the
data frame.

```{r}
skim_lm <- function(.data, fit = FALSE) {
  .data <- .data$model
  if (fit) {
    .data <- .data %>%
      dplyr::bind_cols(
        fitted = data.frame(results$fitted.values),
        residuals = data.frame(results$residuals)
      )
  }
  skimr::skim(.data)
}
```


```{r}
skim_lm(results, fit = TRUE)
```

A second example of the need for a special function is with `dist` objects. The
`UScitiesD` data set is an example of this.

```{r}
class(UScitiesD)
UScitiesD
```

A `dist` object is most often, as in this case, lower triange matrices of
distances, which can be measured in various ways.  There are many packages that
produce dist objects and/or take dist objects as inputs, including those for
cluster analysis and multidimensional scaling. 

A simple solution to this is to follow a similar design to that for `lm`
objects.

```{r}
skim_dist <- function(.data) {
  .data <- data.frame(as.matrix(.data))
  skimr::skim(.data)
}
```

However, this has the limitation of treating the dist data as though it is
simple numeric data.

What we might want to do, instead, is to create a new class, for example,
"distance" that is specifically for distance data. This will allow it to have
its own `sfl` or skimr function list.

As handling gets more complex, rather than make a new function it can be more
powerful to define an `as.data.frame` S3 method for dist objects, which will
allow it to integrate with skimr more completely and uses to use the `skim()`
function directly. In a package you will want to export this.

```{r}
as.data.frame.dist <- function(.data) {
  .data <- data.frame(as.matrix(.data))

  .data[] <- lapply(.data, structure, class = "distance", nms = names(.data))
  .data
}

as.data.frame(UScitiesD)
```

However, until an `sfl` is created, `skimr` will not recognize the class and
fall back to treating the data as if it were character data.

```{r}
skim(UScitiesD)
```

The solution to this is to define an `sfl` (skimr function list) specifically
for the `distance` class.

## Defining sfl's for a package

`skimr` has an opinionated list of functions for each class (e.g. numeric,
factor)  of data. The core package supports many commonly used classes, but
there are many others. You can investigate these defaults by calling
`get_default_skimmer_names()`.

What if your data type, like `distance`, isn't covered by defaults? `skimr`
usually falls back to treating the type as a character, which isn't necessarily
helpful. In this case, you're best off adding your data type with `skim_with()`.

Before we begin, we'll be using the following custom summary statistics
throughout. These functions find the nearest and furthest other location for
each location. 

One thing that is important to be aware of when creating statistics functions
for skimr is that skimr largely uses tibbles rather than base data frames. This
means that many base operations do not work as expected.

```{r}
get_nearest <- function(column) {
  closest <- which.min(column[column != 0])
  cities <- attr(column, "nms")[column != 0]
  toString(cities[closest])
}

get_furthest <- function(column) {
  furthest <- which.max(column[column != 0])
  cities <- attr(column, "nms")[column != 0]
  toString(cities[furthest])
}
```

This function, like all summary functions used by `skimr` has two notable
features.

*  It accepts a vector as its single argument
*  it returns a scalar, or in R terminology, a vector of length 1.

There are a lot of functions that fulfill these criteria:

*  Existing functions from base, stats, or other packages,
*  lambda's created using the Tidyverse-style syntax
*  custom functions that have been defined in the `skimr` package
*  custom functions that you have defined.

Not fulfilling the two criteria can lead to some very confusing behavior within
`skimr`. Beware! An example of this issue is the base `quantile()` function in
default `skimr` percentiles are returned by using `quantile()` five 
times. In the case of these functions, there could be ties which would result in
returning vectors that have length greater than 1. This is handled by collapsing
all of the tied values into a single string. 

Notice, also, that in the case of distance data we may wish to exclude distances
of 0, which indicate the distance from a place to itself. In finding the minimum
our function looks only at the distance to other places.

There are at least two ways that you might want to customize skimr handling of a
special data type within a package or your own work.  The first is to create a
custom skimming function.

````{r}
skim_with_dist <- skim_with(
  distance = sfl(
    nearest = get_nearest,
    furthest = get_furthest
  )
)

skim_with_dist(UScitiesD)
````

The example above creates a new *function*, and you can call that function on
a specific column with `distance` data to get the appropriate summary 
statistics. The `skim_with` factory also uses the default skimrs for things 
like factors, characters, and numerics. Therefore our `skim_with_dist` is like
the regular `skim` function with the added ability to summarize `distance`
columns.

While this works for any data type and you can also include it within any 
package (assuming your users load skimr), there is a second, even better,
approach. To take full advantage of `skimr`, we'll dig a bit into its API.

## Adding new methods

`skimr` has a lookup mechanism, based on the function `get_skimmers()`, to
find default summary functions for each class. This is based on the S3 class
system. You can learn more about it in
[*Advanced R*](https://adv-r.hadley.nz/s3.html).

This requires that you add `skimr` to your list of dependencies.

To export a new set of defaults for a data type, create a method for the generic
function `get_skimmers`. Each of those methods returns an `sfl` (skimr
function list) This is the same list-like data structure used in the
`skim_with()` example above. But note! There is one key difference. When adding
a generic we also want to identify the `skim_type` in the `sfl`. You will
probably want to use `skimr::get_skimmers.distance()` but that will not work in
a vignette.

In a package you will want to export this.

````{r}
#' @importFrom skimr get_skimmers
#' @export
get_skimmers.distance <- function(column) {
  sfl(
    skim_type = "distance",
    nearest = get_nearest,
    furthest = get_furthest
  )
}
````

The same strategy follows for other data types.

*  Create a method
*  return an `sfl`
*  make sure that the `skim_type` is included.

Users of your package should load `skimr` to get the `skim()` function (although
you could import and reexport it). Once loaded, a call to
`get_default_skimmer_names()` will return defaults for your data types as well! 

```{r}
get_default_skimmer_names()
```

They will then be able to use `skim()` directly.

```{r}
skim(UScitiesD)
```


## Conclusion

This is a very simple example. For some packages the custom statistics
will likely  be much more complex. The flexibility of `skimr` allows you to
manage that.

Thanks to Jakub Nowosad, Tiernan Martin, Edzer Pebesma, Michael Sumner, and 
Kyle Butts for inspiring and helping with the development of this code.

---
title: "Extending skimr"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Extending skimr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

The `skim()` function summarizes data types contained within data frames and
objects that have `as.data.frame()` methods to coerce them into data frames. It
comes with a set of default summary functions for a wide variety of data types,
but this is not comprehensive. 

Package authors (and advanced users) can add support for skimming their specific
non-data-frame objects in their packages, and they can provide different
defaults in their own summary functions. This will require including skimr as a
dependency.

# Skimming objects that are not coercible to data frames

This example will illustrate this by creating support for the `lm` object
produced by `lm()`. For any object this involves two required elements and one
optional element.  This is a simple example, but for other types of objects
there may be much more complexity

If you are adding skim support to a package you will also need to add `skimr` to
the list of imports. 

```{r}
library(skimr)
```

The `lm()` function produces a complex object with class "lm".

```{r}
results <- lm(weight ~ feed, data = chickwts)
class(results)
attributes(results)
```

There is no as.data.frame method for an `lm` object.

```{r, eval = FALSE}
as.data.frame(results)
#> Error in as.data.frame.default(results) :
#>  cannot coerce class ‘"lm"’ to a data.frame
```


Unlike the example of having a new type of data in a column of a simple data
frame (for which we would create a `sfl`) frame in the "Using skimr" vignette,
this is a different type of challenge: an object that we might wish to skim, but
that cannot be directly skimmed. Therefore we need to make it into an object
that is either a data frame or coercible to a data frame.

In the case of the lm object, the `model` attribute is already a data frame. So
a very simple way to solve the challenge is to skim `results$model` directly.

```{r}
skim(results$model)
```

This is works, but we could go one step further and create a new function for
doing this directly.

```{r}
skim_lm <- function(.data) {
  .data <- .data$model
  skimr::skim(.data)
}

lm(weight ~ feed, data = chickwts) |> skim_lm()
```

If desired, a more complex function can be created.  For example, the lm object
also contains fitted values and residuals.  We could incorporate these in the
data frame.

```{r}
skim_lm <- function(.data, fit = FALSE) {
  .data <- .data$model
  if (fit) {
    .data <- .data |>
      dplyr::bind_cols(
        fitted = data.frame(results$fitted.values),
        residuals = data.frame(results$residuals)
      )
  }
  skimr::skim(.data)
}
```


```{r}
skim_lm(results, fit = TRUE)
```

A second example of the need for a special function is with `dist` objects. The
`UScitiesD` data set is an example of this.

```{r}
class(UScitiesD)
UScitiesD
```

A `dist` object is most often, as in this case, lower triange matrices of
distances, which can be measured in various ways.  There are many packages that
produce dist objects and/or take dist objects as inputs, including those for
cluster analysis and multidimensional scaling. 

A simple solution to this is to follow a similar design to that for `lm`
objects.

```{r}
skim_dist <- function(.data) {
  .data <- data.frame(as.matrix(.data))
  skimr::skim(.data)
}
```

However, this has the limitation of treating the dist data as though it is
simple numeric data.

What we might want to do, instead, is to create a new class, for example,
"distance" that is specifically for distance data. This will allow it to have
its own `sfl` or skimr function list.

As handling gets more complex, rather than make a new function it can be more
powerful to define an `as.data.frame` S3 method for dist objects, which will
allow it to integrate with skimr more completely and uses to use the `skim()`
function directly. In a package you will want to export this.

```{r}
as.data.frame.dist <- function(.data) {
  .data <- data.frame(as.matrix(.data))

  .data[] <- lapply(.data, structure, class = "distance", nms = names(.data))
  .data
}

as.data.frame(UScitiesD)
```

However, until an `sfl` is created, `skimr` will not recognize the class and
fall back to treating the data as if it were character data.

```{r}
skim(UScitiesD)
```

The solution to this is to define an `sfl` (skimr function list) specifically
for the `distance` class.

## Defining sfl's for a package

`skimr` has an opinionated list of functions for each class (e.g. numeric,
factor)  of data. The core package supports many commonly used classes, but
there are many others. You can investigate these defaults by calling
`get_default_skimmer_names()`.

What if your data type, like `distance`, isn't covered by defaults? `skimr`
usually falls back to treating the type as a character, which isn't necessarily
helpful. In this case, you're best off adding your data type with `skim_with()`.

Before we begin, we'll be using the following custom summary statistics
throughout. These functions find the nearest and furthest other location for
each location. 

One thing that is important to be aware of when creating statistics functions
for skimr is that skimr largely uses tibbles rather than base data frames. This
means that many base operations do not work as expected.

```{r}
get_nearest <- function(column) {
  closest <- which.min(column[column != 0])
  cities <- attr(column, "nms")[column != 0]
  toString(cities[closest])
}

get_furthest <- function(column) {
  furthest <- which.max(column[column != 0])
  cities <- attr(column, "nms")[column != 0]
  toString(cities[furthest])
}
```

This function, like all summary functions used by `skimr` has two notable
features.

*  It accepts a vector as its single argument
*  it returns a scalar, or in R terminology, a vector of length 1.

There are a lot of functions that fulfill these criteria:

*  Existing functions from base, stats, or other packages,
*  lambda's created using the Tidyverse-style syntax
*  custom functions that have been defined in the `skimr` package
*  custom functions that you have defined.

Not fulfilling the two criteria can lead to some very confusing behavior within
`skimr`. Beware! An example of this issue is the base `quantile()` function in
default `skimr` percentiles are returned by using `quantile()` five 
times. In the case of these functions, there could be ties which would result in
returning vectors that have length greater than 1. This is handled by collapsing
all of the tied values into a single string. 

Notice, also, that in the case of distance data we may wish to exclude distances
of 0, which indicate the distance from a place to itself. In finding the minimum
our function looks only at the distance to other places.

There are at least two ways that you might want to customize skimr handling of a
special data type within a package or your own work.  The first is to create a
custom skimming function.

````{r}
skim_with_dist <- skim_with(
  distance = sfl(
    nearest = get_nearest,
    furthest = get_furthest
  )
)

skim_with_dist(UScitiesD)
````

The example above creates a new *function*, and you can call that function on
a specific column with `distance` data to get the appropriate summary 
statistics. The `skim_with` factory also uses the default skimrs for things 
like factors, characters, and numerics. Therefore our `skim_with_dist` is like
the regular `skim` function with the added ability to summarize `distance`
columns.

While this works for any data type and you can also include it within any 
package (assuming your users load skimr), there is a second, even better,
approach. To take full advantage of `skimr`, we'll dig a bit into its API.

## Adding new methods

`skimr` has a lookup mechanism, based on the function `get_skimmers()`, to
find default summary functions for each class. This is based on the S3 class
system. You can learn more about it in
[*Advanced R*](https://adv-r.hadley.nz/s3.html).

This requires that you add `skimr` to your list of dependencies.

To export a new set of defaults for a data type, create a method for the generic
function `get_skimmers`. Each of those methods returns an `sfl` (skimr
function list) This is the same list-like data structure used in the
`skim_with()` example above. But note! There is one key difference. When adding
a generic we also want to identify the `skim_type` in the `sfl`. You will
probably want to use `skimr::get_skimmers.distance()` but that will not work in
a vignette.

In a package you will want to export this.

````{r}
#' @importFrom skimr get_skimmers
#' @export
get_skimmers.distance <- function(column) {
  sfl(
    skim_type = "distance",
    nearest = get_nearest,
    furthest = get_furthest
  )
}
````

The same strategy follows for other data types.

*  Create a method
*  return an `sfl`
*  make sure that the `skim_type` is included.

Users of your package should load `skimr` to get the `skim()` function (although
you could import and reexport it). Once loaded, a call to
`get_default_skimmer_names()` will return defaults for your data types as well! 

```{r}
get_default_skimmer_names()
```

They will then be able to use `skim()` directly.

```{r}
skim(UScitiesD)
```


## Conclusion

This is a very simple example. For some packages the custom statistics
will likely  be much more complex. The flexibility of `skimr` allows you to
manage that.

Thanks to Jakub Nowosad, Tiernan Martin, Edzer Pebesma, Michael Sumner, and 
Kyle Butts for inspiring and helping with the development of this code.

[Theme song: *PSA* by
Jay-Z](https://www.youtube.com/watch?v=-LzdKH1naok)

We announced the testing version of `skimr` v2 on [June 19,
2018](https://github.com/ropensci/skimr/issues/341). After more than a
year of (admittedly intermittent) work, we’re thrilled to be able to say
that the package is ready to go to CRAN. So, what happened over the last
year? And why are we so excited for V2?

Setting the stage
-----------------

Before we can talk about the last year of `skimr` development, we need
to lay out the timeline that got us to this point. For those deeply
enmeshed in `skimr` lore, all
[dozens](https://imgur.com/gallery/R1fdEt3) of you, bear with.

`skimr` was originally an [rOpenSci
unconf17](https://ropensci.org/blog/2017/07/11/skimr/) project, a big
collaboration between eight different participants that resulted in a
conceptual outline of the package and a basic working version.
Participating in the unconf was a truly magical experience, with
everyone bringing a tremendous amount of energy and ideas to the
project, and implementation happening over a flurry of [“fancy git
commits”](https://twitter.com/AmeliaMN/status/867818976666976256).

About six months later, we released our first version on CRAN. The time
between these two milestones was mostly spent on fleshing out all of the
different ideas that were generated during the unconf (like handling
grouped data frames) and fixing all the bugs we discovered along the
way.

Getting the package on CRAN opened the gates for bug reports and feature
requests on [GitHub](https://github.com/ropensci/skimr/issues). About
the same time we pushed our first version to CRAN, Elin got `skimr`’s
rOpenSci’s package [peer
review](https://github.com/ropensci/software-review/issues/175) started
(thank you Jennifer and Jim!), opening another incredibly useful channel
for collecting feedback on the package. All of these new ideas and
suggestions gave us the opportunity to really push `skimr` to the next
level, but doing that would require rethinking the package, from the
ground up.

A month after finishing the peer review (and six months after the
process began), we announced v2. Over the first phase of `skimr`’s life,
we accumulated 700 commits, two release, 400 GitHub stars, 95 percent
code coverage and a lifetime’s worth of [unicode rendering
bugs](https://github.com/ropensci/skimr#support-for-spark-histograms)!

Just kidding! We love our little histograms, even when they don’t love
us back! For those of you that might have never seen `skimr`, using the
package typically boils down to a single function call:

    library(skimr)
    library(dplyr)
    options(width = 90)

    skim(iris)

    ## ── Data Summary ────────────────────────
    ##                            Values
    ## Name                       iris  
    ## Number of rows             150   
    ## Number of columns          5     
    ## _______________________          
    ## Column type frequency:           
    ##   factor                   1     
    ##   numeric                  4     
    ## ________________________         
    ## Group variables            None  
    ## 
    ## ── Variable type: factor ─────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate ordered n_unique top_counts               
    ## 1 Species               0             1 FALSE          3 set: 50, ver: 50, vir: 50
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate  mean    sd    p0   p25   p50   p75  p100 hist 
    ## 1 Sepal.Length          0             1  5.84 0.828   4.3   5.1  5.8    6.4   7.9 ▆▇▇▅▂
    ## 2 Sepal.Width           0             1  3.06 0.436   2     2.8  3      3.3   4.4 ▁▆▇▂▁
    ## 3 Petal.Length          0             1  3.76 1.77    1     1.6  4.35   5.1   6.9 ▇▁▆▇▂
    ## 4 Petal.Width           0             1  1.20 0.762   0.1   0.3  1.3    1.8   2.5 ▇▁▇▅▃

Getting it right
----------------

Under normal circumstances (i.e. not during a hackathon), most software
engineering projects begin with a design phase and series of
increasingly detailed design docs. `skimr` is only a few hundred lines
of code, which means “increasingly detailed design docs” translates to
one doc. But we did actually write it! [It’s
here](https://docs.google.com/document/d/18lBStDZzd1rJq08O-4Sw2qHhuHEZ79QX4sBkeyzWNFY/edit#heading=h.5x0d5h95i329).
And it still goes a good job of laying out some of the big ideas we were
interested in taking on for v2.

-   Eliminating frictions that resulted from differences in the way we
    stored data vs how it was displayed to users
-   Getting away from using a global environment to configure `skimr`
-   Making it easier for others to extend `skimr`
-   Create more useful ways to use `skimr`

Better internal data structures
-------------------------------

In v1, `skimr` stored all of its data in a “long format”, data frame.
Although hidden from the user by its print methods, this format would
appear any time you’d try do something with the results of a `skim()`
call. It looked something like this:

    skim(mtcars) %>% dplyr::filter(stat=="hist")

    # A tibble: 11 x 6
       variable type    stat  level value formatted
       <chr>    <chr>   <chr> <chr> <dbl> <chr>
     1 mpg      numeric hist  .all     NA ▃▇▇▇▃▂▂▂
     2 cyl      numeric hist  .all     NA ▆▁▁▃▁▁▁▇
     3 disp     numeric hist  .all     NA ▇▆▁▂▅▃▁▂
     4 hp       numeric hist  .all     NA ▃▇▃▅▂▃▁▁
     5 drat     numeric hist  .all     NA ▃▇▁▅▇▂▁▁
     6 wt       numeric hist  .all     NA ▃▃▃▇▆▁▁▂
     7 qsec     numeric hist  .all     NA ▃▂▇▆▃▃▁▁
     8 vs       numeric hist  .all     NA ▇▁▁▁▁▁▁▆
     9 am       numeric hist  .all     NA ▇▁▁▁▁▁▁▆
    10 gear     numeric hist  .all     NA ▇▁▁▆▁▁▁▂
    11 carb     numeric hist  .all     NA ▆▇▂▇▁▁▁▁

Big ups to anyone who looked at the rendered output and saw that this
was how you actually filtered the results. Hopefully there are even
better applications of your near-telepathic abilities.

Now, working with `skimr` is a bit more sane.

    skimmed <- iris %>%
      skim() %>%
      dplyr::filter(numeric.sd > 1)

    skimmed

    ## ── Data Summary ────────────────────────
    ##                            Values    
    ## Name                       Piped data
    ## Number of rows             150       
    ## Number of columns          5         
    ## _______________________              
    ## Column type frequency:               
    ##   numeric                  1         
    ## ________________________             
    ## Group variables            None      
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate  mean    sd    p0   p25   p50   p75  p100 hist 
    ## 1 Petal.Length          0             1  3.76  1.77     1   1.6  4.35   5.1   6.9 ▇▁▆▇▂

And

    dplyr::glimpse(skimmed)

    ## Observations: 1
    ## Variables: 15
    ## $ skim_type         <chr> "numeric"
    ## $ skim_variable     <chr> "Petal.Length"
    ## $ n_missing         <int> 0
    ## $ complete_rate     <dbl> 1
    ## $ factor.ordered    <lgl> NA
    ## $ factor.n_unique   <int> NA
    ## $ factor.top_counts <chr> NA
    ## $ numeric.mean      <dbl> 3.758
    ## $ numeric.sd        <dbl> 1.765298
    ## $ numeric.p0        <dbl> 1
    ## $ numeric.p25       <dbl> 1.6
    ## $ numeric.p50       <dbl> 4.35
    ## $ numeric.p75       <dbl> 5.1
    ## $ numeric.p100      <dbl> 6.9
    ## $ numeric.hist      <chr> "▇▁▆▇▂"

It’s still not perfect, as you need to rely on a *pseudo-namespace* to
refer to the column that you want. But this is unfortunately a necessary
trade-off. As the Rstats Bible, errr Hadley Wickham’s *Advanced R*,
states, all elements of [an atomic vector must have the same
type](https://adv-r.hadley.nz/vectors-chap.html). This normally isn’t
something that you have to think too much about, that is until you try
to combine the means of all your `Date` columns with the means of your
`numeric` columns and everything comes out utterly garbled. So instead
of that basket of laughs, we prefix columns names by their data type.

There’s a couple of other nuances here:

-   The data frame `skim()` produces always starts off with some
    metadata columns
-   Functions that always produce the same, regardless of input type,
    can be treated as `base_skimmers` and don’t need a namespace

### Manipulating internal data

A better representation of internal data comes with better tools for
reshaping the data and getting it for other contexts. A common request
in v1 was tooling to handle the `skimr` subtables separately. We now do
this with `partition()`. It replaces the v1 function `skim_to_list()`.

    partition(skimmed)

    ## $numeric
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate  mean    sd    p0   p25   p50   p75  p100 hist 
    ## 1 Petal.Length          0             1  3.76  1.77     1   1.6  4.35   5.1   6.9 ▇▁▆▇▂

You can undo a call to `partition()` with `bind()`, which joins the
subtables into the original `skim_df` object and properly accounts for
metadata. You can skip a step with the function `yank()`, which calls
partition and pulls out a particular subtable

    yank(skimmed, "numeric")

    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate  mean    sd    p0   p25   p50   p75  p100 hist 
    ## 1 Petal.Length          0             1  3.76  1.77     1   1.6  4.35   5.1   6.9 ▇▁▆▇▂

Last, with support something close to the older format with the
`to_long()` function. This can be added for something close to backwards
compatibility. Being realistic on open source sustainability means that
we are not able to support 100% backward compatibility in v2 even with
new functions. Meanwhile you can keep using v1 if you are happy with it.
However, because `skimr`’s dependencies are under ongoing development,
sooner or later skimr v1 will no longer work with updates to them.

### Working with dplyr

Using `skimr` in a `dplyr` pipeline was part of the original package
design, and we’ve needed to devote some extra love to making sure that
everything is as seamless as possible. Part of this is due to the object
produce by `skim()`, which we call `skim_df`. It’s a little weird in
that it needs both metadata and columns in the underlying data frame.

In practice, this means that you can coerce it into a different type
through normal `dplyr` operations. Here’s one:

    select(skimmed, numeric.mean)

    ## # A tibble: 1 x 1
    ##   numeric.mean
    ##          <dbl>
    ## 1         3.76

To get around this, we’ve added some helper functions and methods. The
more `skimr`-like replacement for `select()` is `focus()`, which
preserves metadata columns.

    focus(skimmed, numeric.mean)

    ## ── Data Summary ────────────────────────
    ##                            Values    
    ## Name                       Piped data
    ## Number of rows             150       
    ## Number of columns          5         
    ## _______________________              
    ## Column type frequency:               
    ##   numeric                  1         
    ## ________________________             
    ## Group variables            None      
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable  mean
    ## 1 Petal.Length   3.76

Configuring and extending skimr
-------------------------------

Most of `skimr`’s magic, to [steal a
term](https://resources.rstudio.com/rstudio-conf-2019/our-colour-of-magic-the-open-sourcery-of-fantastic-r-packages),
comes from the fact that you can do most everything with one function.
But believe it or not, there’s actually a bit more to the package.

One big one is customization. We like the `skimr` defaults, but that
doesn’t guarantee you will. So what if you want to do something
different, we have a function factory for that!

    my_skim <- skim_with(numeric = sfl(iqr = IQR, p25 = NULL, p75 = NULL))
    my_skim(faithful)

    ## ── Data Summary ────────────────────────
    ##                            Values  
    ## Name                       faithful
    ## Number of rows             272     
    ## Number of columns          2       
    ## _______________________            
    ## Column type frequency:             
    ##   numeric                  2       
    ## ________________________           
    ## Group variables            None    
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate  mean    sd    p0   p50  p100 hist    iqr
    ## 1 eruptions             0             1  3.49  1.14   1.6     4   5.1 ▇▂▂▇▇  2.29
    ## 2 waiting               0             1 70.9  13.6   43      76  96   ▃▃▂▇▂ 24

Those of you familiar with customizing `skim()` in v1 will notice a
couple differences:

-   we now has an object called `sfl()` for managing `skimr` function
    lists; more below
-   instead of setting global options, we now have a *function factory*

Yes! A function factory. `skim_with()` gives us a new function each time
we call it, and the returned function is configured by the arguments in
`skim_with()`. This works the same way as `ecdf()` in the `stats`
package or `colorRamp` in `grDevices`. Creating new functions has a few
advantages over the previous approach.

-   you can export a `skim()` function in a package or create it in a
    `.Rprofile`
-   you avoid a bunch of potential side effects from setting options
    with `skim_with()`

The other big change is how we now handle different data types. Although
many will never see it, a key piece of `skimr` customization comes from
the `get_skimmers()` generic. It’s used to detect different column types
in your data and set the appropriate summary functions for that type.
It’s also designed to work with `sfl()`. Here’s an example from the
“Supporting additional objects” vignette. Here, we’ll create some
skimmers for
[`sf`](https://cran.r-project.org/web/packages/sf/index.html) data
types:

    get_skimmers.sfc_POINT <- function(column) {
      sfl(
        skim_type = "sfc_POINT",
        n_unique = n_unique,
        valid = ~ sum(sf::st_is_valid(.))
      )
    }

While it was required in `skim_with()`, users must provide a `skim_type`
value when creating new methods. With that, you can export this method
in a new package (be sure to import the generic), and the new default
skimmer is added when you load the package.

    get_default_skimmer_names()

    ...
    $sfc_POINT
    [1] "missing"  "complete" "n"        "n_unique" "valid"
    ...

Even if you don’t go the full route of supporting a new data type,
creating a couple of `skimr` function lists has other benefits. For
example, you can add some to your `.Rprofile` as a way to quickly
configure `skimr` interactively.

    sfc_point_sfl <- sfl(
      n_unique = n_unique,
      valid = ~ sum(sf::st_is_valid(.))
    )

    my_skimmer <- skim_with(sfc_POINT = sfc_point_sfl)

Using skimr in other contexts
-----------------------------

In `skimr` v1, we developed some slightly hacky approaches to getting
nicer `skim()` output in RMarkdown docs. These have been removed in
favor of the
[actually-supported](https://github.com/yihui/knitr/issues/1493)
`knit_print` API. Now, calling `skim()`, within an RMarkdown doc should
produce something nice by default.

    skim(chickwts)

<table>
<caption>Data summary</caption>
<tbody>
<tr class="odd">
<td style="text-align: left;">Name</td>
<td style="text-align: left;">chickwts</td>
</tr>
<tr class="even">
<td style="text-align: left;">Number of rows</td>
<td style="text-align: left;">71</td>
</tr>
<tr class="odd">
<td style="text-align: left;">Number of columns</td>
<td style="text-align: left;">2</td>
</tr>
<tr class="even">
<td style="text-align: left;">_______________________</td>
<td style="text-align: left;"></td>
</tr>
<tr class="odd">
<td style="text-align: left;">Column type frequency:</td>
<td style="text-align: left;"></td>
</tr>
<tr class="even">
<td style="text-align: left;">factor</td>
<td style="text-align: left;">1</td>
</tr>
<tr class="odd">
<td style="text-align: left;">numeric</td>
<td style="text-align: left;">1</td>
</tr>
<tr class="even">
<td style="text-align: left;">________________________</td>
<td style="text-align: left;"></td>
</tr>
<tr class="odd">
<td style="text-align: left;">Group variables</td>
<td style="text-align: left;">None</td>
</tr>
</tbody>
</table>

**Variable type: factor**

<table>
<thead>
<tr class="header">
<th style="text-align: left;">skim_variable</th>
<th style="text-align: right;">n_missing</th>
<th style="text-align: right;">complete_rate</th>
<th style="text-align: left;">ordered</th>
<th style="text-align: right;">n_unique</th>
<th style="text-align: left;">top_counts</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td style="text-align: left;">feed</td>
<td style="text-align: right;">0</td>
<td style="text-align: right;">1</td>
<td style="text-align: left;">FALSE</td>
<td style="text-align: right;">6</td>
<td style="text-align: left;">soy: 14, cas: 12, lin: 12, sun: 12</td>
</tr>
</tbody>
</table>

**Variable type: numeric**

<table>
<thead>
<tr class="header">
<th style="text-align: left;">skim_variable</th>
<th style="text-align: right;">n_missing</th>
<th style="text-align: right;">complete_rate</th>
<th style="text-align: right;">mean</th>
<th style="text-align: right;">sd</th>
<th style="text-align: right;">p0</th>
<th style="text-align: right;">p25</th>
<th style="text-align: right;">p50</th>
<th style="text-align: right;">p75</th>
<th style="text-align: right;">p100</th>
<th style="text-align: left;">hist</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td style="text-align: left;">weight</td>
<td style="text-align: right;">0</td>
<td style="text-align: right;">1</td>
<td style="text-align: right;">261.31</td>
<td style="text-align: right;">78.07</td>
<td style="text-align: right;">108</td>
<td style="text-align: right;">204.5</td>
<td style="text-align: right;">258</td>
<td style="text-align: right;">323.5</td>
<td style="text-align: right;">423</td>
<td style="text-align: left;">▆▆▇▇▃</td>
</tr>
</tbody>
</table>

You get a nice html version of both the summary header and the `skimr`
subtables for each type of data.

In this context, you configure the output the same way you handle other
`knitr` code chunks.

This means that we’re dropping direct support for `kable.skim_df()` and
`pander.skim_df()`. But you can still get pretty similar results to
these functions by using the reshaping functions described above to get
subtables. You can also still use `Pander` and other nice rendering
packages on an ad hoc basis as you would for other data frames or
tibbles.

We also have a similarly-nice rendered output in
[Jupyter](https://github.com/ropensci/skimr/blob/8c2263c4fd4796af0e5e8f32aafc4980bd58d43a/inst/other_docs/skimr_in_jupyter.ipynb)
and RMarkdown notebooks. In the latter, the summary is separated from
the rest of the output when working interactively. We like it that way,
but we’d be happy to hear what the rest of you think!

Wait, that took over a year?
----------------------------

Well, we think that’s a lot! But to be fair, it wasn’t exactly simple to
keep up with `skimr`. Real talk, open source development takes up a lot
of time, and the `skimr` developers have additional important
priorities. Michael’s family added a new baby, and despite swearing up
and down otherwise, he got absolutely nothing not-baby-related done
during his paternity leave (take note new dads!). Elin ended up taking a
much bigger role on at Lehman, really limiting time for any other work.

Even so, these are just the highlights in the normal ebb and flow of
this sort of work. Since it’s no one’s real job, it might not always be
the first focus. And that’s OK! We’ve been really lucky to have a group
of new users that have been very patient with this slow development
cycle while still providing really good feedback throughout. Thank you
all!

We’re really excited about this next step in the `skimr` journey. We’ve
put a huge amount of work into this new version. Hopefully it shows. And
hopefully it inspires some of you to send more feedback and help us find
even more ways to improve!

---
title: "`skimr` for useful and tidy summary statistics"
authors:
  - name: Eduardo Arino de la Rubia
    url: http://github.com/earino
  - name: Shannon Ellis
    url: http://github.com/ShanEllis
  - name: Julia Stewart Lowndes
    url: http://github.com/jules32
  - name: Hope McLeod
    url: http://github.com/homcl
  - name: Amelia McNamara
    url: http://github.com/AmeliaMN
  - name: Michael Quinn
    url: https://github.com/michaelquinn32
  - name: Elin Waring
    url: https://github.com/elinw
  - name: Hao Zhu
    url: https://github.com/haozhu233
date: "07/11/2017"
output:
  md_document:
    variant: markdown_github
categories:
  - blog
tags:
  - unconf
  - testing
  - review
  - packages
  - software
  - community
  - meetings
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

Like every R user who uses summary statistics (so, everyone), our team has to rely on some combination of summary functions beyond `summary()` and `str()`. But we found them all lacking in some way because they can be generic, they don't always provide easy-to-operate-on data structures, and they are not pipeable. What we wanted was a frictionless approach for quickly skimming useful and tidy summary statistics as part of a pipeline. And so at [rOpenSci #unconf17](http://unconf17.ropensci.org/), we developed [`skimr`](https://github.com/ropenscilabs/skimr#skimr). 

In a nutshell, `skimr` will create a `skim_df` object that can be further operated upon or that provides a human-readable printout in the console. It presents reasonable default summary statistics for numerics, factors, etc, and lists counts, and missing and unique values. And the momentum is still going, thanks to our awesome team (see below)!

## Backstory

The idea for skimr as a project for the #unconf17 [was proposed by Amelia McNamara](https://github.com/ropensci/unconf17/issues/50) following [discussions on Twitter](https://twitter.com/AmeliaMN/status/774348524653834241) and an [initial package Hadley Wickham](https://github.com/hadley/precis). 

Once we were together in Los Angeles, we formed a solid team, set up a Google Doc, a Slack channel, the `ropensci/skimr` repo, and grabbed a whiteboard.  

We started off by brainstorming what we liked about existing summary packages and what other features we wanted. We started looking at example data, `mtcars`.

Here's what we liked and disliked, in Amelia's words:

```{r summary, eval=FALSE}
### "I like what we get here because mpg is numeric so these stats make sense:"
summary(mtcars$mpg)
#   Min. 1st Qu.  Median    Mean 3rd Qu.    Max.
#  10.40   15.42   19.20   20.09   22.80   33.90


### "But I don’t like this because cyl should really be a factor and shouldn't
### have these stats:"
summary(mtcars$cyl)
#   Min. 1st Qu.  Median    Mean 3rd Qu.    Max.
#  4.000   4.000   6.000   6.188   8.000   8.000


### "This is OK, but not descriptive enough. It could be clearer what I'm
### looking at."
mosaic::tally(~cyl, data = mtcars) # install.packages('mosaic')
# cyl
# 4  6  8
# 11  7 14


### "But this output isn't labeled, not ideal."
table(mtcars$cyl, mtcars$vs)
#
#     0  1
#  4  1 10
#  6  3  4
#  8 14  0


### "I like this because it returns 'sd', 'n' and 'missing'":
mosaic::favstats(~mpg, data = mtcars)
#  min     Q1 median   Q3  max     mean       sd  n missing
# 10.4 15.425   19.2 22.8 33.9 20.09062 6.026948 32       0
```

Once we had an idea of what we thought would be useful, we did a bit of market research (i.e. we made a short [presentation](https://docs.google.com/presentation/d/13Ky3-Y70STzufLJtCm6GXN8SMj2Y11riDef8b9cBgAQ/edit#slide=id.p) and ran it by other unconfers at lunch. 

## Introducing `skimr`

So what does `skimr` actually do? It allows you to skim useful summary statistics in the console, or use those statistics in a pipeable workflow.

Some features of output in the console: 

- reports missing, complete, n, sd, and quantiles
- reports numeric/int/double separately from factor/chr, and identifies class
- handles dates, logicals
- uses [Hadley's pillars](https://github.com/hadley/pillar), specifically `pillar::spark-bar()`

Here are examples of `skimr` in action: 

### Quick skim in the console: 

**Nicely separates numeric and factor variables:**

![](https://github.com/ropenscilabs/skimr/blob/24c733d7e4752c37e46e4c36693da107f42f3f55/man/figures/skim_iris.png)
<br>

**Clearly displays many numeric variables:**

![](https://github.com/ropenscilabs/skimr/blob/ecb90e22047d4a1b228bcf471650eb79b733e52e/man/figures/skim_mtcars.png)
<br>

**Also works with strings:**

![](https://github.com/ropenscilabs/skimr/blob/ecb90e22047d4a1b228bcf471650eb79b733e52e/man/figures/skim_babynames.png) <br>

### Exploring a skim\_df object

By default `skim` prints beautifully in the console, but it also produces a long, tidy-format `skim_df` object that can be computed on.

```{r, eval = FALSE}
a <- skim(chickwts)
dim(a)
# [1] 22  5
View(a)
```

<img src="https://github.com/ropenscilabs/skimr/blob/ecb90e22047d4a1b228bcf471650eb79b733e52e/man/figures/skim_chickwts_df.png" width="450px">


### Computing with the skim_df object

Maybe you just want to skim a specific portion of your data frame. Use skimr with a pipe!

```{r, eval = FALSE}
skim(mtcars) |> filter(stat == "hist")
# A tibble: 11 × 5
# var    type  stat      level value
#    <chr>   <chr> <chr>      <chr> <dbl>
# 1    mpg numeric  hist ▂▅▇▇▇▃▁▁▂▂     0
# 2    cyl numeric  hist ▆▁▁▁▃▁▁▁▁▇     0
# 3   disp numeric  hist ▇▇▅▁▁▇▃▂▁▃     0
# 4     hp numeric  hist ▆▆▇▂▇▂▃▁▁▁     0
# 5   drat numeric  hist ▃▇▂▂▃▆▅▁▁▁     0
# 6     wt numeric  hist ▂▂▂▂▇▆▁▁▁▂     0
# 7   qsec numeric  hist ▂▃▇▇▇▅▅▁▁▁     0
# 8     vs numeric  hist ▇▁▁▁▁▁▁▁▁▆     0
# 9     am numeric  hist ▇▁▁▁▁▁▁▁▁▆     0
# 10  gear numeric  hist ▇▁▁▁▆▁▁▁▁▂     0
# 11  carb numeric  hist ▆▇▂▁▇▁▁▁▁▁     0
```



### Specifying your own statistics

Another possibility is specifying your own statistics to display with `skimr`:

```{r, eval = FALSE}
funs <- list(
  iqr = IQR,
  quantile = purrr::partial(quantile, probs = .99)
)
skim_with(numeric = funs, append = FALSE)
skim_v(iris$Sepal.Length)

#  A tibble: 2 × 4
#      type     stat level value
#     <chr>    <chr> <chr> <dbl>
# 1 numeric      iqr  .all   1.3
# 2 numeric quantile   99%   7.7
```

## Our awesome team

We had a really fantastic team with diverse backgrounds, and it was really cool how organically everyone found a role for themselves during the development of `skimr`. Between brainstorming sessions, experienced coders began to iteratively develop the code while others worked on documentation and tests, and got more involved. Everyone asked questions and brainstormed together; it was a really welcoming environment. We knew that by the end of the second day of the unconf, we would present our work using only the repo's [README](https://github.com/ropenscilabs/skimr#skimr) file. So we focused on communication throughout the entire development process. 

A lot of the heavy lifting at the unconf was done by Michael, Elin, and Eduardo, and Elin has continued leading development in the month since! 

This was the original team in alphabetical order. We have also had many virtual contributors as well: see the full list of contributors [here](https://github.com/ropenscilabs/skimr/graphs/contributors).

**Eduardo Arino de la Rubia**  
Job Title: Chief Data Scientist at Domino Data Lab    
Project Contributions: Coder

**Shannon Ellis**  
Job Title: Postdoctoral fellow in the Biostatistics Department at the Johns Hopkins Bloomberg School of Public Health   
Project Contributions: Test Scripts

**Julia Stewart Lowndes**  
Job Title: Marine Data Scientist at the National Center for Ecological Analysis and Synthesis   
Project Contributions: Documentation and test scripts

**Hope McLeod**  
Job Title: Data Engineer at Kobalt Music  
Project Contributions: Documentation

**Amelia McNamara**  
Job Title: Visiting Assistant Professor of Statistical & Data Sciences at Smith College   
Project Contributions: Coder

**Michael Quinn**  
Job Title: Quantitative Analyst at Google  
Project Contributions: Coder

**Elin Waring**  
Job Title: Professor at Lehman College Sociology Department, City University of New York   
Project Contributions: Coder

**Hao Zhu**  
Job Title: Programmer Analyst at the Institute for Aging Research  
Project Contributions: Coder

## In summary (ha...)

The work we did together was only possible because of rOpenSci's incredible community and culture. For us to be able to dream up something we wanted to build and have the time and space to actually do it together was really exciting. So thank you rOpenSci and everyone in the greater community!

There is more work to be done on `skimr`, so please check out the [`skimr`](https://github.com/ropenscilabs/skimr) repo for the latest features and improvements!

Updated to use the base pipe.

---
title: "Using Fonts"
output: 
  rmarkdown::html_vignette:
    latex_engine: xelatex
vignette: >
  %\VignetteIndexEntry{Using Fonts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

One of the features of skim is the production of spark graphs for numeric data.
However these graphs may not render properly because of lack of font support 
or for other reasons, such as an operating system that does not support UTF-8. 
In your specific environment this may depend on the fonts 
installed and the operating system and may occur for only 
specific types of documents such as PDF documents.
Skimr supports kable() formatted tables, which is used in conjunction with 
fonts to render the spark graphs.

To produce spark histograms a font that supports block elements must be used. To
support the spark line graphs a font with Braille support must be used. 

Well-known fonts that support block elements include _DejaVu Sans_ and _Arial_. 
Their availability depends on your operating system. 

The yaml at the beginning of a document using custom fonts will generally be
similar to that shown below.

```
---
title: "Untitled"
mainfont: DejaVu Sans
output:
  html_document: default
  pdf_document:
    latex_engine: xelatex
  word_document: default
font-family: Times New Roman
---
```


A further discussion of this (with examples) is available in the 
"Using fonts" template for skimr.  If you are using RStudio you can open that 
template by opening a new markdown file and selecting "From template"
and then choosing it.  Alternatively this file is available inside the 
skimr folder or repository at 
inst/markdown/templates/fonts_in_skimr/skeleton/skeleton.Rmd. 

If you are having difficulties making the spark graphs work, you can opt to turn 
them off using the code below.

```
no_sparks <- skim_with(numeric = sfl(hist = NULL), ts = sfl(line_graph = NULL))

```

---
title: "Skimr defaults"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    latex_engine: xelatex
vignette: >
  %\VignetteIndexEntry{Skimr defaults}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

This vignette simply displays the default data types and summary functions for
`skimr`. Customizing `skimr` is explained in the **Using Skimr** vignette.

## The base skimmers

`skimr` has a group of functions that it applies to all data types. We call
these the "base"" skimmers:

*  `n_missing`: The number of missing values in the column.
*  `complete_rate`: The ratio of non-missing values to the total values in the
   column.

## Default skimmers

To learn more about the functions used in this package, use the function
`get_default_skimmer_names()`.

```{r}
library(skimr)
get_default_skimmer_names()
```

The counterpart to this function is `get_default_skimmers()`, which returns the
functions themselves. If you are interested in a particular class within
`skimr`, pass it as a string to either function.

```{r}
get_default_skimmer_names("numeric")
```

The same information is stored in the `skimmers_used` attribute of the object
returned by `skim()`.

[Theme song: *PSA* by
Jay-Z](https://www.youtube.com/watch?v=-LzdKH1naok)

We announced the testing version of `skimr` v2 on [June 19,
2018](https://github.com/ropenscilabs/skimr/issues/341). After more than
a year of (admittedly intermittent) work, we’re thrilled to be able to
say that the package is ready to go to CRAN. So, what happened over the
last year? And why are we so excited for V2?

## Setting the stage

Before we can talk about the last year of `skimr` development, we need
to lay out the timeline that got us to this point. For those deeply
enmeshed in `skimr` lore, all
[dozens](https://imgur.com/gallery/R1fdEt3) of you, bear with.

`skimr` was originally an [rOpenSci
unconf17](https://ropensci.org/blog/2017/07/11/skimr/) project, a big
collaboration between eight different participants that resulted in a
conceptual outline of the package and a basic working version.
Participating in the unconf was a truly magical experience, with
everyone bringing a tremendous amount of energy and ideas to the
project, and implementation happening over a flurry of [“fancy git
commits”](https://twitter.com/AmeliaMN/status/867818976666976256).

About six months later, we released our first version on CRAN. The time
between these two milestones was mostly spent on fleshing out all of the
different ideas that were generated during the unconf (like handling
grouped data frames) and fixing all the bugs we discovered along the
way.

Getting the package on CRAN opened the gates for bug reports and feature
requests on [GitHub](https://github.com/ropenscilabs/skimr/issues).
About the same time we pushed our first version to CRAN, Elin got
`skimr`’s rOpenSci’s package [peer
review](https://github.com/ropenscilabs/software-review/issues/175)
started (thank you Jennifer and Jim!), opening another incredibly useful
channel for collecting feedback on the package. All of these new ideas
and suggestions gave us the opportunity to really push `skimr` to the
next level, but doing that would require rethinking the package, from
the ground up.

A month after finishing the peer review (and six months after the
process began), we announced v2. Over the first phase of `skimr`’s life,
we accumulated 700 commits, two release, 400 GitHub stars, 95 percent
code coverage and a lifetime’s worth of [unicode rendering
bugs](https://github.com/ropenscilabs/skimr#support-for-spark-histograms)!

Just kidding! We love our little histograms, even when they don’t love
us back! For those of you that might have never seen `skimr`, using the
package typically boils down to a single function call:

    library(skimr)
    library(dplyr)
    options(width = 90)

    skim(iris)

    ## ── Data Summary ────────────────────────
    ##                            Values
    ## Name                       iris  
    ## Number of rows             150   
    ## Number of columns          5     
    ## _______________________          
    ## Column type frequency:           
    ##   factor                   1     
    ##   numeric                  4     
    ## ________________________         
    ## Group variables            None  
    ## 
    ## ── Variable type: factor ─────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate ordered n_unique top_counts               
    ## 1 Species               0             1 FALSE          3 set: 50, ver: 50, vir: 50
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate mean    sd  p0 p25  p50 p75 p100 hist 
    ## 1 Sepal.Length          0             1 5.84 0.828 4.3 5.1 5.8  6.4  7.9 ▆▇▇▅▂
    ## 2 Sepal.Width           0             1 3.06 0.436 2   2.8 3    3.3  4.4 ▁▆▇▂▁
    ## 3 Petal.Length          0             1 3.76 1.77  1   1.6 4.35 5.1  6.9 ▇▁▆▇▂
    ## 4 Petal.Width           0             1 1.20 0.762 0.1 0.3 1.3  1.8  2.5 ▇▁▇▅▃

## Getting it right

Under normal circumstances (i.e. not during a hackathon), most software
engineering projects begin with a design phase and series of
increasingly detailed design docs. `skimr` is only a few hundred lines
of code, which means “increasingly detailed design docs” translates to
one doc. But we did actually write it! [It’s
here](https://docs.google.com/document/d/18lBStDZzd1rJq08O-4Sw2qHhuHEZ79QX4sBkeyzWNFY/edit#heading=h.5x0d5h95i329).
And it still goes a good job of laying out some of the big ideas we were
interested in taking on for v2.

-   Eliminating frictions that resulted from differences in the way we
    stored data vs how it was displayed to users
-   Getting away from using a global environment to configure `skimr`
-   Making it easier for others to extend `skimr`
-   Create more useful ways to use `skimr`

## Better internal data structures

In v1, `skimr` stored all of its data in a “long format”, data frame.
Although hidden from the user by its print methods, this format would
appear any time you’d try do something with the results of a `skim()`
call. It looked something like this:

    skim(mtcars) |> dplyr::filter(stat=="hist")

    # A tibble: 11 x 6
       variable type    stat  level value formatted
       <chr>    <chr>   <chr> <chr> <dbl> <chr>
     1 mpg      numeric hist  .all     NA ▃▇▇▇▃▂▂▂
     2 cyl      numeric hist  .all     NA ▆▁▁▃▁▁▁▇
     3 disp     numeric hist  .all     NA ▇▆▁▂▅▃▁▂
     4 hp       numeric hist  .all     NA ▃▇▃▅▂▃▁▁
     5 drat     numeric hist  .all     NA ▃▇▁▅▇▂▁▁
     6 wt       numeric hist  .all     NA ▃▃▃▇▆▁▁▂
     7 qsec     numeric hist  .all     NA ▃▂▇▆▃▃▁▁
     8 vs       numeric hist  .all     NA ▇▁▁▁▁▁▁▆
     9 am       numeric hist  .all     NA ▇▁▁▁▁▁▁▆
    10 gear     numeric hist  .all     NA ▇▁▁▆▁▁▁▂
    11 carb     numeric hist  .all     NA ▆▇▂▇▁▁▁▁

Big ups to anyone who looked at the rendered output and saw that this
was how you actually filtered the results. Hopefully there are even
better applications of your near-telepathic abilities.

Now, working with `skimr` is a bit more sane.

    skimmed <- iris |>
      skim() |>
      dplyr::filter(numeric.sd > 1)

    skimmed

    ## ── Data Summary ────────────────────────
    ##                            Values
    ## Name                       iris  
    ## Number of rows             150   
    ## Number of columns          5     
    ## _______________________          
    ## Column type frequency:           
    ##   numeric                  1     
    ## ________________________         
    ## Group variables            None  
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate mean   sd p0 p25  p50 p75 p100 hist 
    ## 1 Petal.Length          0             1 3.76 1.77  1 1.6 4.35 5.1  6.9 ▇▁▆▇▂

And

    dplyr::glimpse(skimmed)

    ## Rows: 1
    ## Columns: 15
    ## $ skim_type         <chr> "numeric"
    ## $ skim_variable     <chr> "Petal.Length"
    ## $ n_missing         <int> 0
    ## $ complete_rate     <dbl> 1
    ## $ factor.ordered    <lgl> NA
    ## $ factor.n_unique   <int> NA
    ## $ factor.top_counts <chr> NA
    ## $ numeric.mean      <dbl> 3.758
    ## $ numeric.sd        <dbl> 1.765298
    ## $ numeric.p0        <dbl> 1
    ## $ numeric.p25       <dbl> 1.6
    ## $ numeric.p50       <dbl> 4.35
    ## $ numeric.p75       <dbl> 5.1
    ## $ numeric.p100      <dbl> 6.9
    ## $ numeric.hist      <chr> "▇▁▆▇▂"

It’s still not perfect, as you need to rely on a *pseudo-namespace* to
refer to the column that you want. But this is unfortunately a necessary
trade-off. As the Rstats Bible, errr Hadley Wickham’s *Advanced R*,
states, all elements of [an atomic vector must have the same
type](https://adv-r.hadley.nz/vectors-chap.html). This normally isn’t
something that you have to think too much about, that is until you try
to combine the means of all your `Date` columns with the means of your
`numeric` columns and everything comes out utterly garbled. So instead
of that basket of laughs, we prefix columns names by their data type.

There’s a couple of other nuances here:

-   The data frame `skim()` produces always starts off with some
    metadata columns
-   Functions that always produce the same, regardless of input type,
    can be treated as `base_skimmers` and don’t need a namespace

### Manipulating internal data

A better representation of internal data comes with better tools for
reshaping the data and getting it for other contexts. A common request
in v1 was tooling to handle the `skimr` subtables separately. We now do
this with `partition()`. It replaces the v1 function `skim_to_list()`.

    partition(skimmed)

    ## $numeric
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate mean   sd p0 p25  p50 p75 p100 hist 
    ## 1 Petal.Length          0             1 3.76 1.77  1 1.6 4.35 5.1  6.9 ▇▁▆▇▂

You can undo a call to `partition()` with `bind()`, which joins the
subtables into the original `skim_df` object and properly accounts for
metadata. You can skip a step with the function `yank()`, which calls
partition and pulls out a particular subtable

    yank(skimmed, "numeric")

    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate mean   sd p0 p25  p50 p75 p100 hist 
    ## 1 Petal.Length          0             1 3.76 1.77  1 1.6 4.35 5.1  6.9 ▇▁▆▇▂

Last, with support something close to the older format with the
`to_long()` function. This can be added for something close to backwards
compatibility. Being realistic on open source sustainability means that
we are not able to support 100% backward compatibility in v2 even with
new functions. Meanwhile you can keep using v1 if you are happy with it.
However, because `skimr`’s dependencies are under ongoing development,
sooner or later skimr v1 will no longer work with updates to them.

### Working with dplyr

Using `skimr` in a `dplyr` pipeline was part of the original package
design, and we’ve needed to devote some extra love to making sure that
everything is as seamless as possible. Part of this is due to the object
produce by `skim()`, which we call `skim_df`. It’s a little weird in
that it needs both metadata and columns in the underlying data frame.

In practice, this means that you can coerce it into a different type
through normal `dplyr` operations. Here’s one:

    select(skimmed, numeric.mean)

    ## # A tibble: 1 × 1
    ##   numeric.mean
    ##          <dbl>
    ## 1         3.76

To get around this, we’ve added some helper functions and methods. The
more `skimr`-like replacement for `select()` is `focus()`, which
preserves metadata columns.

    focus(skimmed, numeric.mean)

    ## ── Data Summary ────────────────────────
    ##                            Values
    ## Name                       iris  
    ## Number of rows             150   
    ## Number of columns          5     
    ## _______________________          
    ## Column type frequency:           
    ##   numeric                  1     
    ## ________________________         
    ## Group variables            None  
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable mean
    ## 1 Petal.Length  3.76

## Configuring and extending skimr

Most of `skimr`’s magic, to [steal a
term](https://resources.rstudio.com/rstudio-conf-2019/our-colour-of-magic-the-open-sourcery-of-fantastic-r-packages),
comes from the fact that you can do most everything with one function.
But believe it or not, there’s actually a bit more to the package.

One big one is customization. We like the `skimr` defaults, but that
doesn’t guarantee you will. So what if you want to do something
different, we have a function factory for that!

    my_skim <- skim_with(numeric = sfl(iqr = IQR, p25 = NULL, p75 = NULL))
    my_skim(faithful)

    ## ── Data Summary ────────────────────────
    ##                            Values  
    ## Name                       faithful
    ## Number of rows             272     
    ## Number of columns          2       
    ## _______________________            
    ## Column type frequency:             
    ##   numeric                  2       
    ## ________________________           
    ## Group variables            None    
    ## 
    ## ── Variable type: numeric ────────────────────────────────────────────────────────────────
    ##   skim_variable n_missing complete_rate  mean    sd   p0 p50 p100 hist    iqr
    ## 1 eruptions             0             1  3.49  1.14  1.6   4  5.1 ▇▂▂▇▇  2.29
    ## 2 waiting               0             1 70.9  13.6  43    76 96   ▃▃▂▇▂ 24

Those of you familiar with customizing `skim()` in v1 will notice a
couple differences:

-   we now has an object called `sfl()` for managing `skimr` function
    lists; more below
-   instead of setting global options, we now have a *function factory*

Yes! A function factory. `skim_with()` gives us a new function each time
we call it, and the returned function is configured by the arguments in
`skim_with()`. This works the same way as `ecdf()` in the `stats`
package or `colorRamp` in `grDevices`. Creating new functions has a few
advantages over the previous approach.

-   you can export a `skim()` function in a package or create it in a
    `.Rprofile`
-   you avoid a bunch of potential side effects from setting options
    with `skim_with()`

The other big change is how we now handle different data types. Although
many will never see it, a key piece of `skimr` customization comes from
the `get_skimmers()` generic. It’s used to detect different column types
in your data and set the appropriate summary functions for that type.
It’s also designed to work with `sfl()`. Here’s an example from the
“Supporting additional objects” vignette. Here, we’ll create some
skimmers for
[`sf`](https://cran.r-project.org/web/packages/sf/index.html) data
types:

    get_skimmers.sfc_POINT <- function(column) {
      sfl(
        skim_type = "sfc_POINT",
        n_unique = n_unique,
        valid = ~ sum(sf::st_is_valid(.))
      )
    }

While it was required in `skim_with()`, users must provide a `skim_type`
value when creating new methods. With that, you can export this method
in a new package (be sure to import the generic), and the new default
skimmer is added when you load the package.

    get_default_skimmer_names()

    ...
    $sfc_POINT
    [1] "missing"  "complete" "n"        "n_unique" "valid"
    ...

Even if you don’t go the full route of supporting a new data type,
creating a couple of `skimr` function lists has other benefits. For
example, you can add some to your `.Rprofile` as a way to quickly
configure `skimr` interactively.

    sfc_point_sfl <- sfl(
      n_unique = n_unique,
      valid = ~ sum(sf::st_is_valid(.))
    )

    my_skimmer <- skim_with(sfc_POINT = sfc_point_sfl)

## Using skimr in other contexts

In `skimr` v1, we developed some slightly hacky approaches to getting
nicer `skim()` output in RMarkdown docs. These have been removed in
favor of the
[actually-supported](https://github.com/yihui/knitr/issues/1493)
`knit_print` API. Now, calling `skim()`, within an RMarkdown doc should
produce something nice by default.

    skim(chickwts)

<table>
<caption>Data summary</caption>
<tbody>
<tr>
<td style="text-align: left;">Name</td>
<td style="text-align: left;">chickwts</td>
</tr>
<tr>
<td style="text-align: left;">Number of rows</td>
<td style="text-align: left;">71</td>
</tr>
<tr>
<td style="text-align: left;">Number of columns</td>
<td style="text-align: left;">2</td>
</tr>
<tr>
<td style="text-align: left;">_______________________</td>
<td style="text-align: left;"></td>
</tr>
<tr>
<td style="text-align: left;">Column type frequency:</td>
<td style="text-align: left;"></td>
</tr>
<tr>
<td style="text-align: left;">factor</td>
<td style="text-align: left;">1</td>
</tr>
<tr>
<td style="text-align: left;">numeric</td>
<td style="text-align: left;">1</td>
</tr>
<tr>
<td style="text-align: left;">________________________</td>
<td style="text-align: left;"></td>
</tr>
<tr>
<td style="text-align: left;">Group variables</td>
<td style="text-align: left;">None</td>
</tr>
</tbody>
</table>

**Variable type: factor**

<table>
<colgroup>
<col style="width: 15%" />
<col style="width: 11%" />
<col style="width: 15%" />
<col style="width: 8%" />
<col style="width: 10%" />
<col style="width: 38%" />
</colgroup>
<thead>
<tr>
<th style="text-align: left;">skim_variable</th>
<th style="text-align: right;">n_missing</th>
<th style="text-align: right;">complete_rate</th>
<th style="text-align: left;">ordered</th>
<th style="text-align: right;">n_unique</th>
<th style="text-align: left;">top_counts</th>
</tr>
</thead>
<tbody>
<tr>
<td style="text-align: left;">feed</td>
<td style="text-align: right;">0</td>
<td style="text-align: right;">1</td>
<td style="text-align: left;">FALSE</td>
<td style="text-align: right;">6</td>
<td style="text-align: left;">soy: 14, cas: 12, lin: 12, sun: 12</td>
</tr>
</tbody>
</table>

**Variable type: numeric**

<table style="width:100%;">
<colgroup>
<col style="width: 17%" />
<col style="width: 12%" />
<col style="width: 17%" />
<col style="width: 8%" />
<col style="width: 7%" />
<col style="width: 4%" />
<col style="width: 7%" />
<col style="width: 4%" />
<col style="width: 7%" />
<col style="width: 6%" />
<col style="width: 7%" />
</colgroup>
<thead>
<tr>
<th style="text-align: left;">skim_variable</th>
<th style="text-align: right;">n_missing</th>
<th style="text-align: right;">complete_rate</th>
<th style="text-align: right;">mean</th>
<th style="text-align: right;">sd</th>
<th style="text-align: right;">p0</th>
<th style="text-align: right;">p25</th>
<th style="text-align: right;">p50</th>
<th style="text-align: right;">p75</th>
<th style="text-align: right;">p100</th>
<th style="text-align: left;">hist</th>
</tr>
</thead>
<tbody>
<tr>
<td style="text-align: left;">weight</td>
<td style="text-align: right;">0</td>
<td style="text-align: right;">1</td>
<td style="text-align: right;">261.31</td>
<td style="text-align: right;">78.07</td>
<td style="text-align: right;">108</td>
<td style="text-align: right;">204.5</td>
<td style="text-align: right;">258</td>
<td style="text-align: right;">323.5</td>
<td style="text-align: right;">423</td>
<td style="text-align: left;">▆▆▇▇▃</td>
</tr>
</tbody>
</table>

You get a nice html version of both the summary header and the `skimr`
subtables for each type of data.

In this context, you configure the output the same way you handle other
`knitr` code chunks.

This means that we’re dropping direct support for `kable.skim_df()` and
`pander.skim_df()`. But you can still get pretty similar results to
these functions by using the reshaping functions described above to get
subtables. You can also still use `Pander` and other nice rendering
packages on an ad hoc basis as you would for other data frames or
tibbles.

We also have a similarly-nice rendered output in
[Jupyter](https://github.com/ropenscilabs/skimr/blob/8c2263c4fd4796af0e5e8f32aafc4980bd58d43a/inst/other_docs/skimr_in_jupyter.ipynb)
and RMarkdown notebooks. In the latter, the summary is separated from
the rest of the output when working interactively. We like it that way,
but we’d be happy to hear what the rest of you think!

## Wait, that took over a year?

Well, we think that’s a lot! But to be fair, it wasn’t exactly simple to
keep up with `skimr`. Real talk, open source development takes up a lot
of time, and the `skimr` developers have additional important
priorities. Michael’s family added a new baby, and despite swearing up
and down otherwise, he got absolutely nothing not-baby-related done
during his paternity leave (take note new dads!). Elin ended up taking a
much bigger role on at Lehman, really limiting time for any other work.

Even so, these are just the highlights in the normal ebb and flow of
this sort of work. Since it’s no one’s real job, it might not always be
the first focus. And that’s OK! We’ve been really lucky to have a group
of new users that have been very patient with this slow development
cycle while still providing really good feedback throughout. Thank you
all!

We’re really excited about this next step in the `skimr` journey. We’ve
put a huge amount of work into this new version. Hopefully it shows. And
hopefully it inspires some of you to send more feedback and help us find
even more ways to improve!

Updated to use base pipe and update URLS.

---
title: "Untitled"
mainfont: DejaVu Sans
output:
  html_document: default
  pdf_document:
    latex_engine: xelatex
  word_document: default
font-family: Times New Roman
---

## Getting ready for fonts

Notice that the yaml at the beginning of this file includes a latex_engine which will be used 
when creating a pdf document.

It also includes a mainfont setting called _DejaVu Sans_.  This is not the only font that will 
work to produce the spark graphs. However, it is a free font available through the
`extrafont` package.  If you have not installed extrafont you should do so using the normal
package installation procedures. You should then make sure that the desired font is installed.

The code below will not run automatically when you knit, instead you should run it in the
console. 

```
install.packages(c("extrafont"))
font_install("DejaVu Sans")
```

If there are any difficulties please read the extrafont documentation.

## Experimenting with rendering
 
```{r, message=FALSE}
library(knitr)
library(skimr)
```

Try knitting this document to PDF, HTML, doc or any other format you wish to try. You 
will notice that there are slight differences between them.  To understand the impact
of the engine and font choices you should experiment with different options.

The first example shows what printing the basic skim function looks like. 
You can try knitting to different formats to see how it changes.

```{r }
skim(iris)
```

It is possible that the histograms will not print in all of the formats.

Unfortunately this is outside the control of the skimr team because 
it relates to the operating system you are using, fonts installed, 
and locale. 

---
title: "Using Skimr"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using Skimr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

`skimr` is designed to provide summary statistics about variables in data frames,
tibbles, data tables and vectors. It is
opinionated in its defaults, but easy to modify.

In base R, the most similar functions are `summary()` for vectors and data
frames and `fivenum()` for numeric  vectors:

```{r}
summary(iris)
```

```{r}
summary(iris$Sepal.Length)
```

```{r}
fivenum(iris$Sepal.Length)
```

```{r}
summary(iris$Species)
```

## The `skim()` function

The core function of `skimr` is `skim()`, which is designed to work with
(grouped) data frames, and will try coerce other objects to data frames 
if possible. Like `summary()`, `skim()`'s method for data frames presents 
results for every column; the statistics it provides depend on the class of 
the variable.

### Skimming data frames

By design, the main focus of `skimr` is on data frames; it is intended to fit
well within a data [pipeline](https://r4ds.had.co.nz/pipes.html) and relies
extensively on [tidyverse](https://www.tidyverse.org/) vocabulary, which
focuses on data frames.

Results of `skim()` are *printed* horizontally, with one section per variable
type and one row per variable.

```{r, render = knitr::normal_print}
library(skimr)
skim(iris)
```

The format of the results are a single wide data frame combining the results, 
with some additional attributes and two metadata columns:

- `skim_variable`: name of the original variable
- `skim_type`: class of the variable

Unlike many other objects within `R`, these columns are intrinsic to the
`skim_df` class. Dropping these variables will result in a coercion to a
`tibble`. The `is_skim_df()` function is used to assert that an object is
a skim_df.

```{r}
skim(iris) %>% is_skim_df()

```

```{r, render = knitr::normal_print}
skim(iris) %>%
  dplyr::select(-skim_type, -skim_variable) %>% is_skim_df()
```

```{r, render = knitr::normal_print}
skim(iris) %>%
  dplyr::select(-n_missing) %>% is_skim_df()
```

In order to avoid type coercion, columns for summary statistics for different
types are prefixed with the corresponding `skim_type`. This means that the
columns of the `skim_df` are somewhat sparse, with quite a few missing
values. This is because for some statistics the representations for different
types of variables is different. For example, the mean of a Date variable and
of a numeric variable are represented differently when printing, but this 
cannot be supported in a single vector. The exception to this are 
`n_missing` and `complete_rate` (missing/number of observations) which are the 
same for all types of variables.

```{r, render = knitr::normal_print}
skim(iris) %>%
  tibble::as_tibble()
```

This is in contrast to `summary.data.frame()`, which stores statistics in a
`table`. The distinction is important, because the `skim_df` object is pipeable
and easy to use for additional manipulation: for example, the user could select
all of the variable means, or all summary statistics for a specific variable.

```{r, render = knitr::normal_print}
skim(iris) %>%
  dplyr::filter(skim_variable == "Petal.Length")
```

Most `dplyr` verbs should work as expected.

```{r, render = knitr::normal_print}
skim(iris) %>%
  dplyr::select(skim_type, skim_variable, n_missing)
```

The base skimmers `n_missing` and `complete_rate` are computed for all of the
columns in the data. But all other type-based skimmers have a namespace. You
need to use a `skim_type` prefix to refer to correct column.

```{r, render = knitr::normal_print}
skim(iris) %>%
  dplyr::select(skim_type, skim_variable, numeric.mean)
```

`skim()` also supports grouped data created by `dplyr::group_by()`. 
In this case, one additional column for each grouping variable is added 
to the `skim_df` object.

```{r, render = knitr::normal_print}
iris %>%
  dplyr::group_by(Species) %>%
  skim()
```

Individual columns from a data frame may be selected using tidyverse-style
selectors.

```{r, render = knitr::normal_print}
skim(iris, Sepal.Length, Species)
```

Or with common `select` helpers.

```{r, render = knitr::normal_print}
skim(iris, starts_with("Sepal"))
```

If an individual column is of an unsupported class, it is treated as a
character variable with a warning.

## Skimming vectors

In `skimr` v2, `skim()` will attempt to coerce non-data frames (such as vectors
and matrices) to data frames. In most cases with vectors, the object being
evaluated should be equivalent to wrapping the object in `as.data.frame()`.

For example, the `lynx` data set is class `ts`.

```{r, render = knitr::normal_print}
skim(lynx)
```

Which is the same as coercing to a data frame.

```{r}
all.equal(skim(lynx), skim(as.data.frame(lynx)))
```

## Skimming matrices

`skimr` does not support skimming matrices directly but coerces them to data
frames. Columns in the matrix become variables. This behavior is similar to
`summary.matrix()`). Three possible ways to handle matrices  with `skim()`
parallel the three variations of the mean function for matrices.

```{r, render = knitr::normal_print}
m <- matrix(c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12), nrow = 4, ncol = 3)
m
```

Skimming the matrix produces similar results to `colMeans()`.

```{r, render = knitr::normal_print}
colMeans(m)
skim(m) # Similar to summary.matrix and colMeans()
```

Skimming the transpose of the matrix will give row-wise results.

```{r, render = knitr::normal_print}
rowMeans(m)
skim(t(m))
```

And call `c()` on the matrix to get results across all columns.

```{r, render = knitr::normal_print}
skim(c(m))
mean(m)
```

### Skimming without modification

`skim_tee()` produces the same printed version as `skim()` but returns the
original, unmodified data frame. This allows for continued piping of the
original data.

```{r, render = knitr::normal_print}
iris_setosa <- iris %>%
  skim_tee() %>%
  dplyr::filter(Species == "setosa")
head(iris_setosa)
```

Note, that `skim_tee()` is customized differently than `skim` itself. See below
for more details.

## Reshaping the results from `skim()`

As noted above, `skim()` returns a wide data frame. This is usually the most
sensible format for the majority of operations when investigating data, but
the package has some other functions to help with edge cases.

First, `partition()` returns a named list of the wide data frames for each data
type. Unlike the original data the partitioned data only has columns
corresponding to the skimming functions used for this data type. These data 
frames are, therefore, not `skim_df` objects.

```{r, render = knitr::normal_print}
iris %>%
  skim() %>%
  partition()
```

Alternatively, `yank()` selects only the subtable for a specific type. Think of
it like `dplyr::select` on column types in the original data. Again, unsuitable
columns are dropped.

```{r, render = knitr::normal_print}
iris %>%
  skim() %>%
  yank("numeric")
```

`to_long()` returns a single long data frame with columns `variable`, `type`,
`statistic` and `formatted`. This is similar but not identical to the `skim_df`
object in `skimr` v1.

```{r, render = knitr::normal_print}
iris %>%
  skim() %>%
  to_long() %>% 
  head()
```

Since the `skim_variable` and `skim_type` columns are a core component of the
`skim_df` class, it's possible to get unwanted side effects when using
`dplyr::select()`. Instead, use `focus()` to select columns of the skimmed
results and keep them as a `skim_df`; it always keeps the metadata column.

```{r, render = knitr::normal_print}
iris %>%
  skim() %>%
  focus(n_missing, numeric.mean)
```

## Rendering the results of `skim()`

The `skim_df` object is a wide data frame. The display is
created by default using `print.skim_df()`; users can specify additional
options by explicitly calling `print([skim_df object], ...)`.

For documents rendered by `knitr`, the package provides a custom `knit_print`
method. To use it, the final line of your code chunk should have a `skim_df`
object.

```{r}
skim(Orange)
```

The same type of rendering is available from reshaped `skim_df` objects, those
generated by `partition()` and `yank()` in particular.

```{r}
skim(Orange) %>%
  yank("numeric")
```

## Customizing print options

Although its not a common use case outside of writing vignettes about `skimr`,
you can fall back to default printing methods by adding the chunk option
`render = knitr::normal_print`.

You can also disable the `skimr` summary by setting the chunk option
`skimr_include_summary = FALSE`.

You can change the number of digits shown in the columns of generated statistics
by changing the `skimr_digits` chunk option.

## Modifying `skim()`

`skimr` is opinionated in its choice of defaults, but users can easily add,
replace, or remove the statistics for a class. For interactive use, you can
create your own skimming function with the `skim_with()` factory. `skimr` also
has an API for extensions in other packages. Working with that is covered later.

To add a statistic for a data type, create an `sfl()` (a `skimr` function list)
for each class that you want to change:

```{r}
my_skim <- skim_with(numeric = sfl(new_mad = mad))
my_skim(faithful)
```

As the previous example suggests, the default is to append *new* summary
statistics to the preexisting set. This behavior isn't always desirable,
especially when you want lots of changes. To stop appending, set
`append = FALSE`.

```{r}
my_skim <- skim_with(numeric = sfl(new_mad = mad), append = FALSE)
my_skim(faithful)
```

You can also use `skim_with()` to remove specific statistics by setting them to
`NULL`. This is commonly used to disable the inline histograms and spark graphs.

```{r}
no_hist <- skim_with(ts = sfl(line_graph = NULL))
no_hist(Nile)
```

The same pattern applies to changing skimmers for multiple classes
simultaneously. If you want to partially-apply function arguments, use the
Tidyverse lambda syntax.

```{r}
my_skim <- skim_with(
  numeric = sfl(total = ~ sum(., na.rm = TRUE)),
  factor = sfl(missing = ~ sum(is.na(.))),
  append = FALSE
)

my_skim(iris)
```

To modify the "base" skimmers, refer to them in a similar manner. Since base
skimmers are usually a small group, they must return the same type for all
data types in R, `append` doesn't apply here.

```{r}
my_skim <- skim_with(base = sfl(length = length))
my_skim(faithful)
```

## Extending `skimr`

Packages may wish to export their own `skim()` functions. Use `skim_with()` for
this. In fact, this is how `skimr` generates its version of `skim()`.

```{r}
#' @export
my_package_skim <- skim_with()
```

Alternatively, defaults for another data types can be added to `skimr` with the
`get_skimmers` generic. The method for your data type should return an `sfl()`.
Unlike the `sfl()` used interactively, you also need to set the `skim_type`
argument. It should match the method type in the function signature.

```{r}
get_skimmers.my_type <- function(column) {
  sfl(
    skim_type = "my_type",
    total = sum
  )
}

my_data <- data.frame(
  my_type = structure(1:3, class = c("my_type", "integer"))
)
skim(my_data)
```

An extended example is available in the vignette *Supporting additional
objects*.

## Solutions to common rendering problems

The details of rendering are dependent on the operating system R is running on,
the locale of the installation, and the fonts installed. Rendering may also
differ based on whether it occurs in the console or when knitting to specific
types of documents such as HTML and PDF.

The most commonly reported problems involve rendering the spark graphs (inline
histogram and line chart) on Windows. One common fix is to switch your locale. The
function `fix_windows_histograms()` does this for you.

In order to render the sparkgraphs in html or PDF histogram you may need to
change fonts to one that supports blocks or Braille (depending on which you
need). Please review the separate vignette and associated template for details.

---
slug: skimrv2
title: "(Re)introducing `skimr` v2"
subtitle: "A year in the life of an open source R project"
package_version: 2.0
authors:
  - name: Michael Quinn
    url: https://github.com/michaelquinn32
  - name: Elin Waring
    url: https://github.com/elinw
date: "10/09/2019"
categories: blog
tags:
  - software peer review
  - R
  - packages
  - community
  - skimr
---

[Theme song: *PSA* by Jay-Z](https://www.youtube.com/watch?v=-LzdKH1naok)

We announced the testing version of `skimr` v2 on
[June 19, 2018](https://github.com/ropensci/skimr/issues/341). After more than
a year of (admittedly intermittent) work, we're thrilled to be able to say that
the package is ready to go to CRAN. So, what happened over the last year? And
why are we so excited for V2?

## Setting the stage

Before we can talk about the last year of `skimr` development, we need to lay
out the timeline that got us to this point. For those deeply enmeshed in `skimr`
lore, all [dozens](https://imgur.com/gallery/R1fdEt3) of you, bear with.

`skimr` was originally an [rOpenSci
unconf17](https://ropensci.org/blog/2017/07/11/skimr/) project, a big
collaboration between eight different participants that resulted in a conceptual
outline of the package and a basic working version. Participating in the unconf
was a truly magical experience, with everyone bringing a tremendous amount of
energy and ideas to the project, and implementation happening over a flurry of
["fancy git commits"](https://twitter.com/AmeliaMN/status/867818976666976256).

About six months later, we released our first version on CRAN. The time between
these two milestones was mostly spent on fleshing out all of the different ideas
that were generated during the unconf (like handling grouped data frames) and
fixing all the bugs we discovered along the way.

Getting the package on CRAN opened the gates for bug reports and feature
requests on [GitHub](https://github.com/ropensci/skimr/issues). About the same
time we pushed our first version to CRAN, Elin got `skimr`'s rOpenSci's package
[peer review](https://github.com/ropensci/software-review/issues/175) started
(thank you Jennifer and Jim!), opening another incredibly useful channel for
collecting feedback on the package. All of these new ideas and suggestions gave
us the opportunity to really push `skimr` to the next level, but doing that
would require rethinking the package, from the ground up.

A month after finishing the peer review (and six months after the process
began), we announced v2. Over the first phase of `skimr`'s life, we accumulated
700 commits, two release, 400 GitHub stars, 95 percent code coverage and a
lifetime's worth of [unicode rendering
bugs](https://github.com/ropensci/skimr#support-for-spark-histograms)!

Just kidding! We love our little histograms, even when they don't love us back!
For those of you that might have never seen `skimr`, using the package typically
boils down to a single function call:

```{r render = knitr::normal_print, message=FALSE}
library(skimr)
library(dplyr)
options(width = 90)

skim(iris)
```

## Getting it right

Under normal circumstances (i.e. not during a hackathon), most software
engineering projects begin with a design phase and series of increasingly
detailed design docs. `skimr` is only a few hundred lines of code, which means
"increasingly detailed design docs" translates to one doc. But we did actually
write it! [It's
here](https://docs.google.com/document/d/18lBStDZzd1rJq08O-4Sw2qHhuHEZ79QX4sBkeyzWNFY/edit#heading=h.5x0d5h95i329).
And it still goes a good job of laying out some of the big ideas we were
interested in taking on for v2.

* Eliminating frictions that resulted from differences in the way we stored data
  vs how it was displayed to users
* Getting away from using a global environment to configure `skimr`
* Making it easier for others to extend `skimr`
* Create more useful ways to use `skimr`

## Better internal data structures

In v1, `skimr` stored all of its data in a "long format", data frame. Although
hidden from the user by its print methods, this format would appear any time
you'd try do something with the results of a `skim()` call. It looked something
like this:

```{r eval = FALSE}
skim(mtcars) %>% dplyr::filter(stat=="hist")
```
```
# A tibble: 11 x 6
   variable type    stat  level value formatted
   <chr>    <chr>   <chr> <chr> <dbl> <chr>
 1 mpg      numeric hist  .all     NA ▃▇▇▇▃▂▂▂
 2 cyl      numeric hist  .all     NA ▆▁▁▃▁▁▁▇
 3 disp     numeric hist  .all     NA ▇▆▁▂▅▃▁▂
 4 hp       numeric hist  .all     NA ▃▇▃▅▂▃▁▁
 5 drat     numeric hist  .all     NA ▃▇▁▅▇▂▁▁
 6 wt       numeric hist  .all     NA ▃▃▃▇▆▁▁▂
 7 qsec     numeric hist  .all     NA ▃▂▇▆▃▃▁▁
 8 vs       numeric hist  .all     NA ▇▁▁▁▁▁▁▆
 9 am       numeric hist  .all     NA ▇▁▁▁▁▁▁▆
10 gear     numeric hist  .all     NA ▇▁▁▆▁▁▁▂
11 carb     numeric hist  .all     NA ▆▇▂▇▁▁▁▁
```

Big ups to anyone who looked at the rendered output and saw that this was how
you actually filtered the results. Hopefully there are even better applications
of your near-telepathic abilities.

Now, working with `skimr` is a bit more sane.

```{r render = knitr::normal_print}
skimmed <- iris %>%
  skim() %>%
  dplyr::filter(numeric.sd > 1)

skimmed
```

And

```{r render = knitr::normal_print}
dplyr::glimpse(skimmed)
```

It's still not perfect, as you need to rely on a *pseudo-namespace* to refer to
the column that you want. But this is unfortunately a necessary trade-off. As
the Rstats Bible, errr Hadley Wickham's *Advanced R*, states, all elements of
[an atomic vector must have the same
type](https://adv-r.hadley.nz/vectors-chap.html). This normally isn't something
that you have to think too much about, that is until you try to combine the
means of all your `Date` columns with the means of your `numeric` columns and
everything comes out utterly garbled. So instead of that basket of laughs, we
prefix columns names by their data type.

There's a couple of other nuances here:

* The data frame `skim()` produces always starts off with some metadata columns
* Functions that always produce the same, regardless of input type, can
  be treated as `base_skimmers` and don't need a namespace

### Manipulating internal data

A better representation of internal data comes with better tools for reshaping
the data and getting it for other contexts. A common request in v1 was tooling
to handle the `skimr` subtables separately. We now do this with `partition()`.
It replaces the v1 function `skim_to_list()`.

```{r render = knitr::normal_print}
partition(skimmed)
```

You can undo a call to `partition()` with `bind()`, which joins the subtables
into the original `skim_df` object and properly accounts for metadata. You
can skip a step with the function `yank()`, which calls partition and pulls
out a particular subtable


```{r render = knitr::normal_print}
yank(skimmed, "numeric")
```

Last, with support something close to the older format with the `to_long()`
function. This can be added for something close to backwards compatibility.
Being realistic on open source sustainability means that we are not able to
support 100% backward compatibility in v2 even with new functions.  Meanwhile
you can keep using v1 if you are happy with it.  However,
because `skimr`'s dependencies are under ongoing development, sooner or later
skimr v1 will no longer work with updates to them.

### Working with dplyr

Using `skimr` in a `dplyr` pipeline was part of the original package design, and
we've needed to devote some extra love to making sure that everything is as
seamless as possible. Part of this is due to the object produce by `skim()`,
which we call `skim_df`. It's a little weird in that it needs both metadata and
columns in the underlying data frame.

In practice, this means that you can coerce it into a different type through
normal `dplyr` operations. Here's one:

```{r render = knitr::normal_print}
select(skimmed, numeric.mean)
```

To get around this, we've added some helper functions and methods. The more
`skimr`-like replacement for `select()` is `focus()`, which preserves
metadata columns.

```{r render = knitr::normal_print}
focus(skimmed, numeric.mean)
```

## Configuring and extending skimr

Most of `skimr`'s magic, to [steal a
term](https://resources.rstudio.com/rstudio-conf-2019/our-colour-of-magic-the-open-sourcery-of-fantastic-r-packages),
comes from the fact that you can do
most everything with one function. But believe it or not, there's actually a bit
more to the package.

One big one is customization. We like the `skimr` defaults, but that doesn't
guarantee you will. So what if you want to do something different, we have
a function factory for that!

```{r render = knitr::normal_print}
my_skim <- skim_with(numeric = sfl(iqr = IQR, p25 = NULL, p75 = NULL))
my_skim(faithful)
```

Those of you familiar with customizing `skim()` in v1 will notice a couple
differences:

* we now has an object called `sfl()` for managing `skimr` function lists; more
  below
* instead of setting global options, we now have a *function factory*

Yes! A function factory. `skim_with()` gives us a new function each time
we call it, and the returned function is configured by the arguments in
`skim_with()`. This works the same way as `ecdf()` in the `stats` package or
`colorRamp` in `grDevices`. Creating new functions has a few advantages over
the previous approach.

* you can export a `skim()` function in a package or create it in a `.Rprofile`
* you avoid a bunch of potential side effects from setting options with
  `skim_with()`

The other big change is how we now handle different data types. Although many
will never see it, a key piece of `skimr` customization comes from the
`get_skimmers()` generic. It's used to detect different column types in your
data and set the appropriate summary functions for that type. It's also
designed to work with `sfl()`. Here's an example from the "Supporting additional
objects" vignette. Here, we'll create some skimmers for
[`sf`](https://cran.r-project.org/web/packages/sf/index.html) data types:

```{r eval = FALSE}
get_skimmers.sfc_POINT <- function(column) {
  sfl(
    skim_type = "sfc_POINT",
    n_unique = n_unique,
    valid = ~ sum(sf::st_is_valid(.))
  )
}
```

While it was required in `skim_with()`, users must provide a `skim_type` value
when creating new methods. With that, you can export this method in a new
package (be sure to import the generic), and the new default skimmer is added
when you load the package.

```{r, eval = FALSE}
get_default_skimmer_names()
```
```
...
$sfc_POINT
[1] "missing"  "complete" "n"        "n_unique" "valid"
...
```

Even if you don't go the full route of supporting a new data type, creating a
couple of `skimr` function lists has other benefits. For example, you can add
some to your `.Rprofile` as a way to quickly configure `skimr` interactively.

```{r eval = FALSE}
sfc_point_sfl <- sfl(
  n_unique = n_unique,
  valid = ~ sum(sf::st_is_valid(.))
)

my_skimmer <- skim_with(sfc_POINT = sfc_point_sfl)
```

## Using skimr in other contexts

In `skimr` v1, we developed some slightly hacky approaches to getting nicer
`skim()` output in RMarkdown docs. These have been removed in favor of the
[actually-supported](https://github.com/yihui/knitr/issues/1493) `knit_print`
API. Now, calling `skim()`, within an RMarkdown doc should produce something
nice by default.

```{r}
skim(chickwts)
```

You get a nice html version of both the summary header and the `skimr` subtables
for each type of data.

In this context, you configure the output the same way you handle other `knitr`
code chunks.

~~~
```{r skimr_digits = 4, skimr_summary = TRUE}
```
~~~

This means that we're dropping direct support for `kable.skim_df()` and
`pander.skim_df()`. But you can still get pretty similar results to these
functions by using the reshaping functions described above to get subtables. You
can also still use `Pander` and other nice rendering packages on an ad hoc basis
as you would for other data frames or tibbles.

We also have a similarly-nice rendered output in
[Jupyter](https://github.com/ropensci/skimr/blob/8c2263c4fd4796af0e5e8f32aafc4980bd58d43a/inst/other_docs/skimr_in_jupyter.ipynb)
and RMarkdown notebooks. In the latter, the summary is separated from the rest
of the output when working interactively. We like it that way, but we'd be happy
to hear what the rest of you think!

## Wait, that took over a year?

Well, we think that's a lot! But to be fair, it wasn't exactly simple to keep up
with `skimr`. Real talk, open source development takes up a lot of time, and
the `skimr` developers have additional important priorities.  Michael's family
added a new baby, and despite swearing up and down otherwise,
he got absolutely nothing not-baby-related done during his paternity leave (take
note new dads!). Elin
ended up taking a much bigger role on at Lehman, really limiting time for any
other work.

Even so, these are just the highlights in the normal ebb and flow of this sort
of work. Since it's no one's real job, it might not always be the first focus.
And that's OK! We've been really lucky to have a group of new users that have
been very patient with this slow development cycle while still providing really
good feedback throughout. Thank you all!

We're really excited about this next step in the `skimr` journey. We've put a
huge amount of work into this new version. Hopefully it shows. And hopefully
it inspires some of you to send more feedback and help us find even more ways
to improve!

Like every R user who uses summary statistics (so, everyone), our team has to rely on some combination of summary functions beyond `summary()` and `str()`. But we found them all lacking in some way because they can be generic, they don't always provide easy-to-operate-on data structures, and they are not pipeable. What we wanted was a frictionless approach for quickly skimming useful and tidy summary statistics as part of a pipeline. And so at [rOpenSci \#unconf17](http://unconf17.ropensci.org/), we developed [`skimr`](https://github.com/ropenscilabs/skimr#skimr).

In a nutshell, `skimr` will create a `skim_df` object that can be further operated upon or that provides a human-readable printout in the console. It presents reasonable default summary statistics for numerics, factors, etc, and lists counts, and missing and unique values. And the momentum is still going, thanks to our awesome team (see below)!

Backstory
---------

The idea for skimr as a project for the \#unconf17 [was proposed by Amelia McNamara](https://github.com/ropensci/unconf17/issues/50) following [discussions on Twitter](https://twitter.com/AmeliaMN/status/774348524653834241) and an [initial package Hadley Wickham](https://github.com/hadley/precis).

Once we were together in Los Angeles, we formed a solid team, set up a Google Doc, a Slack channel, the `ropensci/skimr` repo, and grabbed a whiteboard.

We started off by brainstorming what we liked about existing summary packages and what other features we wanted. We started looking at example data, `mtcars`.

Here's what we liked and disliked, in Amelia's words:

``` r
### "I like what we get here because mpg is numeric so these stats make sense:" 
summary(mtcars$mpg) 
#   Min. 1st Qu.  Median    Mean 3rd Qu.    Max. 
#  10.40   15.42   19.20   20.09   22.80   33.90 


### "But I don’t like this because cyl should really be a factor and shouldn't have these stats:"
summary(mtcars$cyl)
#   Min. 1st Qu.  Median    Mean 3rd Qu.    Max. 
#  4.000   4.000   6.000   6.188   8.000   8.000 


### "This is OK, but not descriptive enough. It could be clearer what I'm looking at."
mosaic::tally(~cyl, data=mtcars) # install.packages('mosaic')
#cyl
# 4  6  8 
#11  7 14 


### "But this output isn't labeled, not ideal." 
table(mtcars$cyl, mtcars$vs)
#   
#     0  1
#  4  1 10
#  6  3  4
#  8 14  0


### "I like this because it returns 'sd', 'n' and 'missing'":
mosaic::favstats(~mpg, data=mtcars) 
#  min     Q1 median   Q3  max     mean       sd  n missing
# 10.4 15.425   19.2 22.8 33.9 20.09062 6.026948 32       0
```

Once we had an idea of what we thought would be useful, we did a bit of market research (i.e. we made a short [presentation](https://docs.google.com/presentation/d/13Ky3-Y70STzufLJtCm6GXN8SMj2Y11riDef8b9cBgAQ/edit#slide=id.p) and ran it by other unconfers at lunch.

Introducing `skimr`
-------------------

So what does `skimr` actually do? It allows you to skim useful summary statistics in the console, or use those statistics in a pipeable workflow.

Some features of output in the console:

-   reports missing, complete, n, sd, and quantiles
-   reports numeric/int/double separately from factor/chr, and identifies class
-   handles dates, logicals
-   uses [Hadley's pillars](https://github.com/hadley/pillar), specifically `pillar::spark-bar()`

Here are examples of `skimr` in action:

### Quick skim in the console:

**Nicely separates numeric and factor variables:**

![](https://github.com/ropenscilabs/skimr/blob/24c733d7e4752c37e46e4c36693da107f42f3f55/man/figures/skim_iris.png) <br>

**Clearly displays many numeric variables:**

![](https://github.com/ropenscilabs/skimr/blob/ecb90e22047d4a1b228bcf471650eb79b733e52e/man/figures/skim_mtcars.png) <br>

**Also works with strings:**

![](https://github.com/ropenscilabs/skimr/blob/ecb90e22047d4a1b228bcf471650eb79b733e52e/man/figures/skim_babynames.png) <br>

### Exploring a skim\_df object

By default `skim` prints beautifully in the console, but it also produces a long, tidy-format `skim_df` object that can be computed on.

``` r
a <-  skim(chickwts)
dim(a)
# [1] 22  5
View(a)
```

<img src="https://github.com/ropenscilabs/skimr/blob/ecb90e22047d4a1b228bcf471650eb79b733e52e/man/figures/skim_chickwts_df.png" width="450px">

### Computing with the skim\_df object

Maybe you just want to skim a specific portion of your data frame. Use skimr with a pipe!

``` r
> skim(mtcars) %>% filter(stat=="hist")
# A tibble: 11 × 5
     var    type  stat      level value
   <chr>   <chr> <chr>      <chr> <dbl>
1    mpg numeric  hist ▂▅▇▇▇▃▁▁▂▂     0
2    cyl numeric  hist ▆▁▁▁▃▁▁▁▁▇     0
3   disp numeric  hist ▇▇▅▁▁▇▃▂▁▃     0
4     hp numeric  hist ▆▆▇▂▇▂▃▁▁▁     0
5   drat numeric  hist ▃▇▂▂▃▆▅▁▁▁     0
6     wt numeric  hist ▂▂▂▂▇▆▁▁▁▂     0
7   qsec numeric  hist ▂▃▇▇▇▅▅▁▁▁     0
8     vs numeric  hist ▇▁▁▁▁▁▁▁▁▆     0
9     am numeric  hist ▇▁▁▁▁▁▁▁▁▆     0
10  gear numeric  hist ▇▁▁▁▆▁▁▁▁▂     0
11  carb numeric  hist ▆▇▂▁▇▁▁▁▁▁     0
```

### Specifying your own statistics

Another possibility is specifying your own statistics to display with `skimr`:

``` r
 funs <- list(iqr = IQR,
    quantile = purrr::partial(quantile, probs = .99))
  skim_with(numeric = funs, append = FALSE)
  skim_v(iris$Sepal.Length)
  
#  A tibble: 2 × 4
#      type     stat level value
#     <chr>    <chr> <chr> <dbl>
# 1 numeric      iqr  .all   1.3
# 2 numeric quantile   99%   7.7
```

Our awesome team
----------------

We had a really fantastic team with diverse backgrounds, and it was really cool how organically everyone found a role for themselves during the development of `skimr`. Between brainstorming sessions, experienced coders began to iteratively develop the code while others worked on documentation and tests, and got more involved. Everyone asked questions and brainstormed together; it was a really welcoming environment. We knew that by the end of the second day of the unconf, we would present our work using only the repo's [README](https://github.com/ropenscilabs/skimr#skimr) file. So we focused on communication throughout the entire development process.

A lot of the heavy lifting at the unconf was done by Michael, Elin, and Eduardo, and Elin has continued leading development in the month since!

This was the original team in alphabetical order. We have also had many virtual contributors as well: see the full list of contributors [here](https://github.com/ropenscilabs/skimr/graphs/contributors).

**Eduardo Arino de la Rubia**
Job Title: Chief Data Scientist at Domino Data Lab
Project Contributions: Coder

**Shannon Ellis**
Job Title: Postdoctoral fellow in the Biostatistics Department at the Johns Hopkins Bloomberg School of Public Health
Project Contributions: Test Scripts

**Julia Stewart Lowndes**
Job Title: Marine Data Scientist at the National Center for Ecological Analysis and Synthesis
Project Contributions: Documentation and test scripts

**Hope McLeod**
Job Title: Data Engineer at Kobalt Music
Project Contributions: Documentation

**Amelia McNamara**
Job Title: Visiting Assistant Professor of Statistical & Data Sciences at Smith College
Project Contributions: Coder

**Michael Quinn**
Job Title: Quantitative Analyst at Google
Project Contributions: Coder

**Elin Waring**
Job Title: Professor at Lehman College Sociology Department, City University of New York
Project Contributions: Coder

**Hao Zhu**
Job Title: Programmer Analyst at the Institute for Aging Research
Project Contributions: Coder

In summary (ha...)
------------------

The work we did together was only possible because of rOpenSci's incredible community and culture. For us to be able to dream up something we wanted to build and have the time and space to actually do it together was really exciting. So thank you rOpenSci and everyone in the greater community!

There is more work to be done on `skimr`, so please check out the [`skimr`](https://github.com/ropenscilabs/skimr) repo for the latest features and improvements!

---
output:
  html_document: default
  pdf_document:
    latex_engine: xelatex
---

```{r}
library(skimr)
data <- read.csv("https://github.com/ropensci/skimr/files/11214724/githubdata.csv")
```

```{r}
skim(data)
```
```{r}
hist(data$h, breaks =12)
```

# dplyr::filter works as expected

    Code
      dplyr::filter(skimmed_iris, skim_type == "numeric")
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       iris  
      Number of rows             150   
      Number of columns          5     
      _______________________          
      Column type frequency:           
        numeric                  4     
      ________________________         
      Group variables            None  
      
      -- Variable type: numeric ------------------------------------------------------
        skim_variable n_missing complete_rate mean    sd  p0 p25  p50 p75 p100 hist 
      1 Sepal.Length          0             1 5.84 0.828 4.3 5.1 5.8  6.4  7.9 ▆▇▇▅▂
      2 Sepal.Width           0             1 3.06 0.436 2   2.8 3    3.3  4.4 ▁▆▇▂▁
      3 Petal.Length          0             1 3.76 1.77  1   1.6 4.35 5.1  6.9 ▇▁▆▇▂
      4 Petal.Width           0             1 1.20 0.762 0.1 0.3 1.3  1.8  2.5 ▇▁▇▅▃
    Code
      dplyr::filter(skimmed_iris, skim_type == "no_type")
    Output
      # A tibble: 0 x 15
      # i 15 variables: skim_type <chr>, skim_variable <chr>, n_missing <int>,
      #   complete_rate <dbl>, factor.ordered <lgl>, factor.n_unique <int>,
      #   factor.top_counts <chr>, numeric.mean <dbl>, numeric.sd <dbl>,
      #   numeric.p0 <dbl>, numeric.p25 <dbl>, numeric.p50 <dbl>, numeric.p75 <dbl>,
      #   numeric.p100 <dbl>, numeric.hist <chr>

# dplyr::select works as expected

    Code
      with_type
    Output
      # A tibble: 5 x 2
        skim_type skim_variable
        <chr>     <chr>        
      1 factor    Species      
      2 numeric   Sepal.Length 
      3 numeric   Sepal.Width  
      4 numeric   Petal.Length 
      5 numeric   Petal.Width  

---

    Code
      without_type
    Output
      # A tibble: 5 x 1
        numeric.mean
               <dbl>
      1        NA   
      2         5.84
      3         3.06
      4         3.76
      5         1.20

# dplyr::mutate works as expected

    Code
      input
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       iris  
      Number of rows             150   
      Number of columns          5     
      _______________________          
      Column type frequency:           
        factor                   1     
        numeric                  4     
      ________________________         
      Group variables            None  
      
      -- Variable type: factor -------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique
      1 Species               0             1 FALSE          3
        top_counts               
      1 set: 50, ver: 50, vir: 50
      
      -- Variable type: numeric ------------------------------------------------------
        skim_variable n_missing complete_rate mean    sd  p0 p25  p50 p75 p100 hist 
      1 Sepal.Length          0             1 5.84 0.828 4.3 5.1 5.8  6.4  7.9 ▆▇▇▅▂
      2 Sepal.Width           0             1 3.06 0.436 2   2.8 3    3.3  4.4 ▁▆▇▂▁
      3 Petal.Length          0             1 3.76 1.77  1   1.6 4.35 5.1  6.9 ▇▁▆▇▂
      4 Petal.Width           0             1 1.20 0.762 0.1 0.3 1.3  1.8  2.5 ▇▁▇▅▃
        mean2
      1 34.1 
      2  9.35
      3 14.1 
      4  1.44

# dplyr::slice works as expected

    Code
      input
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       iris  
      Number of rows             150   
      Number of columns          5     
      _______________________          
      Column type frequency:           
        factor                   1     
        numeric                  2     
      ________________________         
      Group variables            None  
      
      -- Variable type: factor -------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique
      1 Species               0             1 FALSE          3
        top_counts               
      1 set: 50, ver: 50, vir: 50
      
      -- Variable type: numeric ------------------------------------------------------
        skim_variable n_missing complete_rate mean    sd  p0 p25 p50 p75 p100 hist 
      1 Sepal.Length          0             1 5.84 0.828 4.3 5.1 5.8 6.4  7.9 ▆▇▇▅▂
      2 Sepal.Width           0             1 3.06 0.436 2   2.8 3   3.3  4.4 ▁▆▇▂▁

# dplyr::arrange works as expected

    Code
      dplyr::arrange(skimmed_iris, desc(numeric.mean))
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       iris  
      Number of rows             150   
      Number of columns          5     
      _______________________          
      Column type frequency:           
        factor                   1     
        numeric                  4     
      ________________________         
      Group variables            None  
      
      -- Variable type: factor -------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique
      1 Species               0             1 FALSE          3
        top_counts               
      1 set: 50, ver: 50, vir: 50
      
      -- Variable type: numeric ------------------------------------------------------
        skim_variable n_missing complete_rate mean    sd  p0 p25  p50 p75 p100 hist 
      1 Sepal.Length          0             1 5.84 0.828 4.3 5.1 5.8  6.4  7.9 ▆▇▇▅▂
      2 Petal.Length          0             1 3.76 1.77  1   1.6 4.35 5.1  6.9 ▇▁▆▇▂
      3 Sepal.Width           0             1 3.06 0.436 2   2.8 3    3.3  4.4 ▁▆▇▂▁
      4 Petal.Width           0             1 1.20 0.762 0.1 0.3 1.3  1.8  2.5 ▇▁▇▅▃

# The summary print method prints the correct object

    Code
      skim_summary_input
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       iris  
      Number of rows             150   
      Number of columns          5     
      _______________________          
      Column type frequency:           
        factor                   1     
        numeric                  4     
      ________________________         
      Group variables            None  

# The summary print method prints the correct object when piped

    Code
      summary_input
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       iris  
      Number of rows             150   
      Number of columns          5     
      _______________________          
      Column type frequency:           
        factor                   1     
        numeric                  4     
      ________________________         
      Group variables            None  

# Using skim_tee prints returns the object

    Code
      skim_object <- skim_tee(chickwts)
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       data  
      Number of rows             71    
      Number of columns          2     
      _______________________          
      Column type frequency:           
        factor                   1     
        numeric                  1     
      ________________________         
      Group variables            None  
      
      -- Variable type: factor -------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique
      1 feed                  0             1 FALSE          6
        top_counts                        
      1 soy: 14, cas: 12, lin: 12, sun: 12
      
      -- Variable type: numeric ------------------------------------------------------
        skim_variable n_missing complete_rate mean   sd  p0  p25 p50  p75 p100 hist 
      1 weight                0             1 261. 78.1 108 204. 258 324.  423 ▆▆▇▇▃

# skim_tee prints only selected columns, but returns full object

    Code
      obj <- skim_tee(iris, Species)
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       data  
      Number of rows             150   
      Number of columns          5     
      _______________________          
      Column type frequency:           
        factor                   1     
      ________________________         
      Group variables            None  
      
      -- Variable type: factor -------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique
      1 Species               0             1 FALSE          3
        top_counts               
      1 set: 50, ver: 50, vir: 50

# skim_tee supports dplyr helpers

    Code
      obj <- skim_tee(iris, starts_with("Sepal"))
    Output
      -- Data Summary ------------------------
                                 Values
      Name                       data  
      Number of rows             150   
      Number of columns          5     
      _______________________          
      Column type frequency:           
        numeric                  2     
      ________________________         
      Group variables            None  
      
      -- Variable type: numeric ------------------------------------------------------
        skim_variable n_missing complete_rate mean    sd  p0 p25 p50 p75 p100 hist 
      1 Sepal.Length          0             1 5.84 0.828 4.3 5.1 5.8 6.4  7.9 ▆▇▇▅▂
      2 Sepal.Width           0             1 3.06 0.436 2   2.8 3   3.3  4.4 ▁▆▇▂▁

# Skim_tee works with groups

    Code
      obj <- skim_tee(iris_grouped, Sepal.Length, skim_fun = my_skim)
    Output
      -- Data Summary ------------------------
                                 Values 
      Name                       data   
      Number of rows             150    
      Number of columns          5      
      _______________________           
      Column type frequency:            
        numeric                  1      
      ________________________          
      Group variables            Species
      
      -- Variable type: numeric ------------------------------------------------------
        skim_variable Species    n_missing complete_rate mean    sd  p0  p25 p50 p75
      1 Sepal.Length  setosa             0             1 5.01 0.352 4.3 4.8  5   5.2
      2 Sepal.Length  versicolor         0             1 5.94 0.516 4.9 5.6  5.9 6.3
      3 Sepal.Length  virginica          0             1 6.59 0.636 4.9 6.22 6.5 6.9
        p100
      1  5.8
      2  7  
      3  7.9

# skim of a simple data.table produces output as expected

    Code
      skimmed_DT_letters
    Output
      -- Data Summary ------------------------
                                 Values    
      Name                       DT_letters
      Number of rows             26        
      Number of columns          1         
      Key                        NULL      
      _______________________              
      Column type frequency:               
        character                1         
      ________________________             
      Group variables            None      
      
      -- Variable type: character ---------------------------------------------------------------
        skim_variable n_missing complete_rate min max empty n_unique whitespace
      1 abc                   0             1   1   1     0       26          0

# skim of data.table produces output as expected

    Code
      skim(DT_factors)
    Output
      -- Data Summary ------------------------
                                 Values    
      Name                       DT_factors
      Number of rows             26        
      Number of columns          3         
      Key                        NULL      
      _______________________              
      Column type frequency:               
        character                1         
        factor                   1         
        numeric                  1         
      ________________________             
      Group variables            None      
      
      -- Variable type: character ---------------------------------------------------------------
        skim_variable n_missing complete_rate min max empty n_unique whitespace
      1 abc                   0             1   1   1     0       26          0
      
      -- Variable type: factor ------------------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique top_counts   
      1 grps                  0             1 FALSE          2 AA: 18, BB: 8
      
      -- Variable type: numeric -----------------------------------------------------------------
        skim_variable n_missing complete_rate   mean    sd    p0    p25     p50   p75 p100 hist 
      1 values                0             1 0.0121 0.937 -2.21 -0.335 -0.0306 0.742 1.36 ▂▂▃▇▆

---

    Code
      skim(DT_factors)
    Output
      -- Data Summary ------------------------
                                 Values    
      Name                       DT_factors
      Number of rows             26        
      Number of columns          3         
      Key                        abc, grps 
      _______________________              
      Column type frequency:               
        character                1         
        factor                   1         
        numeric                  1         
      ________________________             
      Group variables            None      
      
      -- Variable type: character ---------------------------------------------------------------
        skim_variable n_missing complete_rate min max empty n_unique whitespace
      1 abc                   0             1   1   1     0       26          0
      
      -- Variable type: factor ------------------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique top_counts   
      1 grps                  0             1 FALSE          2 AA: 18, BB: 8
      
      -- Variable type: numeric -----------------------------------------------------------------
        skim_variable n_missing complete_rate   mean    sd    p0    p25     p50   p75 p100 hist 
      1 values                0             1 0.0121 0.937 -2.21 -0.335 -0.0306 0.742 1.36 ▂▂▃▇▆

---

    Code
      skim(DF_factors)
    Output
      -- Data Summary ------------------------
                                 Values    
      Name                       DF_factors
      Number of rows             26        
      Number of columns          3         
      _______________________              
      Column type frequency:               
        character                1         
        factor                   1         
        numeric                  1         
      ________________________             
      Group variables            None      
      
      -- Variable type: character ---------------------------------------------------------------
        skim_variable n_missing complete_rate min max empty n_unique whitespace
      1 abc                   0             1   1   1     0       26          0
      
      -- Variable type: factor ------------------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique top_counts   
      1 grps                  0             1 FALSE          2 AA: 18, BB: 8
      
      -- Variable type: numeric -----------------------------------------------------------------
        skim_variable n_missing complete_rate   mean    sd    p0    p25     p50   p75 p100 hist 
      1 values                0             1 0.0121 0.937 -2.21 -0.335 -0.0306 0.742 1.36 ▂▂▃▇▆

---

    Code
      skim(tibble_factors)
    Output
      -- Data Summary ------------------------
                                 Values        
      Name                       tibble_factors
      Number of rows             26            
      Number of columns          3             
      _______________________                  
      Column type frequency:                   
        character                1             
        factor                   1             
        numeric                  1             
      ________________________                 
      Group variables            None          
      
      -- Variable type: character ---------------------------------------------------------------
        skim_variable n_missing complete_rate min max empty n_unique whitespace
      1 abc                   0             1   1   1     0       26          0
      
      -- Variable type: factor ------------------------------------------------------------------
        skim_variable n_missing complete_rate ordered n_unique top_counts   
      1 grps                  0             1 FALSE          2 AA: 18, BB: 8
      
      -- Variable type: numeric -----------------------------------------------------------------
        skim_variable n_missing complete_rate   mean    sd    p0    p25     p50   p75 p100 hist 
      1 values                0             1 0.0121 0.937 -2.21 -0.335 -0.0306 0.742 1.36 ▂▂▃▇▆

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reshape.R
\name{focus}
\alias{focus}
\title{Only show a subset of summary statistics after skimming}
\usage{
focus(.data, ...)
}
\arguments{
\item{.data}{A \code{skim_df} object.}

\item{...}{One or more unquoted expressions separated by commas. Variable
names can be used as if they were positions in the data frame, so
expressions like x:y can be used to select a range of variables.}
}
\description{
This function is a variant of \code{\link[dplyr:select]{dplyr::select()}} designed to work with
\code{skim_df} objects. When using \code{focus()}, \code{skimr} metadata columns are kept,
and \code{skimr} print methods are still utilized. Otherwise, the signature and
behavior is identical to \code{\link[dplyr:select]{dplyr::select()}}.
}
\examples{
# Compare
iris |>
  skim() |>
  dplyr::select(n_missing)

iris |>
  skim() |>
  focus(n_missing)

# This is equivalent to
iris |>
  skim() |>
  dplyr::select(skim_variable, skim_type, n_missing)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reshape.R
\name{focus}
\alias{focus}
\title{Only show a subset of summary statistics after skimming}
\usage{
focus(.data, ...)
}
\arguments{
\item{.data}{A \code{skim_df} object.}

\item{...}{One or more unquoted expressions separated by commas. Variable
names can be used as if they were positions in the data frame, so
expressions like x:y can be used to select a range of variables.}
}
\description{
This function is a variant of \code{\link[dplyr:select]{dplyr::select()}} designed to work with
\code{skim_df} objects. When using \code{focus()}, \code{skimr} metadata columns are kept,
and \code{skimr} print methods are still utilized. Otherwise, the signature and
behavior is identical to \code{\link[dplyr:select]{dplyr::select()}}.
}
\examples{
# Compare
iris \%>\%
  skim() \%>\%
  dplyr::select(n_missing)

iris \%>\%
  skim() \%>\%
  focus(n_missing)

# This is equivalent to
iris \%>\%
  skim() \%>\%
  dplyr::select(skim_variable, skim_type, n_missing)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim_obj.R
\name{skim-attr}
\alias{skim-attr}
\alias{data_rows}
\alias{data_cols}
\alias{df_name}
\alias{dt_key}
\alias{group_names}
\alias{base_skimmers}
\alias{skimmers_used}
\title{Functions for accessing skim_df attributes}
\usage{
data_rows(object)

data_cols(object)

df_name(object)

dt_key(object)

group_names(object)

base_skimmers(object)

skimmers_used(object)
}
\arguments{
\item{object}{A \code{skim_df} or \code{skim_list}.}
}
\value{
Data contained within the requested \code{skimr} attribute.
}
\description{
These functions simplify access to attributes contained within a \code{skim_df}.
While all attributes are read-only, being able to extract this information
is useful for different analyses. These functions should always be preferred
over calling base R's attribute functions.
}
\section{Functions}{
\itemize{
\item \code{data_rows()}: Get the number of rows in the skimmed data frame.

\item \code{data_cols()}: Get the number of columns in the skimmed data frame.

\item \code{df_name()}: Get the name of the skimmed data frame. This is only
available in contexts where the name can be looked up. This is often not
the case within a pipeline.

\item \code{dt_key()}: Get the key of the skimmed data.table. This is only
available in contexts where \code{data} is of class \code{data.table}.

\item \code{group_names()}: Get the names of the groups in the original data frame.
Only available if the data was grouped. Otherwise, \code{NULL}.

\item \code{base_skimmers()}: Get the names of the base skimming functions used.

\item \code{skimmers_used()}: Get the names of the skimming functions used, separated
by data type.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats.R
\name{stats}
\alias{stats}
\alias{n_missing}
\alias{n_complete}
\alias{complete_rate}
\alias{n_whitespace}
\alias{sorted_count}
\alias{top_counts}
\alias{inline_hist}
\alias{n_empty}
\alias{min_char}
\alias{max_char}
\alias{n_unique}
\alias{ts_start}
\alias{ts_end}
\alias{inline_linegraph}
\alias{list_lengths_min}
\alias{list_lengths_median}
\alias{list_lengths_max}
\alias{list_min_length}
\alias{list_max_length}
\title{Summary statistic functions}
\usage{
n_missing(x)

n_complete(x)

complete_rate(x)

n_whitespace(x)

sorted_count(x)

top_counts(x, max_char = 3, max_levels = 4)

inline_hist(x, n_bins = 8)

n_empty(x)

min_char(x)

max_char(x)

n_unique(x)

ts_start(x)

ts_end(x)

inline_linegraph(x, length.out = 16)

list_lengths_min(x)

list_lengths_median(x)

list_lengths_max(x)

list_min_length(x)

list_max_length(x)
}
\arguments{
\item{x}{A vector}

\item{max_char}{In \code{top} = 3, max_levels = 4}

\item{max_levels}{The maximum number of levels to be displayed.}

\item{n_bins}{In \code{inline_hist}, the number of histogram bars.}

\item{length.out}{In \code{inline_linegraph}, the length of the character time
series.}
}
\description{
\code{skimr} provides extensions to a variety of functions with R's stats package
to simplify creating summaries of data. All functions are vectorized over the
first argument. Additional arguments should be set in the \code{\link[=sfl]{sfl()}} that sets
the appropriate skimmers for a data type. You can use these, along with other
vectorized R functions, for creating custom sets of summary functions for
a given data type.
}
\section{Functions}{
\itemize{
\item \code{n_missing()}: Calculate the sum of \code{NA} and \code{NULL} (i.e. missing) values.

\item \code{n_complete()}: Calculate the sum of not \code{NA} and \code{NULL} (i.e. missing)
values.

\item \code{complete_rate()}: Calculate complete values; complete values are not missing.

\item \code{n_whitespace()}: Calculate the number of rows containing only whitespace
values using s+ regex.

\item \code{sorted_count()}: Create a contingency table and arrange its levels in
descending order. In case of ties, the ordering of results is alphabetical
and depends upon the locale. \code{NA} is treated as a ordinary value for
sorting.

\item \code{top_counts()}: Compute and collapse a contingency table into a single
character scalar. Wraps \code{\link[=sorted_count]{sorted_count()}}.

\item \code{inline_hist()}: Generate inline histogram for numeric variables. The
character length of the histogram is controlled by the formatting options
for character vectors.

\item \code{n_empty()}: Calculate the number of blank values in a character vector.
A "blank" is equal to "".

\item \code{min_char()}: Calculate the minimum number of characters within a
character vector.

\item \code{max_char()}: Calculate the maximum number of characters within a
character vector.

\item \code{n_unique()}: Calculate the number of unique elements but remove \code{NA}.

\item \code{ts_start()}: Get the start for a time series without the frequency.

\item \code{ts_end()}: Get the finish for a time series without the frequency.

\item \code{inline_linegraph()}: Generate inline line graph for time series variables. The
character length of the line graph is controlled by the formatting options
for character vectors.
Based on the function in the pillar package.

\item \code{list_lengths_min()}: Get the length of the shortest list in a vector of lists.

\item \code{list_lengths_median()}: Get the median length of the lists.

\item \code{list_lengths_max()}: Get the maximum length of the lists.

\item \code{list_min_length()}: Get the length of the shortest list in a vector of lists.

\item \code{list_max_length()}: Get the length of the longest list in a vector of lists.

}}
\seealso{
\code{\link[=get_skimmers]{get_skimmers()}} for customizing the functions called by \code{\link[=skim]{skim()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim_obj.R
\name{skim-obj}
\alias{skim-obj}
\alias{has_type_column}
\alias{has_variable_column}
\alias{has_skimr_attributes}
\alias{has_skim_type_attribute}
\alias{has_skimmers}
\alias{is_data_frame}
\alias{is_skim_df}
\alias{is_one_skim_df}
\alias{is_skim_list}
\alias{could_be_skim_df}
\alias{assert_is_skim_df}
\alias{assert_is_skim_list}
\alias{assert_is_one_skim_df}
\title{Test if an object is compatible with \code{skimr}}
\usage{
has_type_column(object)

has_variable_column(object)

has_skimr_attributes(object)

has_skim_type_attribute(object)

has_skimmers(object)

is_data_frame(object)

is_skim_df(object)

is_one_skim_df(object)

is_skim_list(object)

could_be_skim_df(object)

assert_is_skim_df(object)

assert_is_skim_list(object)

assert_is_one_skim_df(object)
}
\arguments{
\item{object}{Any \code{R} object.}
}
\description{
Objects within \code{skimr} are identified by a class, but they require additional
attributes and data columns for all operations to succeed. These checks help
ensure this. While they have some application externally, they are mostly
used internally.
}
\details{
Most notably, a \code{skim_df} has columns \code{skim_type} and \code{skim_variable}. And
has the following special attributes
\itemize{
\item \code{data_rows}: n rows in the original data
\item \code{data_cols}: original number of columns
\item \code{df_name}: name of the original data frame
\item \code{dt_key}: name of the key if original is a data.table
\item \code{groups}: if there were group variables
\item \code{base_skimmers}: names of functions applied to all skim types
\item \code{skimmers_used}: names of functions used to skim each type
}

The functions in these checks work like \code{\link[=all.equal]{all.equal()}}. The return \code{TRUE} if
the check passes, or otherwise notifies why the check failed. This makes them
more useful when throwing errors.
}
\section{Functions}{
\itemize{
\item \code{has_type_column()}: Does the object have the \code{skim_type} column?

\item \code{has_variable_column()}: Does the object have the \code{skim_variable} column?

\item \code{has_skimr_attributes()}: Does the object have the appropriate \code{skimr} attributes?

\item \code{has_skim_type_attribute()}: Does the object have a \code{skim_type} attribute? This makes
it a \code{one_skim_df}.

\item \code{has_skimmers()}: Does the object have skimmers?

\item \code{is_data_frame()}: Is the object a data frame?

\item \code{is_skim_df()}: Is the object a \code{skim_df}?

\item \code{is_one_skim_df()}: Is the object a \code{one_skim_df}? This is similar to a
\code{skim_df}, but does not have the \code{type} column. That is stored as an
attribute instead.

\item \code{is_skim_list()}: Is the object a \code{skim_list}?

\item \code{could_be_skim_df()}: Is this a data frame with \code{skim_variable} and
\code{skim_type} columns?

\item \code{assert_is_skim_df()}: Stop if the object is not a \code{skim_df}.

\item \code{assert_is_skim_list()}: Stop if the object is not a \code{skim_list}.

\item \code{assert_is_one_skim_df()}: Stop if the object is not a \code{one_skim_df}.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim_print.R
\name{knit_print}
\alias{knit_print}
\alias{knit_print.skim_df}
\alias{knit_print.skim_list}
\alias{knit_print.one_skim_df}
\alias{knit_print.summary_skim_df}
\title{Provide a default printing method for knitr.}
\usage{
\method{knit_print}{skim_df}(x, options = NULL, ...)

\method{knit_print}{skim_list}(x, options = NULL, ...)

\method{knit_print}{one_skim_df}(x, options = NULL, ...)

\method{knit_print}{summary_skim_df}(x, options = NULL, ...)
}
\arguments{
\item{x}{An R object to be printed}

\item{options}{Options passed into the print function.}

\item{...}{Additional arguments passed to the S3 method. Currently ignored,
except two optional arguments \code{options} and \code{inline}; see
the references below.}
}
\value{
A \code{knit_asis} object. Which is used by \code{knitr} when rendered.
}
\description{
Instead of standard R output, \code{knitr} and \code{RMarkdown} documents will have
formatted \code{\link[knitr:kable]{knitr::kable()}} output on return. You can disable this by setting
the chunk option \code{render = normal_print}.
}
\details{
The summary statistics for the original data frame can be disabled by setting
the \code{knitr} chunk option \code{skimr_include_summary = FALSE}. See
\link[knitr:opts_chunk]{knitr::opts_chunk} for more information. You can change the number of digits
shown in the printed table with the \code{skimr_digits} chunk option.

Alternatively, you can call \code{\link[=yank]{yank()}} to get the particular
\code{skim_df} objects and format them however you like. One warning though.
Because histograms contain unicode characters, they can have unexpected
print results, as R as varying levels of unicode support. This affects
Windows users most commonly. Call \code{vignette("Using_fonts")} for more details.
}
\section{Methods (by class)}{
\itemize{
\item \code{knit_print(skim_df)}: Default \code{knitr} print for \code{skim_df} objects.

\item \code{knit_print(skim_list)}: Default \code{knitr} print for a \code{skim_list}.

\item \code{knit_print(one_skim_df)}: Default \code{knitr} print within a partitioned \code{skim_df}.

\item \code{knit_print(summary_skim_df)}: Default \code{knitr} print for \code{skim_df} summaries.

}}
\seealso{
\code{\link[knitr:kable]{knitr::kable()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.skim_df}
\alias{summary.skim_df}
\title{Summary function for skim_df}
\usage{
\method{summary}{skim_df}(object, ...)
}
\arguments{
\item{object}{a skim dataframe.}

\item{...}{Additional arguments affecting the summary produced. Not used.}
}
\value{
A summary of the skim data frame.
}
\description{
This is a method of the generic function \code{\link[=summary]{summary()}}.
}
\examples{
a <- skim(mtcars)
summary(a)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim_print.R
\name{knit_print}
\alias{knit_print}
\alias{knit_print.skim_df}
\alias{knit_print.skim_list}
\alias{knit_print.one_skim_df}
\alias{knit_print.summary_skim_df}
\title{Provide a default printing method for knitr.}
\usage{
\method{knit_print}{skim_df}(x, options = NULL, ...)

\method{knit_print}{skim_list}(x, options = NULL, ...)

\method{knit_print}{one_skim_df}(x, options = NULL, ...)

\method{knit_print}{summary_skim_df}(x, options = NULL, ...)
}
\arguments{
\item{x}{An R object to be printed}

\item{options}{Options passed into the print function.}

\item{...}{Additional arguments passed to the S3 method. Currently ignored,
except two optional arguments \code{options} and \code{inline}; see
the references below.}
}
\value{
A \code{knit_asis} object. Which is used by \code{knitr} when rendered.
}
\description{
Instead of standard R output, \code{knitr} and \code{RMarkdown} documents will have
formatted \code{\link[knitr:kable]{knitr::kable()}} output on return. You can disable this by setting
the chunk option \code{render = normal_print}.
}
\details{
The summary statistics for the original data frame can be disabled by setting
the \code{knitr} chunk option \code{skimr_include_summary = FALSE}. See
\link[knitr:opts_chunk]{knitr::opts_chunk} for more information. You can change the number of digits
shown in the printed table with the \code{skimr_digits} chunk option.

Alternatively, you can call \code{\link[=collapse]{collapse()}} or \code{\link[=yank]{yank()}} to get the particular
\code{skim_df} objects and format them however you like. One warning though.
Because histograms contain unicode characters, they can have unexpected
print results, as R as varying levels of unicode support. This affects
Windows users most commonly. Call \code{vignette("Using_fonts")} for more details.
}
\section{Methods (by class)}{
\itemize{
\item \code{knit_print(skim_df)}: Default \code{knitr} print for \code{skim_df} objects.

\item \code{knit_print(skim_list)}: Default \code{knitr} print for a \code{skim_list}.

\item \code{knit_print(one_skim_df)}: Default \code{knitr} print within a partitioned \code{skim_df}.

\item \code{knit_print(summary_skim_df)}: Default \code{knitr} print for \code{skim_df} summaries.

}}
\seealso{
\code{\link[knitr:kable]{knitr::kable()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_skimmers.R
\name{get_skimmers}
\alias{get_skimmers}
\alias{get_skimmers.default}
\alias{get_skimmers.numeric}
\alias{get_skimmers.factor}
\alias{get_skimmers.character}
\alias{get_skimmers.logical}
\alias{get_skimmers.complex}
\alias{get_skimmers.Date}
\alias{get_skimmers.POSIXct}
\alias{get_skimmers.difftime}
\alias{get_skimmers.Timespan}
\alias{get_skimmers.ts}
\alias{get_skimmers.list}
\alias{get_skimmers.AsIs}
\alias{get_skimmers.haven_labelled}
\alias{modify_default_skimmers}
\title{Retrieve the summary functions for a specific data type}
\usage{
get_skimmers(column)

\method{get_skimmers}{default}(column)

\method{get_skimmers}{numeric}(column)

\method{get_skimmers}{factor}(column)

\method{get_skimmers}{character}(column)

\method{get_skimmers}{logical}(column)

\method{get_skimmers}{complex}(column)

\method{get_skimmers}{Date}(column)

\method{get_skimmers}{POSIXct}(column)

\method{get_skimmers}{difftime}(column)

\method{get_skimmers}{Timespan}(column)

\method{get_skimmers}{ts}(column)

\method{get_skimmers}{list}(column)

\method{get_skimmers}{AsIs}(column)

\method{get_skimmers}{haven_labelled}(column)

modify_default_skimmers(skim_type, new_skim_type = NULL, new_funs = list())
}
\arguments{
\item{column}{An atomic vector or list. A column from a data frame.}

\item{skim_type}{A character scalar. The class of the object with default
skimmers.}

\item{new_skim_type}{The type to assign to the looked up set of skimmers.}

\item{new_funs}{Replacement functions for those in}
}
\value{
A \code{skim_function_list} object.
}
\description{
These functions are used to set the default skimming functions for a data
type. They are combined with the base skim function list (\code{sfl}) in
\code{\link[=skim_with]{skim_with()}}, to create the summary tibble for each type.
}
\details{
When creating your own set of skimming functions, call \code{\link[=sfl]{sfl()}} within a
\code{\link[=get_skimmers]{get_skimmers()}} method for your particular type. Your call to \code{\link[=sfl]{sfl()}} should
also provide a matching class in the \code{skim_type} argument.  Otherwise, it
will not be possible to dynamically reassign your default functions when
working interactively.

Call \code{\link[=get_default_skimmers]{get_default_skimmers()}} to see the functions for each type of summary
function currently supported. Call \code{\link[=get_default_skimmer_names]{get_default_skimmer_names()}} to just see
the names of these functions. Use \code{\link[=modify_default_skimmers]{modify_default_skimmers()}} for a method
for changing the \code{skim_type} or functions for a default \code{sfl}. This is useful
for creating new default \code{sfl}'s.
}
\section{Methods (by class)}{
\itemize{
\item \code{get_skimmers(default)}: The default method for skimming data. Only used when
a column's data type doesn't match currently installed types. Call
\link{get_default_skimmer_names} to see these defaults.

\item \code{get_skimmers(numeric)}: Summary functions for numeric columns, covering both
\code{\link[=double]{double()}} and \code{\link[=integer]{integer()}} classes: \code{\link[=mean]{mean()}}, \code{\link[=sd]{sd()}}, \code{\link[=quantile]{quantile()}} and
\code{\link[=inline_hist]{inline_hist()}}.

\item \code{get_skimmers(factor)}: Summary functions for factor columns:
\code{\link[=is.ordered]{is.ordered()}}, \code{\link[=n_unique]{n_unique()}} and \code{\link[=top_counts]{top_counts()}}.

\item \code{get_skimmers(character)}: Summary functions for character columns. Also, the
default for unknown columns: \code{\link[=min_char]{min_char()}}, \code{\link[=max_char]{max_char()}}, \code{\link[=n_empty]{n_empty()}},
\code{\link[=n_unique]{n_unique()}} and \code{\link[=n_whitespace]{n_whitespace()}}.

\item \code{get_skimmers(logical)}: Summary functions for logical/ boolean columns:
\code{\link[=mean]{mean()}}, which produces rates for each value, and \code{\link[=top_counts]{top_counts()}}.

\item \code{get_skimmers(complex)}: Summary functions for complex columns: \code{\link[=mean]{mean()}}.

\item \code{get_skimmers(Date)}: Summary functions for \code{Date} columns: \code{\link[=min]{min()}},
\code{\link[=max]{max()}}, \code{\link[=median]{median()}} and \code{\link[=n_unique]{n_unique()}}.

\item \code{get_skimmers(POSIXct)}: Summary functions for \code{POSIXct} columns: \code{\link[=min]{min()}},
\code{\link[=max]{max()}}, \code{\link[=median]{median()}} and \code{\link[=n_unique]{n_unique()}}.

\item \code{get_skimmers(difftime)}: Summary functions for \code{difftime} columns: \code{\link[=min]{min()}},
\code{\link[=max]{max()}}, \code{\link[=median]{median()}} and \code{\link[=n_unique]{n_unique()}}.

\item \code{get_skimmers(Timespan)}: Summary functions for \code{Timespan} columns: \code{\link[=min]{min()}},
\code{\link[=max]{max()}}, \code{\link[=median]{median()}} and \code{\link[=n_unique]{n_unique()}}.

\item \code{get_skimmers(ts)}: Summary functions for \code{ts} columns: \code{\link[=min]{min()}},
\code{\link[=max]{max()}}, \code{\link[=median]{median()}} and \code{\link[=n_unique]{n_unique()}}.

\item \code{get_skimmers(list)}: Summary functions for \code{list} columns: \code{\link[=n_unique]{n_unique()}},
\code{\link[=list_min_length]{list_min_length()}} and \code{\link[=list_max_length]{list_max_length()}}.

\item \code{get_skimmers(AsIs)}: Summary functions for \code{AsIs} columns: \code{\link[=n_unique]{n_unique()}},
\code{\link[=list_min_length]{list_min_length()}} and \code{\link[=list_max_length]{list_max_length()}}.

\item \code{get_skimmers(haven_labelled)}: Summary functions for \code{haven_labelled} columns.
Finds the appropriate skimmers for the underlying data in the vector.

}}
\examples{
# Defining default skimming functions for a new class, `my_class`.
# Note that the class argument is required for dynamic reassignment.
get_skimmers.my_class <- function(column) {
  sfl(
    skim_type = "my_class",
    mean,
    sd
  )
}

# Integer and double columns are both "numeric" and are treated the same
# by default. To switch this behavior in another package, add a method.
get_skimmers.integer <- function(column) {
  sfl(
    skim_type = "integer",
    p50 = ~ stats::quantile(
      .,
      probs = .50, na.rm = TRUE, names = FALSE, type = 1
    )
  )
}
x <- mtcars[c("gear", "carb")]
class(x$carb) <- "integer"
skim(x)
\dontrun{
# In a package, to revert to the V1 behavior of skimming separately with the
# same functions, assign the numeric `get_skimmers`.
get_skimmers.integer <- skimr::get_skimmers.numeric

# Or, in a local session, use `skim_with` to create a different `skim`.
new_skim <- skim_with(integer = skimr::get_skimmers.numeric())

# To apply a set of skimmers from an old type to a new type
get_skimmers.new_type <- function(column) {
  modify_default_skimmers("old_type", new_skim_type = "new_type")
}
}
}
\seealso{
\code{\link[=sfl]{sfl()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fix_windows_histograms}
\alias{fix_windows_histograms}
\title{Fix unicode histograms on Windows}
\usage{
fix_windows_histograms()
}
\description{
This functions changes your session's locale to address issues with printing
histograms on Windows on versions of R below 4.2.1.
}
\details{
There are known issues with printing the spark-histogram characters when
printing a data frame, appearing like this: "<U+2582><U+2585><U+2587>".
This longstanding problem originates in the low-level code for printing
dataframes. This was addressed in R version 4.2.1.
}
\seealso{
\code{\link[=skim_without_charts]{skim_without_charts()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim_print.R
\name{repr}
\alias{repr}
\alias{repr_text.skim_df}
\alias{repr_text.skim_list}
\alias{repr_text.one_skim_df}
\title{Skimr printing within Jupyter notebooks}
\usage{
\method{repr_text}{skim_df}(obj, ...)

\method{repr_text}{skim_list}(obj, ...)

\method{repr_text}{one_skim_df}(obj, ...)
}
\arguments{
\item{obj}{The object to \link{print} and then return the output.}

\item{...}{ignored.}
}
\value{
None. \code{invisible(NULL)}.
}
\description{
This reproduces printed results in the console. By default Jupyter kernels
render the final object in the cell. We want the version printed by
\code{skimr} instead of the data that it contains.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_skimmers.R
\name{get_default_skimmers}
\alias{get_default_skimmers}
\alias{get_one_default_skimmer}
\alias{get_default_skimmer_names}
\alias{get_one_default_skimmer_names}
\alias{get_sfl}
\title{View default skimmer names and functions}
\usage{
get_default_skimmers(skim_type = NULL)

get_one_default_skimmer(skim_type)

get_default_skimmer_names(skim_type = NULL)

get_one_default_skimmer_names(skim_type)

get_sfl(skim_type)
}
\arguments{
\item{skim_type}{The class of the column being skimmed.}
}
\description{
These utility functions look up the currently-available defaults for one or
more \code{skim_type}'s. They work with all defaults in the \code{skimr} package, as
well as the defaults in any package that extends \code{skimr}. See
\code{\link[=get_skimmers]{get_skimmers()}} for writing lookup methods for different.
}
\details{
The functions differ in return type and whether or not the result is in
a list. \code{\link[=get_default_skimmers]{get_default_skimmers()}} and \code{\link[=get_one_default_skimmer]{get_one_default_skimmer()}} return
functions. The former returns functions within a typed list, i.e.
\code{list(numeric = list(...functions...))}.

The functions differ in return type and whether or not the result is in
a list. \code{\link[=get_default_skimmer_names]{get_default_skimmer_names()}} and \code{\link[=get_one_default_skimmer_names]{get_one_default_skimmer_names()}}
return a list of character vectors or a single character vector.

\code{\link[=get_sfl]{get_sfl()}} returns the skimmer function list for a particular \code{skim_type}.
It differs from \code{\link[=get_default_skimmers]{get_default_skimmers()}} in that the returned \code{sfl} contains
a list of functions and a \code{skim_type}.
}
\section{Functions}{
\itemize{
\item \code{get_one_default_skimmer()}: Get the functions associated with one
\code{skim_type}.

\item \code{get_default_skimmer_names()}: Get the names of the functions used in one
or more \code{skim_type}'s.

\item \code{get_one_default_skimmer_names()}: Get the names of the functions used in one
\code{skim_type}.

\item \code{get_sfl()}: Get the \code{sfl} for a \code{skim_type}.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reshape.R
\name{to_long}
\alias{to_long}
\alias{to_long.default}
\alias{to_long.skim_df}
\title{Create "long" skim output}
\usage{
to_long(.data, ..., skim_fun = skim)

\method{to_long}{default}(.data, ..., skim_fun = skim)

\method{to_long}{skim_df}(.data, ..., skim_fun = skim)
}
\arguments{
\item{.data}{A data frame or an object that can be coerced into a data frame.}

\item{...}{Columns to select for skimming. When none are provided, the
default is to skim all columns.}

\item{skim_fun}{The skim function used.}
}
\value{
A tibble
}
\description{
Skim results returned as a tidy long data frame with four columns:
variable, type, stat and formatted.
}
\section{Methods (by class)}{
\itemize{
\item \code{to_long(default)}: Skim a data frame and convert the results to a
long data frame.

\item \code{to_long(skim_df)}: Transform a skim_df to a long data frame.

}}
\examples{
to_long(iris)
to_long(skim(iris))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vctrs.R
\name{skimr-vctrs}
\alias{skimr-vctrs}
\alias{vec_ptype2.skim_df.skim_df}
\alias{vec_ptype2.skim_df.tbl_df}
\alias{vec_ptype2.tbl_df.skim_df}
\alias{vec_cast.skim_df.skim_df}
\alias{vec_cast.skim_df.tbl_df}
\alias{vec_cast.tbl_df.skim_df}
\title{Functions for working with the vctrs package}
\usage{
\method{vec_ptype2}{skim_df.skim_df}(x, y, ...)

\method{vec_ptype2}{skim_df.tbl_df}(x, y, ...)

\method{vec_ptype2}{tbl_df.skim_df}(x, y, ...)

\method{vec_cast}{skim_df.skim_df}(x, to, ...)

\method{vec_cast}{skim_df.tbl_df}(x, to, ...)

\method{vec_cast}{tbl_df.skim_df}(x, to, ...)
}
\description{
These make it clear that we need to use the tibble behavior when joining,
concatenating or casting \code{skim_df} objects. For a better discussion, on
why this is important and how these functions work, see:
\url{https://vctrs.r-lib.org/reference/howto-faq-coercion-data-frame.html}.
}
\details{
\verb{vec_ptype2.*} handles finding common prototypes between \code{skim_df} and
similar objects. \verb{vec_cast.*} handles casting between objects. Note that
as of \verb{dplyr 1.0.2}, \code{\link[dplyr:bind_rows]{dplyr::bind_rows()}} does not full support combining
attributes and \code{\link[vctrs:vec_bind]{vctrs::vec_rbind()}} is preferred instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reshape.R
\name{partition}
\alias{partition}
\alias{bind}
\alias{yank}
\title{Separate a big \code{skim_df} into smaller data frames, by type.}
\usage{
partition(data)

bind(data)

yank(data, skim_type)
}
\arguments{
\item{data}{A \code{skim_df}.}

\item{skim_type}{A character scalar. The subtable to extract from a
\code{skim_df}.}
}
\value{
A \code{skim_list} of \code{skim_df}'s, by type.
}
\description{
The data frames produced by \code{\link[=skim]{skim()}} are wide and sparse, filled with
columns that are mostly \code{NA}. For that reason, it can be convenient to
work with "by type" subsets of the original data frame. These smaller
subsets have their \code{NA} columns removed.
}
\details{
\code{partition()} creates a list of smaller \code{skim_df} data frames. Each entry
in the list is a data type from the original \code{skim_df}. The inverse of
\code{partition()} is \code{bind()}, which takes the list and produces the original
\code{skim_df}. While \code{partition()} keeps all of the subtables as list entries,
\code{yank()} gives you a single subtable for a data type.
}
\section{Functions}{
\itemize{
\item \code{bind()}: The inverse of a \code{partition()}. Rebuild the original
\code{skim_df}.

\item \code{yank()}: Extract a subtable from a \code{skim_df} with a particular
type.

}}
\examples{
# Create a wide skimmed data frame (a skim_df)
skimmed <- skim(iris)

# Separate into a list of subtables by type
separate <- partition(skimmed)

# Put back together
identical(bind(separate), skimmed)
# > TRUE

# Alternatively, get the subtable of a particular type
yank(skimmed, "factor")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfl.R
\name{sfl}
\alias{sfl}
\title{Create a skimr function list}
\usage{
sfl(..., skim_type = "")
}
\arguments{
\item{...}{Inherited from dplyr::data_masking() for dplyr version 1 or later
or dplyr::funs() for older versions of dplyr.
A list of functions
specified by:
\itemize{
\item Their name, \code{"mean"}
\item The function itself, \code{mean}
\item A call to the function with \code{.} as a dummy argument,
\code{mean(., na.rm = TRUE)}
\item An anonymous function in \pkg{purrr} notation, \code{~mean(., na.rm = TRUE)}
}}

\item{skim_type}{A character scalar. This is used to match locally-provided
skimmers with defaults. See \code{\link[=get_skimmers]{get_skimmers()}} for more detail.}
}
\value{
A \code{skimr_function_list}, which contains a list of \code{fun_calls},
returned by \code{dplyr::funs()} and a list of skimming functions to drop.
}
\description{
This constructor is used to create a named list of functions. It also you
also pass \code{NULL} to identify a skimming function that you wish to remove.
Only functions that return a single value, working with \code{\link[dplyr:summarise]{dplyr::summarize()}},
can be used within \code{sfl}.
}
\details{
\code{sfl()} will automatically generate callables and names for a variety of
inputs, including functions, formulas and strings. Nonetheless, we recommend
providing names when reasonable to get better \code{\link[=skim]{skim()}} output.
}
\examples{
# sfl's can take a variety of input formats and will generate names
# if not provided.
sfl(mad, "var", ~ length(.)^2)

# But these can generate unpredictable names in your output.
# Better to set your own names.
sfl(mad = mad, variance = "var", length_sq = ~ length(.)^2)

# sfl's can remove individual skimmers from defaults by passing NULL.
sfl(hist = NULL)

# When working interactively, you don't need to set a type.
# But you should when defining new defaults with `get_skimmers()`.
get_skimmers.my_new_class <- function(column) {
  sfl(n_missing, skim_type = "my_new_class")
}
}
\seealso{
\code{\link[dplyr:funs]{dplyr::funs()}}, \code{\link[=skim_with]{skim_with()}} and \code{\link[=get_skimmers]{get_skimmers()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vctrs.R
\name{skimr-vctrs}
\alias{skimr-vctrs}
\alias{vec_ptype2.skim_df.skim_df}
\alias{vec_ptype2.skim_df.tbl_df}
\alias{vec_ptype2.tbl_df.skim_df}
\alias{vec_cast.skim_df.skim_df}
\alias{vec_cast.skim_df.tbl_df}
\alias{vec_cast.tbl_df.skim_df}
\title{Functions for working with the vctrs package}
\usage{
\method{vec_ptype2}{skim_df.skim_df}(x, y, ...)

\method{vec_ptype2}{skim_df.tbl_df}(x, y, ...)

\method{vec_ptype2}{tbl_df.skim_df}(x, y, ...)

\method{vec_cast}{skim_df.skim_df}(x, to, ...)

\method{vec_cast}{skim_df.tbl_df}(x, to, ...)

\method{vec_cast}{tbl_df.skim_df}(x, to, ...)
}
\description{
These make it clear that we need to use the tibble behavior when joining,
concatenating or casting \code{skim_df} objects. For a better discussion, on
why this is important and how these functions work, see:
\url{https://vctrs.r-lib.org/reference/howto-faq-coercion-data-frame.html}.
}
\details{
\verb{vec_ptype2.*} handles finding common prototypes between \code{skim_df} and
similar objects. \verb{vec_cast.*} handles casting between objects. Note that
as of \verb{dplyr 1.0.2}, \code{\link[dplyr:bind]{dplyr::bind_rows()}} does not full support combining
attributes and \code{\link[vctrs:vec_bind]{vctrs::vec_rbind()}} is preferred instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fix_windows_histograms}
\alias{fix_windows_histograms}
\title{Fix unicode histograms on Windows}
\usage{
fix_windows_histograms()
}
\description{
This functions changes your session's locale to address issues with printing
histograms on Windows.
}
\details{
There are known issues with printing the spark-histogram characters when
printing a data frame, appearing like this: "<U+2582><U+2585><U+2587>".
This longstanding problem originates in the low-level code for printing
dataframes.
}
\seealso{
\code{\link[=skim_without_charts]{skim_without_charts()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim_with.R
\name{skim_with}
\alias{skim_with}
\title{Set or add the summary functions for a particular type of data}
\usage{
skim_with(
  ...,
  base = sfl(n_missing = n_missing, complete_rate = complete_rate),
  append = TRUE
)
}
\arguments{
\item{...}{One or more (\code{sfl}) \code{skimmer_function_list} objects, with an
argument name that matches a particular data type.}

\item{base}{An \code{sfl} that sets skimmers for all column types.}

\item{append}{Whether the provided options should be in addition to the
defaults already in \code{skim}. Default is \code{TRUE}.}
}
\value{
A new \code{skim()} function. This is callable. See \code{\link[=skim]{skim()}} for more
details.
}
\description{
While skim is designed around having an opinionated set of defaults, you
can use this function to change the summary statistics that it returns.
}
\details{
\code{skim_with()} is a closure: a function that returns a new function. This
lets you have several skimming functions in a single R session, but it
also means that you need to assign the return of \code{skim_with()} before
you can use it.

You assign values within \code{skim_with} by using the \code{\link[=sfl]{sfl()}} helper (\code{skimr}
function list). This helper behaves mostly like \code{\link[dplyr:funs]{dplyr::funs()}}, but lets
you also identify which skimming functions you want to remove, by setting
them to \code{NULL}. Assign an \code{sfl} to each column type that you wish to modify.

Functions that summarize all data types, and always return the same type
of value, can be assigned to the \code{base} argument. The default base skimmers
compute the number of missing values \code{\link[=n_missing]{n_missing()}} and the rate of values
being complete, i.e. not missing, \code{\link[=complete_rate]{complete_rate()}}.

When \code{append = TRUE} and local skimmers have names matching the names of
entries in the default \code{skim_function_list}, the values in the default list
are overwritten. Similarly, if \code{NULL} values are passed within \code{sfl()}, these
default skimmers are dropped. Otherwise, if \code{append = FALSE}, only the
locally-provided skimming functions are used.

Note that \code{append} only applies to the \code{typed} skimmers (i.e. non-base).
See \code{\link[=get_default_skimmer_names]{get_default_skimmer_names()}} for a list of defaults.
}
\examples{
# Use new functions for numeric functions. If you don't provide a name,
# one will be automatically generated.
my_skim <- skim_with(numeric = sfl(median, mad), append = FALSE)
my_skim(faithful)

# If you want to remove a particular skimmer, set it to NULL
# This removes the inline histogram
my_skim <- skim_with(numeric = sfl(hist = NULL))
my_skim(faithful)

# This works with multiple skimmers. Just match names to overwrite
my_skim <- skim_with(numeric = sfl(iqr = IQR, p25 = NULL, p75 = NULL))
my_skim(faithful)

# Alternatively, set `append = FALSE` to replace the skimmers of a type.
my_skim <- skim_with(numeric = sfl(mean = mean, sd = sd), append = FALSE)

# Skimmers are unary functions. Partially apply arguments during assigment.
# For example, you might want to remove NA values.
my_skim <- skim_with(numeric = sfl(iqr = ~ IQR(., na.rm = TRUE)))

# Set multiple types of skimmers simultaneously.
my_skim <- skim_with(numeric = sfl(mean), character = sfl(length))

# Or pass the same as a list, unquoting the input.
my_skimmers <- list(numeric = sfl(mean), character = sfl(length))
my_skim <- skim_with(!!!my_skimmers)

# Use the v1 base skimmers instead.
my_skim <- skim_with(base = sfl(
  missing = n_missing,
  complete = n_complete,
  n = length
))

# Remove the base skimmers entirely
my_skim <- skim_with(base = NULL)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim_print.R
\name{print}
\alias{print}
\alias{print.skim_df}
\alias{print.skim_list}
\alias{print.summary_skim_df}
\title{Print \code{skim} objects}
\usage{
\method{print}{skim_df}(
  x,
  include_summary = TRUE,
  n = Inf,
  width = Inf,
  summary_rule_width = getOption("skimr_summary_rule_width", default = 40),
  ...
)

\method{print}{skim_list}(x, n = Inf, width = Inf, ...)

\method{print}{summary_skim_df}(x, .summary_rule_width = 40, ...)
}
\arguments{
\item{x}{Object to format or print.}

\item{include_summary}{Whether a summary of the data frame should be printed}

\item{n}{Number of rows to show. If \code{NULL}, the default, will print all rows
if less than the \code{print_max} \link[pillar:pillar_options]{option}.
Otherwise, will print as many rows as specified by the
\code{print_min} \link[pillar:pillar_options]{option}.}

\item{width}{Width of text output to generate. This defaults to \code{NULL}, which
means use the \code{width} \link[pillar:pillar_options]{option}.}

\item{summary_rule_width}{Width of Data Summary cli rule, defaults to 40.}

\item{...}{Passed on to \link[pillar:tbl_format_setup]{pillar::tbl_format_setup()}}

\item{.summary_rule_width}{the width for the main rule above the summary.}
}
\description{
\code{skimr} has custom print methods for all supported objects. Default printing
methods for \code{knitr}/ \code{rmarkdown} documents is also provided.
}
\section{Methods (by class)}{
\itemize{
\item \code{print(skim_df)}: Print a skimmed data frame (\code{skim_df} from \code{\link[=skim]{skim()}}).

\item \code{print(skim_list)}: Print a \code{skim_list}, a list of \code{skim_df} objects.

\item \code{print(summary_skim_df)}: Print method for a \code{summary_skim_df} object.

}}
\section{Printing options}{


For better or for worse, \code{skimr} often produces more output than can fit in
the standard R console. Fortunately, most modern environments like RStudio
and Jupyter support more than 80 character outputs. Call
\code{options(width = 90)} to get a better experience with \code{skimr}.

The print methods in \code{skimr} wrap those in the \link[tibble:formatting]{tibble}
package. You can control printing behavior using the same global options.
}

\section{Behavior in \code{dplyr} pipelines}{


Printing a \code{skim_df} requires specific columns that might be dropped when
using \code{\link[dplyr:select]{dplyr::select()}} or \code{\link[dplyr:summarise]{dplyr::summarize()}} on a \code{skim_df}. In those
cases, this method falls back to \code{\link[tibble:formatting]{tibble::print.tbl()}}.
}

\section{Options for controlling print behavior}{


You can control the width rule line for the printed subtables with an option:
\code{skimr_table_header_width}.
}

\seealso{
\code{\link[tibble:trunc_mat]{tibble::trunc_mat()}} For a list of global options for customizing
print formatting. \code{\link[cli:num_ansi_colors]{cli::num_ansi_colors()}} for the variety of issues that
affect tibble's color support.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R, R/skimr-package.R
\docType{package}
\name{skimr-package}
\alias{skimr-package}
\title{skimr: Compact and Flexible Summaries of Data}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A simple to use summary function that can be used with pipes and displays nicely in the console. The default summary statistics may be modified by the user as can the default formatting. Support for data frames and vectors is included, and users can implement their own skim methods for specific object types as described in a vignette. Default summaries include support for inline spark graphs. Instructions for managing these on specific operating systems are given in the "Using skimr" vignette and the README.

\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A simple to use summary function that can be used with pipes and displays nicely in the console. The default summary statistics may be modified by the user as can the default formatting. Support for data frames and vectors is included, and users can implement their own skim methods for specific object types as described in a vignette. Default summaries include support for inline spark graphs. Instructions for managing these on specific operating systems are given in the "Using skimr" vignette and the README.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/skimr/}
  \item \url{https://github.com/ropensci/skimr/}
  \item Report bugs at \url{https://github.com/ropensci/skimr/issues}
}


Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/skimr/}
  \item \url{https://github.com/ropensci/skimr/}
  \item Report bugs at \url{https://github.com/ropensci/skimr/issues}
}

}
\author{
\strong{Maintainer}: Elin Waring \email{elin.waring@gmail.com}

Authors:
\itemize{
  \item Michael Quinn \email{msquinn@google.com}
  \item Amelia McNamara \email{amcnamara@smith.edu}
  \item Eduardo Arino de la Rubia \email{earino@gmail.com}
  \item Hao Zhu \email{haozhu233@gmail.com}
  \item Shannon Ellis \email{sellis18@jhmi.edu}
}

Other contributors:
\itemize{
  \item Julia Lowndes \email{lowndes@nceas.ucsb.edu} [contributor]
  \item Hope McLeod \email{hmgit2@gmail.com} [contributor]
  \item Hadley Wickham \email{hadley@rstudio.com} [contributor]
  \item Kirill Müller \email{krlmlr+r@mailbox.org} [contributor]
  \item  RStudio, Inc. (Spark functions) [copyright holder]
  \item Connor Kirkpatrick \email{hello@connorkirkpatrick.com} [contributor]
  \item Scott Brenstuhl \email{brenstsr@miamioh.edu} [contributor]
  \item Patrick Schratz \email{patrick.schratz@gmail.com} [contributor]
  \item lbusett \email{lbusett@gmail.com} [contributor]
  \item Mikko Korpela \email{mvkorpel@iki.fi} [contributor]
  \item Jennifer Thompson \email{thompson.jennifer@gmail.com} [contributor]
  \item Harris McGehee \email{mcgehee.harris@gmail.com} [contributor]
  \item Mark Roepke \email{mroepke5@gmail.com} [contributor]
  \item Patrick Kennedy \email{pkqstr@protonmail.com} [contributor]
  \item Daniel Possenriede \email{possenriede@gmail.com} [contributor]
  \item David Zimmermann \email{david_j_zimmermann@hotmail.com} [contributor]
  \item Kyle Butts \email{buttskyle96@gmail.com} [contributor]
  \item Bastian Torges \email{bastian.torges@gmail.com} [contributor]
  \item Rick Saporta \email{Rick@TheFarmersDog.com} [contributor]
  \item Henry Morgan Stewart \email{henry.morganstewart@gmail.com} [contributor]
  \item Olivier Roy [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim_print.R
\name{print}
\alias{print}
\alias{print.skim_df}
\alias{print.skim_list}
\alias{print.summary_skim_df}
\title{Print \code{skim} objects}
\usage{
\method{print}{skim_df}(
  x,
  include_summary = TRUE,
  n = Inf,
  width = Inf,
  summary_rule_width = getOption("skimr_summary_rule_width", default = 40),
  ...
)

\method{print}{skim_list}(x, n = Inf, width = Inf, ...)

\method{print}{summary_skim_df}(x, .summary_rule_width = 40, ...)
}
\arguments{
\item{x}{Object to format or print.}

\item{include_summary}{Whether a summary of the data frame should be printed}

\item{n}{Number of rows to show. If \code{NULL}, the default, will print all rows
if less than the \code{print_max} \link[pillar:pillar_options]{option}.
Otherwise, will print as many rows as specified by the
\code{print_min} \link[pillar:pillar_options]{option}.}

\item{width}{Width of text output to generate. This defaults to \code{NULL}, which
means use the \code{width} \link[pillar:pillar_options]{option}.}

\item{summary_rule_width}{Width of Data Summary cli rule, defaults to 40.}

\item{...}{Passed on to \code{\link[=tbl_format_setup]{tbl_format_setup()}}.}

\item{.summary_rule_width}{the width for the main rule above the summary.}
}
\description{
\code{skimr} has custom print methods for all supported objects. Default printing
methods for \code{knitr}/ \code{rmarkdown} documents is also provided.
}
\section{Methods (by class)}{
\itemize{
\item \code{print(skim_df)}: Print a skimmed data frame (\code{skim_df} from \code{\link[=skim]{skim()}}).

\item \code{print(skim_list)}: Print a \code{skim_list}, a list of \code{skim_df} objects.

\item \code{print(summary_skim_df)}: Print method for a \code{summary_skim_df} object.

}}
\section{Printing options}{


For better or for worse, \code{skimr} often produces more output than can fit in
the standard R console. Fortunately, most modern environments like RStudio
and Jupyter support more than 80 character outputs. Call
\code{options(width = 90)} to get a better experience with \code{skimr}.

The print methods in \code{skimr} wrap those in the \link[tibble:formatting]{tibble}
package. You can control printing behavior using the same global options.
}

\section{Behavior in \code{dplyr} pipelines}{


Printing a \code{skim_df} requires specific columns that might be dropped when
using \code{\link[dplyr:select]{dplyr::select()}} or \code{\link[dplyr:summarise]{dplyr::summarize()}} on a \code{skim_df}. In those
cases, this method falls back to \code{\link[tibble:formatting]{tibble::print.tbl()}}.
}

\section{Options for controlling print behavior}{


You can control the width rule line for the printed subtables with an option:
\code{skimr_table_header_width}.
}

\seealso{
\code{\link[tibble:trunc_mat]{tibble::trunc_mat()}} For a list of global options for customizing
print formatting. \code{\link[crayon:has_color]{crayon::has_color()}} for the variety of issues that
affect tibble's color support.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{mutate.skim_df}
\alias{mutate.skim_df}
\title{Mutate a skim_df}
\usage{
\method{mutate}{skim_df}(.data, ...)
}
\arguments{
\item{.data}{A \code{skim_df}, which behaves like a \code{tbl.}}

\item{...}{Name-value pairs of expressions, each with length 1 or the same
length as the number of rows in the group, if using \code{\link[dplyr:group_by]{dplyr::group_by()}}, or
in the entire input (if not using groups). The name of each argument will
be the name of a new variable, and the value will be its corresponding
value. Use \code{NULL} value in \code{\link[dplyr:mutate]{dplyr::mutate()}} to drop a variable. New
variables overwrite existing variables of the same name.

The arguments in \code{...} are automatically quoted with \code{\link[rlang:defusing-advanced]{rlang::quo()}} and
evaluated with \code{\link[rlang:eval_tidy]{rlang::eval_tidy()}} in the context of the data frame. They
support unquoting \link[rlang:topic-inject]{rlang::quasiquotation} and splicing. See
\code{vignette("programming", package = "dplyr")} for an introduction to these
concepts.}
}
\value{
A \code{skim_df} object, which also inherits the class(es) of the input
data. In many ways, the object behaves like a \code{\link[tibble:tibble]{tibble::tibble()}}.
}
\description{
\code{\link[dplyr:mutate]{dplyr::mutate()}} currently drops attributes, but we need to keep them around
for other skim behaviors. Otherwise the behavior is exactly the same. For
more information, see \url{https://github.com/tidyverse/dplyr/issues/3429}.
}
\seealso{
\code{\link[dplyr:mutate]{dplyr::mutate()}} for the function's expected behavior.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skimr-package.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{\%>\%}
\alias{contains}
\alias{ends_with}
\alias{everything}
\alias{matches}
\alias{num_range}
\alias{one_of}
\alias{starts_with}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{magrittr}{\code{\link[magrittr:pipe]{\%>\%}}}

  \item{tidyselect}{\code{\link[tidyselect:starts_with]{contains}}, \code{\link[tidyselect:starts_with]{ends_with}}, \code{\link[tidyselect]{everything}}, \code{\link[tidyselect:starts_with]{matches}}, \code{\link[tidyselect:starts_with]{num_range}}, \code{\link[tidyselect]{one_of}}, \code{\link[tidyselect]{starts_with}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim.R
\name{skim}
\alias{skim}
\alias{skim_tee}
\alias{skim_without_charts}
\title{Skim a data frame, getting useful summary statistics}
\usage{
skim(data, ..., .data_name = NULL)

skim_tee(data, ..., skim_fun = skim)

skim_without_charts(data, ..., .data_name = NULL)
}
\arguments{
\item{data}{A tibble, or an object that can be coerced into a tibble.}

\item{...}{Columns to select for skimming. When none are provided, the
default is to skim all columns.}

\item{.data_name}{The name to use for the data. Defaults to the same as data.}

\item{skim_fun}{The skim function used.}

\item{skim}{The skimming function to use in \code{skim_tee()}.}
}
\value{
A \code{skim_df} object, which also inherits the class(es) of the input
data. In many ways, the object behaves like a \code{\link[tibble:tibble]{tibble::tibble()}}.
}
\description{
\code{skim()} is an alternative to \code{\link[=summary]{summary()}}, quickly providing a broad
overview of a data frame. It handles data of all types, dispatching a
different set of summary functions based on the types of columns in the data
frame.
}
\details{
Each call produces a \code{skim_df}, which is a fundamentally a tibble with a
special print method. One unusual feature of this data frame is pseudo-
namespace for columns. \code{skim()} computes statistics by data type, and it
stores them in the data frame as \verb{<type>.<statistic>}. These types are
stripped when printing the results. The "base" skimmers (\code{n_missing} and
\code{complete_rate}) are the only columns that don't follow this behavior.
See \code{\link[=skim_with]{skim_with()}} for more details on customizing \code{skim()} and
\code{\link[=get_default_skimmers]{get_default_skimmers()}} for a list of default functions.

If you just want to see the printed output, call \code{skim_tee()} instead.
This function returns the original data. \code{skim_tee()} uses the default
\code{skim()}, but you can replace it with the \code{skim} argument.

The data frame produced by \code{skim} is wide and sparse. To avoid type coercion
\code{skimr} uses a type namespace for all summary statistics. Columns for numeric
summary statistics all begin \code{numeric}; for factor summary statistics
begin \code{factor}; and so on.

See \code{\link[=partition]{partition()}} and \code{\link[=yank]{yank()}} for methods for transforming this wide data
frame. The first function splits it into a list, with each entry
corresponding to a data type. The latter pulls a single subtable for a
particular type from the \code{skim_df}.

\code{skim()} is designed to operate in pipes and to generally play nicely with
other \code{tidyverse} functions. This means that you can use \code{tidyselect} helpers
within \code{skim} to select or drop specific columns for summary. You can also
further work with a \code{skim_df} using \code{dplyr} functions in a pipeline.
}
\section{Customizing skim}{

\code{skim()} is an intentionally simple function, with minimal arguments like
\code{\link[=summary]{summary()}}. Nonetheless, this package provides two broad approaches to
how you can customize \code{skim()}'s behavior. You can customize the functions
that are called to produce summary statistics with \code{\link[=skim_with]{skim_with()}}.
}

\section{Unicode rendering}{

If the rendered examples show unencoded values such as \verb{<U+2587>} you will
need to change your locale to allow proper rendering. Please review the
\emph{Using Skimr} vignette for more information
(\code{vignette("skimr", package = "skimr")}).

Otherwise, we export \code{skim_without_charts()} to produce summaries without the
spark graphs. These are the source of the unicode dependency.
}

\examples{
skim(iris)

# Use tidyselect
skim(iris, Species)
skim(iris, starts_with("Sepal"))
skim(iris, where(is.numeric))

# Skim also works groupwise
iris |>
  dplyr::group_by(Species) |>
  skim()

# Which five numeric columns have the greatest mean value?
# Look in the `numeric.mean` column.
iris |>
  skim() |>
  dplyr::select(numeric.mean) |>
  dplyr::slice_head(n = 5)

# Which of my columns have missing values? Use the base skimmer n_missing.
iris |>
  skim() |>
  dplyr::filter(n_missing > 0)

# Use skim_tee to view the skim results and
# continue using the original data.
chickwts |>
  skim_tee() |>
  dplyr::filter(feed == "sunflower")

# Produce a summary without spark graphs
iris |>
  skim_without_charts()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skim.R
\name{skim}
\alias{skim}
\alias{skim_tee}
\alias{skim_without_charts}
\title{Skim a data frame, getting useful summary statistics}
\usage{
skim(data, ..., .data_name = NULL)

skim_tee(data, ..., skim_fun = skim)

skim_without_charts(data, ..., .data_name = NULL)
}
\arguments{
\item{data}{A tibble, or an object that can be coerced into a tibble.}

\item{...}{Columns to select for skimming. When none are provided, the
default is to skim all columns.}

\item{.data_name}{The name to use for the data. Defaults to the same as data.}

\item{skim_fun}{The skim function used.}

\item{skim}{The skimming function to use in \code{skim_tee()}.}
}
\value{
A \code{skim_df} object, which also inherits the class(es) of the input
data. In many ways, the object behaves like a \code{\link[tibble:tibble]{tibble::tibble()}}.
}
\description{
\code{skim()} is an alternative to \code{\link[=summary]{summary()}}, quickly providing a broad
overview of a data frame. It handles data of all types, dispatching a
different set of summary functions based on the types of columns in the data
frame.
}
\details{
Each call produces a \code{skim_df}, which is a fundamentally a tibble with a
special print method. One unusual feature of this data frame is pseudo-
namespace for columns. \code{skim()} computes statistics by data type, and it
stores them in the data frame as \verb{<type>.<statistic>}. These types are
stripped when printing the results. The "base" skimmers (\code{n_missing} and
\code{complete_rate}) are the only columns that don't follow this behavior.
See \code{\link[=skim_with]{skim_with()}} for more details on customizing \code{skim()} and
\code{\link[=get_default_skimmers]{get_default_skimmers()}} for a list of default functions.

If you just want to see the printed output, call \code{skim_tee()} instead.
This function returns the original data. \code{skim_tee()} uses the default
\code{skim()}, but you can replace it with the \code{skim} argument.

The data frame produced by \code{skim} is wide and sparse. To avoid type coercion
\code{skimr} uses a type namespace for all summary statistics. Columns for numeric
summary statistics all begin \code{numeric}; for factor summary statistics
begin \code{factor}; and so on.

See \code{\link[=partition]{partition()}} and \code{\link[=yank]{yank()}} for methods for transforming this wide data
frame. The first function splits it into a list, with each entry
corresponding to a data type. The latter pulls a single subtable for a
particular type from the \code{skim_df}.

\code{skim()} is designed to operate in pipes and to generally play nicely with
other \code{tidyverse} functions. This means that you can use \code{tidyselect} helpers
within \code{skim} to select or drop specific columns for summary. You can also
further work with a \code{skim_df} using \code{dplyr} functions in a pipeline.
}
\section{Customizing skim}{

\code{skim()} is an intentionally simple function, with minimal arguments like
\code{\link[=summary]{summary()}}. Nonetheless, this package provides two broad approaches to
how you can customize \code{skim()}'s behavior. You can customize the functions
that are called to produce summary statistics with \code{\link[=skim_with]{skim_with()}}.
}

\section{Unicode rendering}{

If the rendered examples show unencoded values such as \verb{<U+2587>} you will
need to change your locale to allow proper rendering. Please review the
\emph{Using Skimr} vignette for more information
(\code{vignette("Using_skimr", package = "skimr")}).

Otherwise, we export \code{skim_without_charts()} to produce summaries without the
spark graphs. These are the source of the unicode dependency.
}

\examples{
skim(iris)

# Use tidyselect
skim(iris, Species)
skim(iris, starts_with("Sepal"))
skim(iris, where(is.numeric))

# Skim also works groupwise
iris \%>\%
  dplyr::group_by(Species) \%>\%
  skim()

# Which five numeric columns have the greatest mean value?
# Look in the `numeric.mean` column.
iris \%>\%
  skim() \%>\%
  dplyr::select(numeric.mean) \%>\%
  dplyr::top_n(5)

# Which of my columns have missing values? Use the base skimmer n_missing.
iris \%>\%
  skim() \%>\%
  dplyr::filter(n_missing > 0)

# Use skim_tee to view the skim results and
# continue using the original data.
chickwts \%>\%
  skim_tee() \%>\%
  dplyr::filter(feed == "sunflower")

# Produce a summary without spark graphs
iris \%>\%
  skim_without_charts()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skimr-package.R
\docType{package}
\name{skimr-package}
\alias{skimr-package}
\alias{skimr}
\title{Skim a data frame}
\description{
This package provides an alternative to the default summary functions
within R. The package's API is tidy, functions take data frames, return
data frames and can work as part of a pipeline. The returned \code{skimr}
object is subsettable and offers a human readable output.
}
\details{
\code{skimr} is opinionated, providing a strong set of summary statistics
that are generated for a variety of different data types. It is also
provides an API for customization. Users can change both the functions
dispatched and the way the results are formatted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{deprecated-v1}
\alias{deprecated-v1}
\alias{skim_to_wide}
\alias{skim_to_list}
\alias{skim_format}
\title{Deprecated functions from skimr v1}
\usage{
skim_to_wide(.data, ...)

skim_to_list(.data, ...)

skim_format(...)
}
\arguments{
\item{.data}{A tibble, or an object that can be coerced into a tibble.}

\item{...}{Columns to select for skimming. When none are provided, the
default is to skim all columns.}
}
\value{
Either A \code{skim_df} or a \code{skim_list} object.
}
\description{
Skimr used to offer functions that combined skimming with a secondary effect,
like reshaping the data, building a list or printing the results. Some of
these behaviors are no longer necessary. \code{\link[=skim]{skim()}} always returns a wide
data frame. Others have been replaced by functions that do a single thing.
\code{\link[=partition]{partition()}} creates a list-like object from a skimmed data frame.
}
\section{Functions}{
\itemize{
\item \code{skim_to_wide()}: \code{\link[=skim]{skim()}} always produces a wide data frame.

\item \code{skim_to_list()}: \code{\link[=partition]{partition()}} creates a list.

\item \code{skim_format()}: \code{\link[=print]{print()}} and \code{\link[=skim_with]{skim_with()}} set options.

}}
